#include "qcalc.h"
#include "qwui.h"

using namespace QSAS;

extern qwuiWindow* qsasWindow;

// QcalcListVirtualItem class
// -----------------------------

/*!
    \class QcalcListVirtualItem
    \brief  This class inherits calculator QguListItem class. It includes all the parameters that are required for an object \n
	created at the calculator (either an operation or float constant).

	Those options are: the list of operands, the operator, working list name, virtual name (which is the operation name e.g. a + b),\n
	the angle unit (if trigonometric operation), vector component (if vector decomposition operation), a pointer to	the related tab \n
	(created when an operation is added to the list view), a flag that signifies if the object is a constant float and
	a QdObject_var member for storing either the result of the operation or the constant float value. The main member functions are: \n
	QcalcListVirtualItem::DoOperation() and QcalcListVirtualItem::DoMathOperation() that are recursively called from
	within the main loop of QcalcWindow::DoCalculation().
*/


QcalcListVirtualItem::QcalcListVirtualItem( QguListView   *parent,
                                            QString       txt,
		                                  QString       WL,
                                            QguListType   type,
	                                       QguListItem   *after )
              :  QguListItem( parent, txt, type, after )
{
     this->QcalcInitItem(txt, WL, type);
}


QcalcListVirtualItem::QcalcListVirtualItem( QguListItem *parent,
                                            QString      txt   ,
	                                       QString      WL    ,
	                                       QguListType  type  ,
	                                       QguListItem *after )
              :  QguListItem( parent, txt, type, after )
{
     this->QcalcInitItem(txt, WL, type);
}

void QcalcListVirtualItem::QcalcInitItem( QString txt, QString WL, QguListType  type )
{
     if ( QuSlotDebug() )
		printf("QcalcListVirtualItem::QcalcInitItem named %s \n", txt.latin1());
	
	 // initialise flag (virtualArrDim is a var ptr, so safe
	 this->isVector = false; 
	
     m_qstrOperator    = QString("");
     m_qdObjResult     = NULL;
     m_bCalcDone       = FALSE;
	 if(txt == "Restored") this->setText( Obj_Name_, "" );  
	 else this->setText( Obj_Name_, txt );     
     
	 QString qstrAlias = QString( this->GetNextAliaseName().latin1() );
     this->setText( Alias_, qstrAlias );

     if ( type == ListTypeEntry )
     {
		QcalcListView *view = dynamic_cast<QcalcListView *>(this->listView());
		if ( view == NULL ) return;
        
		m_pcalcTab = new QcalcTab;

		if ( !m_pcalcTab ) return; // failed to allocate space

		m_pcalcTab->nSlots = 0;
		m_pcalcTab->calcListview = view;
		this->m_pcalcTab->createCalcTabWidgets(this);
		m_pcalcTab->details_groupbox->setTitle( txt );
		view->owner->tabWidget->addTab( this->m_pcalcTab, qstrAlias );
		m_pcalcTab->setName( qstrAlias );
	
		if(WL != "") // object from working list
		{

			m_eObjType   = WORKING_LIST_OBJECT;
			WLcontext = QString( WL.latin1() ); 
			m_pcalcTab->operandStack[0]->setText( WLcontext );

			m_pcalcTab->SetOperator(strOpJoin);
			
			#ifdef QT_VERSION
                        #if QT_VERSION <0x030000
				if(view->hasFocus()) view->owner->tabWidget->showPage(m_pcalcTab); // object was dropped, not created by op 
			#else
				if(view->hasMouse()) view->owner->tabWidget->showPage(m_pcalcTab); // object was dropped, not created by 
			#endif // end of #if
			#endif // end of #ifdef
			QdObject_var dobj = this->getDataObject();
			
			if(!(dobj.is_nil()) ){
				if( QuIsThreeVector(dobj) ){
					this->isVector = true; 
				}
				else{
					QdRMatrix_var Mat_obj = QdRMatrix_var::narrow(dobj);
					QdRMatrixSeq_var MatSeq_obj = QdRMatrixSeq_var::narrow(dobj);
					if ( !Mat_obj.is_nil()  ) {
						this->virtualArrDim = new Dim(Mat_obj->dimdata());
					}
					else if(!MatSeq_obj.is_nil() ){
						this->virtualArrDim = new Dim(MatSeq_obj[0].dimdata());
					}
					// else virtualArrDim is_nil()
				}
			}
			// if Time Tag object is not set, use this
			if(  ( view->owner->_TagSource->text() == QString(" ") || view->owner->_TagSource->text().isEmpty() )
			   &&  view->owner->_TimeTagSelect->currentItem() == 1) {

				QdTimeSeq_var tobj = get_timetags(dobj);
				if(!(tobj.is_nil()) ){
		    			view->owner->_TagSource->setText(WLcontext);
					view->owner->_TagSource->validateInput(WLcontext);
				}
			}
		}
		else{   
			m_eObjType     = VIRTUAL_OBJECT;
			WLcontext = QString( "" );
			m_pcalcTab->SetOperator(strOpAdd);
			view->owner->tabWidget->showPage(m_pcalcTab);
			m_pcalcTab->saveWL_check->setChecked(true);
			m_pcalcTab->toggleWLName();
		}
		this->setChoices( );
		
	}
	else m_pcalcTab = NULL;
 	
}


QcalcListVirtualItem::~QcalcListVirtualItem()
{

}


void QcalcListVirtualItem::setChoices(  )
{
	QcalcListView *view = dynamic_cast<QcalcListView *>(this->listView());

	if ( view == NULL )
	{
		return;
	}
	m_pcalcTab->_algorithm->setCurrentItem(view->owner->_algorithm_Def);

	m_pcalcTab->_width_field->setText(view->owner->_width_field_Def.latin1());

	m_pcalcTab->_gap->setCurrentItem(view->owner->_gap_Def);

	m_pcalcTab->_gap_field->setText(view->owner->_gap_field_Def.latin1());
 
}


bool QcalcListVirtualItem::setArrayFlags(){

	// the flags specify the object type output by this item BEFORE component selection.
	// Use getIsVector() method to get the types as they are delivered by the item (WL items may have components selected)
	// return value indicates if the Array/Vector flags have changed
	
	if  ( QuSlotDebug() )
		printf("QcalcListVirtualItem::setArrayFlags %ld \n", this);

    bool oldVec = this->isVector;
    Dim_var oldArr;
    Dim_var ArrDim0;
    Dim_var ArrDim1;
    if ( !(this->virtualArrDim.is_nil()) ) oldArr = new Dim(*(this->virtualArrDim));

    if ( this->m_eObjType == VIRTUAL_OBJECT )
    {
		bool isVector0;
		bool isVector1;
    		// redetermine if output is vector or array for virtual item
		this->virtualArrDim = new Dim(); // empty dim
		this->isVector = false;

  		// try to resolve input types 

		QcalcListVirtualItem * qItem = dynamic_cast <QcalcListVirtualItem *> 
		  ( m_pcalcTab->calcListview->findItem(this->m_pcalcTab->operandStack[0]->slot->textObject(), Alias_) );
   		
		isVector0 = false;
		if( qItem) {
			isVector0 = qItem->getIsVector();
		
			if(isVector0){
				// allow for component selection in slot
				if(this->m_pcalcTab->operandStack[0]->slot->getComponent() != WL_OBJ) isVector0 = false;
			}
			else{
				// set array dim based on first input (so for no operator can change array dimensions)

				ArrDim0 = qItem->getVirtualArrDim();

			}
		}

	 	isVector1 = false;
		if (this->m_pcalcTab->nSlots > 1) {
			 qItem = dynamic_cast <QcalcListVirtualItem *> 
		  ( m_pcalcTab->calcListview->findItem(this->m_pcalcTab->operandStack[1]->slot->textObject(), Alias_) );
			if( qItem) {
				isVector1 = qItem->getIsVector(); 
			
				if(isVector1) {
					// allow for component selection in slot
					if(this->m_pcalcTab->operandStack[1]->slot->getComponent() != WL_OBJ) isVector1 = false;
				}
				else{
					// re-set array dim based on second input (1st may be scalar)
					ArrDim1 = qItem->getVirtualArrDim();
				}
			}
		}
	
		// set isVector flag feedback

		if( !ArrDim0.is_nil() &&  !ArrDim1.is_nil() ) {
		  if( ArrDim0->size() > ArrDim1->size() ) this->virtualArrDim = ArrDim0;
		  else  this->virtualArrDim = ArrDim1;
		}
		else if( !ArrDim0.is_nil() ) this->virtualArrDim = ArrDim0;
		else if( !ArrDim1.is_nil() ) this->virtualArrDim = ArrDim1;
		
		// set operand feedback

		if(this->m_qstrOperator == strOpAdd || this->m_qstrOperator == strOpSub){
			if( isVector0 && isVector1 ) this->isVector = true; 
		}
		else if(this->m_qstrOperator == strOpMult){
			if( isVector0 ||  isVector1) this->isVector = true; 
			if( isVector0 &&  isVector1) this->isVector = false;  // undo true for disallowed op
			
			if( !ArrDim0.is_nil() &&  !ArrDim1.is_nil() ) {
				// construct dim for product
				Dim_var vDim;
				
				// get reduced array from slot else keep dim from list object
				vDim = this->m_pcalcTab->operandStack[0]->slot->getSlotArrDim();
				if ( !vDim.is_nil() ) ArrDim0 = new Dim(*vDim); 
				
				vDim = this->m_pcalcTab->operandStack[1]->slot->getSlotArrDim();
				if ( !vDim.is_nil() ) ArrDim1 = new Dim(*vDim); 
				
				vector <int> dimList;
				if (ArrDim0->rank() > 0 && ArrDim1->rank() > 0) {
					for(int i=0; i < ArrDim0->rank()-1; i++){
						// ist array loses last dimension
						int elems0 = (*ArrDim0)[i];
						dimList.push_back(elems0);
					}
					for(int i=1; i < ArrDim1->rank(); i++){
						// second array loses 1st dimension
						int elems1 = (*ArrDim1)[i];
						dimList.push_back(elems1);
					}
					if(dimList.size() > 0){
						vDim = new Dim(dimList);	
						this->virtualArrDim = vDim;
					}
					else{
						Dim_var emptyDim;
						this->virtualArrDim = emptyDim; // scalar return
					}
				}
				else if ( ArrDim0->rank() <= 0 && ArrDim1->rank() > 0){
					this->virtualArrDim = ArrDim1;
				}
				else if ( ArrDim1->rank() <= 0 && ArrDim0->rank() > 0){
					this->virtualArrDim = ArrDim0;
				}
				else{
					Dim_var emptyDim;
					this->virtualArrDim = emptyDim; // scalar return
				} 
			}
		}
		else if(this->m_qstrOperator == strOpDiv){
			// only numerator may be vector
			if( isVector0 ) this->isVector = true; 
		}
		else if(this->m_qstrOperator == strOpMod){
			// always scalar
			Dim_var emptyDim;
			this->virtualArrDim = emptyDim;
		}
		else if(this->m_qstrOperator == strOpPower){
			// always scalar
			Dim_var emptyDim;
			this->virtualArrDim = emptyDim;
		}
		else if(this->m_qstrOperator == strOpChSign){
			if( isVector0 ) this->isVector = true; 
		}
		else if(this->m_qstrOperator == strOpCos){
		}
		else if(this->m_qstrOperator == strOpSin){
		}
		else if(this->m_qstrOperator == strOpTan){
		}
		else if(this->m_qstrOperator == strOpCot){
		}
		else if(this->m_qstrOperator == strOpCosh){
		}
		else if(this->m_qstrOperator == strOpSinh){
		}
		else if(this->m_qstrOperator == strOptanh){
		}
		else if(this->m_qstrOperator == strOpCoth){
		}
		else if(this->m_qstrOperator == strOpACos){
		}
		else if(this->m_qstrOperator == strOpASin){
		}
		else if(this->m_qstrOperator == strOpATan){
		}
		else if(this->m_qstrOperator == strOpACot){
		}
		else if(this->m_qstrOperator == strOpACosh){
		}
		else if(this->m_qstrOperator == strOpASinh){
		}
		else if(this->m_qstrOperator == strOpAtanh){
		}
		else if(this->m_qstrOperator == strOpACoth){
		}
		else if(this->m_qstrOperator == strOpVecProd){
			this->isVector = true; 
		}
		else if(this->m_qstrOperator == strOpDotProd){
		}
		else if(this->m_qstrOperator == strOpRemainder){
		}
		else if(this->m_qstrOperator == strOPAbsValue){
		}
		else if(this->m_qstrOperator == strOpLog){
		}
		else if(this->m_qstrOperator == strOpLog10){
		}
		else if(this->m_qstrOperator == strOpInvers){
		}
		else if(this->m_qstrOperator == strOpSqrt){
		}
		else if(this->m_qstrOperator == strOpExp){
		}
		else if(this->m_qstrOperator == strOpUnits){
			// returns same dimension object as input
			if( !ArrDim0.is_nil()  )  this->virtualArrDim = ArrDim0;
			if( isVector0 ) this->isVector = true; 
		}
		else if(this->m_qstrOperator == strOpFills){
			// returns same dimension object as input
			if( !ArrDim0.is_nil()  )  this->virtualArrDim = ArrDim0;
			if( isVector0 ) this->isVector = true; 
		}
    }
    else{
	 	 // for WL items don't change setting from validation
		QdObject_var dobj =QuGetObjByName(this->m_pcalcTab->operandStack[0]->slot->textObject());
		if(dobj.is_nil()){
			// nothing has changed, no object
			return false;
		}
		else{

			this->isVector = QuIsThreeVector(dobj);
			
			if(  !(this->isVector) && QuArrayDims(dobj) > 0 )  {

				QdRMatrix_var Mat_obj = QdRMatrix_var::narrow(dobj);
				QdRMatrixSeq_var MatSeq_obj = QdRMatrixSeq_var::narrow(dobj);
				if ( !Mat_obj.is_nil()  ) {
					this->virtualArrDim = new Dim(Mat_obj->dimdata());
				}
				else if(!MatSeq_obj.is_nil() ){
					this->virtualArrDim = new Dim(MatSeq_obj[0].dimdata());
				}
				// else virtualArrDim is_nil()

			}
		}
		return true; // this method does not change WL item flags, but may need to propogate selection
    }

    if( oldVec != this->isVector ) return true;
    if( oldArr.is_nil() != this->virtualArrDim.is_nil() ) return true; //array status changed
    if( !(oldArr.is_nil()) &&  !(this->virtualArrDim.is_nil()) && oldArr->size() != this->virtualArrDim->size() ) return true; // size changed
    return false; // no changes made
}
    

bool QcalcListVirtualItem::validateLink( bool valid )
{
	// checks that the WL context of an item still refers
	// to a valid item on the working list
	// folders don't change value of valid

	if ( this->getEntryType() != ListTypeEntry )
	{
		return valid;
	}

	// If the object is virtual, skip it
	if ( this->m_eObjType == VIRTUAL_OBJECT )
	{
		return valid;
	}

	int nCount = qsasWindow->WorkingList->Data->childCount();

	// if WL item is no longer there, set local item colour red
	QString ctxt = this->getWLcontext();

	if ( qsasWindow->WorkingList->getItemByName(ctxt) != NULL )
	{
		setColour( valid,true );
	}
	else
	{
		valid = false;
		setColour( valid, false );
	}

	return valid;
}



void QcalcListVirtualItem::SetItemVirtualName( bool bEmitChange )
{
    bool OKconst;
    
	if  ( QuSlotDebug() )
		printf("QcalcListVirtualItem::SetItemVirtualName %ld %ld\n", this, bEmitChange);

    QcalcListView *qview = ( QcalcListView * )this->listView();
    if ( qview == NULL || qview->owner == NULL) return;

    if(!this->m_pcalcTab) return;
    
    int nNumOperands = this->m_pcalcTab->nSlots;
    QString                qstrVirtName = "";
	
	if(m_qstrOperator == strOpJoin)
	{
		qstrVirtName = QString( m_pcalcTab->operandStack[0]->slot->text().latin1() );
		if (qstrVirtName.isEmpty()) qstrVirtName = QString("??");
	}
	else if ( m_qstrOperator == "" ) 
	{
		qstrVirtName = QString(" operation not set ");
	}
	else
	{
		qstrVirtName   = QString( m_qstrOperator.latin1() );
		for(int i=0; i < this->m_pcalcTab->nSlots; i++)
		{
		   	// is this operand set?
		   	int nIndCh = qstrVirtName.findRev( getOpChar(i) ); // must search in reverse in case alias contains OpChar
			if ( nIndCh != -1 )
			{
		 		m_pcalcTab->operandStack[i]->slot->ObjectName.toDouble(&OKconst); // OK =  numeric value for slot   
		   		if ( qview->findItem(m_pcalcTab->operandStack[i]->slot->ObjectName, Alias_) == 0 
				     && !OKconst
					&& !(m_pcalcTab->operandStack[i]->slot->acceptText) ) 
		   		{
					qstrVirtName.replace( nIndCh, 1, "?" );
		   		}
	   			else
				{
					// use actual text (with component) in slot not base object name
					qstrVirtName.replace( nIndCh, 1, m_pcalcTab->operandStack[i]->slot->text().latin1() );
				}
			}
			// else don't know how to rename
		}
	}

    if ( !qstrVirtName.isEmpty() )
    {
	   // Rename this item within the current list view of objects
		this->setText( Obj_Name_, qstrVirtName );
		// Set the text within the selected object group box
		this->m_pcalcTab->details_groupbox->setTitle( qstrVirtName );

		if ( bEmitChange ) emit qview->emitListModified();
    }
}


void QcalcListVirtualItem::DoOperation()
{
	if ( m_bCalcDone )  return;
		
	QcalcListView * qView = dynamic_cast <QcalcListView *> (this->listView());
	if(!qView) return;
		
	// Recursive call for virtual item to ensure inputs all exist first
	if ( m_eObjType == VIRTUAL_OBJECT )
	{  
		QcalcListVirtualItem*  qOperandItem = NULL;
 		for ( int nIndx = 0; nIndx < this->m_pcalcTab->nSlots; nIndx++ )
 		{	
			// is this operand set?
			QString Op_i = m_pcalcTab->operandStack[nIndx]->slot->textObject();
			qOperandItem = dynamic_cast< QcalcListVirtualItem* >( qView->findItem(Op_i, Alias_) );

			// slot must contain either a constant or an alias 
			if ( qOperandItem == NULL ) continue; // trap for value don't need to do op

			qOperandItem->DoOperation();
  		}

	}
	
	if ( !(this->m_pcalcTab->isOpAllowed()) ){
		// not allowed, but don't keep trying to do it
		m_bCalcDone = true;
		char bufLog[ 1024 ];
		sprintf(  bufLog,"Input(s) NOT VALID for operation %s -> %s",  text(Alias_).latin1(), text(Obj_Name_).latin1() );
		QcalcAppendTextDisplay( bufLog );
		qView->owner->messageHandler( QString("Alias ")+text(Alias_)+QString(" Invalid Inputs to operation"), true );
		return;
	}

    this->DoMathOperation();	
    
}

bool QcalcListVirtualItem::CheckOperatorSelfLink( QString& strAlias )
{
	bool                   bReturn      = true;
	QcalcListVirtualItem*  operandItem = NULL;

	if ( this->m_eObjType == WORKING_LIST_OBJECT )
	{
		return ( true );
	}

	if ( strAlias.isEmpty() )
	{
		strAlias = QString( this->text(Alias_).latin1() );
	}

 	QcalcTab* qTab  = this->m_pcalcTab;
	if (qTab == NULL) return true; // if no tab there are  no links to check
	
	for ( int nIndex = 0; nIndex < qTab->operandStack.size(); nIndex++ )
	{
		QString linkStr = qTab->operandStack[nIndex]->slot->text();
		if(linkStr == strAlias)
		{
			return false;  // self link found here, cascade this failure up the chain
		}
		else
		{
			QcalcListView *qview = ( QcalcListView * )this->listView();
			operandItem = dynamic_cast< QcalcListVirtualItem* >( qview->findItem(linkStr, Alias_));
			if ( operandItem )
			{
				// test links below on the chain
				bReturn = operandItem->CheckOperatorSelfLink( strAlias );
			}
		}
	}

	// this method does not test for unresolved links, only circular links
	return ( bReturn );
}

QdObject_var QcalcListVirtualItem::getResultObject()
{
	QdObject_var nil_obj;
	QcalcListView *qview = ( QcalcListView * )this->listView();
	if ( qview == NULL )  return nil_obj;
	if ( qview->owner == NULL )   return nil_obj;
   
	if(this->m_eObjType == VIRTUAL_OBJECT)
	{
		// Virtual Object
		return this->m_qdObjResult;
	}
	else
	{
		// WL Object
		if ( qview->owner->multiJn_check1->isChecked() )
		{ 
	         return this->m_qdObjJoinedFirst;
		}
		else
		{
		    // get the selected component for this WL item
		    // note this slot does not itself accept an alias so this is not an infinite recursion
		    return this->m_pcalcTab->operandStack[0]->slot->getSlotObject();
		}
	}

}
void QcalcListVirtualItem::DoMathOperation()
{
	QcalcTab* qTab  = NULL;
	qTab = this->m_pcalcTab;
	if (!qTab )  return;
	if ( qTab->nSlots == 0 ) return;

	QcalcListView *qview = ( QcalcListView * )this->listView();
	if ( qview == NULL )  return;
	if ( qview->owner == NULL )   return;

	vector <QdObject_var>   qd_Operands;
	QdOperator      qdOpDetails;
	QString         qstrError       = QString("");
	bool            joinOnly        = false;
	int             numOps          = 0;
    
	char bufLog[ 1024 ];
	sprintf(  bufLog,"Start Operation %s -> %s", text(Alias_).latin1(),  Q_LIST_VIEW_ITEM::text(Obj_Name_).latin1() );
	QcalcAppendTextDisplay( bufLog );
	
	sprintf(  bufLog,"End Operation %s -> %s", text(Alias_).latin1(),  Q_LIST_VIEW_ITEM::text(Obj_Name_).latin1() );
	
	if (m_qstrOperator == strOpJoin) {
		joinOnly = true;
	}
	else  if (qview->owner->m_mapOps.FindOperator( m_qstrOperator, qdOpDetails ) )
	{
		numOps = qdOpDetails.GetNumOperands();
	}
	else
	{
		qstrError = QString("Error: When performing operation '") + 
		            text(Obj_Name_) + 
				  QString("', cannot be found within the internal operator map");
		QcalcAppendTextDisplay( qstrError.latin1() );
		QcalcAppendTextDisplay( bufLog );
		return;
	}
	
	// Get Operand Objects: either joining now or already joined
	qd_Operands.clear();

	// Check if pair joining is required

	if ( qview->owner->multiJn_check1->isChecked() )
	{ 
		// already joined
		
		if(joinOnly){
		 	qd_Operands.push_back(this->m_qdObjJoinedFirst); // case that this tab is WL item
		}
		else for ( int nIndx = 0; nIndx < qTab->nSlots; nIndx++ )
		{
			QdObject_var operand_i = qTab->operandStack[nIndx]->slot->getSlotObject();
			if ( operand_i.is_nil() )
			{
				qstrError = QString("Error: When performing operation '" ) +
				            text(Obj_Name_) + 
						  QString("' - Operand  (" + qTab->operandStack[nIndx]->slot->text() + ") is null");
				QcalcAppendTextDisplay( qstrError.latin1() );
				QcalcAppendTextDisplay( bufLog );
				return;
			}
			else
			{
				qd_Operands.push_back(operand_i);
			}
	
		} // end of for loop over slots
	}
	else
	{
		// not multi-joined first, join arguments of this tab

		if ( qview->owner->joinSlotObjects(qTab,  qd_Operands)  )
		{
			// join failed
			qstrError = QString("Error: When joining for operation '") + text(Obj_Name_) + QString("'") ;
			QcalcAppendTextDisplay( qstrError.latin1() );
			QcalcAppendTextDisplay( bufLog );
			return;
		}
	} // end over if else on joined
	if(qd_Operands.size() == 0) return;
	
	if(joinOnly) // if join or component from pulldown, just return first operand
	{
	     m_qdObjResult = qd_Operands[0];
		m_bCalcDone = true;
		return;
	}
	// Do the calculation
	OperatorFnPtr pOpRole    = qdOpDetails.GetOpRole();
	QString       qstrErrMsg = QString("");

	// Check if this is a trig operation
	if ( qdOpDetails.GetTrigFlag() )
	{
		int  nAngleUnit  = 2;  // default to hyperbolic (no units) if angle units slot hidden
		if ( m_pcalcTab )
		{
			if ( !(m_pcalcTab->AngleUnits->isHidden()) )
			{
				nAngleUnit = m_pcalcTab->AngleUnits->pulldown->currentItem();
				if ( nAngleUnit == -1 )
				{
					nAngleUnit = 0; // default is degrees
				}
				
			}
		}
		// Construct operand B dynamically
		QdRScalar_var   qdUnit = new QdRScalar(nAngleUnit );
		if ( qdUnit.is_nil() )
		{
			// This should never happen - but ...
			qstrError = QString("Error: When performing operation '") + 
			            text(Obj_Name_) + 
					  QString("' - Memory cannot be allocated for angle units");
			QcalcAppendTextDisplay( qstrError.latin1() );
			QcalcAppendTextDisplay( bufLog );
			return;
		}

		try
		{
			m_qdObjResult = ( *( pOpRole ) )( qstrErrMsg, qd_Operands[0],   qdUnit   );

		}
		catch( Exception &e )
		{
			qstrError = QString("Error: When performing operation '") + text(Obj_Name_) ;
			QcalcAppendTextDisplay( qstrError.latin1() );
			QcalcAppendTextDisplay ( e.msg().c_str() );
			return;
		}
		catch( ... )
		{
			qstrError = QString("Error: When performing operation '") + 
			            text(Obj_Name_) + 
					  QString("' - Unkown exception thrown");
			QcalcAppendTextDisplay( qstrError.latin1() );
			QcalcAppendTextDisplay( bufLog );
			return;
		}
	}
	else
	{
		try
		{
			if(numOps == 2)
			{
				m_qdObjResult = ( *( pOpRole ) )( qstrErrMsg, qd_Operands[0], qd_Operands[1]  );
			}
			else if (numOps == 1)
			{
				m_qdObjResult = ( *( pOpRole ) )( qstrErrMsg, qd_Operands[0],  qd_Operands[0]  ); // second operand arg is ignored
			}
			else return; // unary and binary only so far. Will migrate to using vector of operands
		}
		catch( Exception &e )
		{
			qstrError = QString("Error: When performing operation '") + text(Obj_Name_) ;
			QcalcAppendTextDisplay( qstrError.latin1() );
			QcalcAppendTextDisplay ( e.msg().c_str() );
		return;
		}
		catch( ... )
		{
			qstrError = QString("Error: When performing operation '") + 
			            text(Obj_Name_) + 
					  QString("' - Unkown exception thrown");

			QcalcAppendTextDisplay( qstrError.latin1() );
			QcalcAppendTextDisplay( bufLog );
			return;
		}
	}

	if ( m_qdObjResult.is_nil() )
	{
		// Warning within the window status
		qstrError = QString("Error: When performing operation '") + 
		text(Obj_Name_) + 
		QString("' - Result is null");
		QcalcAppendTextDisplay( qstrError.latin1() );
	}

	m_bCalcDone = true;
	QcalcAppendTextDisplay( bufLog );
}

QString  QcalcListVirtualItem::GetNextAliaseName()
{
	QString  qstrAlias  = QString("a");
	bool     bAliasUsed;
     int      nAliasChar = 97;
	int      nCounter = 0;
	
	// search for next unused alias starting from 'a' 
	// this allows us to reuse deleted aliases
	
	bAliasUsed = AliasAlreadyUsed( qstrAlias );
	
	while ( bAliasUsed )
	{
		if  ( nAliasChar > 122 /*90*/ )
		{
			nAliasChar = 97;
			nCounter++;
		}


		char chAlias = nAliasChar;
		if ( nCounter > 0 )
		{
			QString  qstrCounter;
			qstrAlias = QString( QChar( chAlias ) ) + qstrCounter.setNum( nCounter );
		}
		else
		{
			qstrAlias = QString( QChar( chAlias ) );
		}

		// keep incrementing until unused alias is found
		bAliasUsed = AliasAlreadyUsed( qstrAlias );
		nAliasChar++;
	}

	return ( qstrAlias );
}

bool QcalcListVirtualItem::AliasAlreadyUsed( QString name )
{
 	// alias name must be unique in whole calc list
	// go to list view and iterate over whole list

	QcalcListView * qView = dynamic_cast <QcalcListView *> (this->listView());
	if ( qView == NULL) return true; // can't find list so don't allow alias name

	Q_LIST_VIEW_ITEM_ITERATOR it( qView );
	while ( it.current() ) {
		QcalcListVirtualItem *item = dynamic_cast <QcalcListVirtualItem *> (it.current());
		if (item->text(Alias_) == name && item != this) return true;
		++it;
	}
    
	return false;
}


void QcalcListVirtualItem::reNameAlias( const QString& NewVal )
{
	if  ( QuSlotDebug() )
		printf("QcalcListVirtualItem::reNameAlias %ld of %s \n", this, NewVal.latin1());
		
	QString qstrOldAlias = QString(this->text( Alias_ ).latin1());
	QcalcTab* qTab = this->m_pcalcTab;
	QcalcListView *view = dynamic_cast<QcalcListView*> (this->listView());
	
	if( !qTab || !view) return;
	
	if ( this->getEntryType() != ListTypeEntry ) return;

	this->setText( Alias_, NewVal );

	if ( qTab )
	{
		if ( qTab->objAliase_field )
		{
			qTab->objAliase_field->setText( NewVal );
		}
		m_pcalcTab->calcListview->owner->tabWidget->changeTab( qTab, NewVal );
	}

	// Change all the objects depending on that alias
	Q_LIST_VIEW_ITEM_ITERATOR  it( view );
	QcalcListVirtualItem*  qItem = NULL;
	while ( it.current() )
	{	
		qItem = dynamic_cast< QcalcListVirtualItem* >( it.current() );
		if ( qItem )
		{
			if ( qItem->m_eObjType == VIRTUAL_OBJECT  )
			{
				qTab = qItem->m_pcalcTab;
				if(qTab) 
				{
					for(int nOp=0; nOp < qTab->nSlots; nOp++)
					{
					   if (qTab->operandStack[nOp]->slot->textObject() == qstrOldAlias)
				   	   {
				   	   	qTab->operandStack[nOp]->slot->setAliasKeepComp(NewVal);
				   	   }
					}
					qTab->SetTabFeedback();
				}
			}
		}
		++it;
		qItem = NULL;
	}

	return ;
}
 
QdObject_var QcalcListVirtualItem::getDataObject()
{
	// returns is_nil() object if not resolved
	// returns appropriate resultant object for this list item
	
	QdObject_var nil_obj;
	QdObject_var ret_obj;
	
	if ( this->m_pcalcTab == NULL
	  || this->getEntryType() != ListTypeEntry ) return nil_obj;
	
	if(this->m_eObjType == WORKING_LIST_OBJECT)
	{
		if (m_qdObjJoinedFirst.is_nil())
		{
			ret_obj = (this->m_pcalcTab->operandStack[0])->slot->getSlotObject();
		}
		else
		{
			ret_obj = m_qdObjJoinedFirst;
		}
	}
	else if( this->m_eObjType == VIRTUAL_OBJECT)
	{
		ret_obj = m_qdObjResult;
	}
	
	return ret_obj;
}


bool QcalcListVirtualItem::getIsVector( )
{
	if (this->m_eObjType == WORKING_LIST_OBJECT) {

		if ( this->isVector) {
			if(this->m_pcalcTab->operandStack[0]->slot->getComponent() == WL_OBJ
				|| this->m_pcalcTab->operandStack[0]->slot->getComponent() == SLICE_COMP){
			   return true;
			}
			else return false; // component selected
		}
	 	else return false; //  object on WL not a vector
	}
	else if (this->m_eObjType == VIRTUAL_OBJECT) return this->isVector;  // Virtual items don't deliver components

     else return false; 
}

bool QcalcListVirtualItem::getIsArray( )
{
	bool isArray = false;
	if(!(this->virtualArrDim.is_nil()) && this->virtualArrDim->rank() > 0 ) isArray = true;

	if (this->m_eObjType == WORKING_LIST_OBJECT) {

		if ( isArray) {
			dimTripleStack arrTriple = this->m_pcalcTab->operandStack[0]->slot->parseArrayReduce();
			if(arrTriple.size() == 0)  return true;  // input object, not reduced
			else {
				for(int i=0; i<arrTriple.size(); i++){
					if(arrTriple[i][0] == 0) return true;  // elements option chosen for this dim, still an array
				}
				return false; // all dimensions sliced or summed, now a scalar
			}
		}
	 	else return false; //  object on WL not an array
	}
	else if (this->m_eObjType == VIRTUAL_OBJECT) return isArray;  // Virtual items don't deliver components

     else return false; 
}

Dim_var QcalcListVirtualItem::getVirtualArrDim( )
{
	
	if (this->m_eObjType == WORKING_LIST_OBJECT) {
		Dim_var vDim;
		if ( !(this->virtualArrDim.is_nil()) && this->virtualArrDim->rank() > 0) {
			// construct dim for selection in ArrayReduce via text()

			vDim = this->m_pcalcTab->operandStack[0]->slot->getSlotArrDim();
		}
	 	if( vDim.is_nil() )return this->virtualArrDim;  // no sub-selection made from ArrayReduce
		else return vDim; 
	}
	
	else  return this->virtualArrDim;  // Virtual items don't deliver components

}


Vostream& WriteCalcItem( Vostream& v_os,   QcalcListVirtualItem& item )
{

	if ( item.m_eObjType == WORKING_LIST_OBJECT )
	{
       	v_os << item.WLcontext.latin1();
		item.m_pcalcTab->operandStack[0]->slot->with = item.with;
     	
	}
	v_os << item.m_qstrOperator.latin1();
	
	v_os << item.isVector;
	
	bool noArrayDim = item.virtualArrDim.is_nil();
	v_os << noArrayDim;
	if( !noArrayDim ){ 
		v_os << item.virtualArrDim;
	}
	
     v_os << item.m_pcalcTab->saveWL_check->isChecked();
	v_os << item.m_pcalcTab->_nameWL_field->text().latin1();
	v_os << item.m_pcalcTab->WLnameUnset;
	
	v_os << item.m_pcalcTab->tb4->text().latin1();
	
	v_os << item.m_pcalcTab->_width_field->text().latin1();
	v_os << item.m_pcalcTab->_algorithm->currentItem();
	v_os << item.m_pcalcTab->_gap_field->text().latin1();
	v_os << item.m_pcalcTab->_gap->currentItem();
	
	v_os << item.m_pcalcTab->nSlots;
	for(int i=0; i < item.m_pcalcTab->nSlots; i++)
	{
		v_os << *(item.m_pcalcTab->operandStack[i]->slot);
		v_os << item.m_pcalcTab->operandStack[i]->slot->slotObjVector;
	}
	v_os << *(item.m_pcalcTab->AngleUnits->slot);
	

   
	return ( v_os );
}

Vistream& ReadCalcItem( Vistream& v_is,   QcalcListVirtualItem& item )
{

	if ( item.m_eObjType == WORKING_LIST_OBJECT )
	{
       	string name_str;
       	v_is >> name_str;
      	item.WLcontext = QString(name_str.c_str());
 	}
	QcalcListView * qView = dynamic_cast <QcalcListView *> (item.listView());
	
	// Operator
	string strOperator;
	v_is >> strOperator;
	item.m_qstrOperator = QString( strOperator.c_str() );

	v_is >>  item.isVector;  
	
	if(qView->owner->qcalc_version > 1.1 ) {
		bool noArrayDim;
		v_is >> noArrayDim;
		if(!noArrayDim){
			Dim_var vArrDim = new Dim();
			v_is >> vArrDim;
			item.virtualArrDim = vArrDim;
		}
	}
	else {
		bool isArray;
		v_is >>  isArray;
		// not used, vitrualArrDim will be is_nil()
	}
	
	bool isChk;
	v_is >> isChk;
	item.m_pcalcTab->saveWL_check->setChecked(isChk);
	
	string nameWL;
	v_is >> nameWL;
	item.m_pcalcTab->_nameWL_field->setText(nameWL.c_str());
	item.m_NameForSave =  QString(nameWL.c_str());
	
	if(qView->owner->qcalc_version > 1.0 ) v_is >> item.m_pcalcTab->WLnameUnset;
	
	string opTxt;
	v_is >> opTxt;
	item.m_pcalcTab->SetOperator( QString(opTxt.c_str()) );

	string widthTxt;
	v_is >> widthTxt;
	item.m_pcalcTab->_width_field->setText(widthTxt.c_str());

	int alg;
	v_is >> alg;
	item.m_pcalcTab->_algorithm->setCurrentItem(alg);
	 
	string gapTxt;
	v_is >> gapTxt;
	item.m_pcalcTab->_gap_field->setText(gapTxt.c_str());

	int gap;
	v_is >> gap;
	item.m_pcalcTab->_gap->setCurrentItem(gap);
	 
	v_is >> item.m_pcalcTab->nSlots;
	for(int i=0; i < item.m_pcalcTab->nSlots; i++)
	{
		v_is >> *(item.m_pcalcTab->operandStack[i]->slot);
		v_is >> item.m_pcalcTab->operandStack[i]->slot->slotObjVector;
	}
	v_is >> *(item.m_pcalcTab->AngleUnits->slot);
	
	return ( v_is );
}


Vostream& operator<<( Vostream& v_os, QcalcListVirtualItem& item )
{

	v_os << (int)item.m_eObjType;
	v_os << item.with;
	v_os << (int)item.isLocked();
	v_os << (int)item.entryType;
	
	v_os << item.text(Obj_Name_).latin1();
	v_os << item.text(Alias_).latin1();
	
	if ( item.getEntryType() == ListTypeEntry )
	{
		WriteCalcItem( v_os, item );
	}

#if QT_VERSION < 0x040000
	QList< QcalcListVirtualItem > lst;
#else
        QList< QcalcListVirtualItem * > lst;
#endif
	QcalcListVirtualItem* childItem = dynamic_cast< QcalcListVirtualItem* >( item.firstChild() );
	int            cnt       = 0;
	while ( childItem )
	{
		lst.append( childItem );
		cnt++;
		childItem = dynamic_cast< QcalcListVirtualItem* >( childItem->nextSibling() );
	}

	v_os << cnt;
	for( int i = 0;i < cnt; i++ )
	{
		QcalcListVirtualItem* nextItem = lst.at( i );
		nextItem->with = item.with;
		v_os << nextItem;
	}

	return v_os;
}


Vistream& operator>>( Vistream& v_is, QcalcListVirtualItem& item )
{   
	int obj_type;
	v_is >> obj_type;
	item.m_eObjType=(ItemTypeEnum)obj_type;
	v_is >> item.with;
	int nLock;
	v_is >> nLock;
	item.setLocked( nLock );
     int entry_type;
     v_is >> entry_type;
	item.entryType=(QguListType)entry_type;

	string txt0;
	v_is >> txt0;
	item.setText(Obj_Name_, txt0.c_str());

	string txt1;
	v_is >> txt1;
	item.setText(Alias_, txt1.c_str() );
		
	if ( item.getEntryType() == ListTypeEntry )
	{
		ReadCalcItem( v_is,   item );
	}

	if ( item.getEntryType() == ListTypeGroup )
	{
		item.setOpen( true );
	}

	int cnt = 0;
	v_is >> cnt;
	for ( int i = 0; i < cnt; i++ )
	{
		QcalcListVirtualItem* childItem = new QcalcListVirtualItem( &item );
		v_is >> *childItem;
		childItem->reNameAlias(childItem->text(Alias_));
	}

	return v_is;

}

QcalcInput::QcalcInput( QWidget   *parent,  QString  name, QString label, QcalcTab *qTab )
#if QT_VERSION < 0x040000
              :  QHBox( parent, name)
#else
	:QWidget(parent)
#endif
{
    if(qTab == 0) return;
    
#if QT_VERSION >= 0x040000
    QHBoxLayout* layout=new QHBoxLayout;
#if QT_VERSION < 0x040300
    layout->setSpacing(QSAS_profiles->current()->GUImargins/2);
    layout->setMargin(QSAS_profiles->current()->GUImargins/2);
#else
    layout->setContentsMargins(0,0,0,0);
#endif
#endif
    
    inputLabel =  new QLabel( label+"  ", this, name);
    inputLabel->setAlignment(Qt::AlignRight);
    
    slot = new QguWLinput( this, name );
    slot->setMinimumWidth( 300 );
    slot->setMaxLength( 200 );
    slot->acceptAlias = true;
    slot->acceptNumber = true;
    slot->acceptTimeSlice = true;
    slot->aliasList = dynamic_cast<QcalcListView *>(qTab->virtualItem->listView());  
    slot->setEnabled( TRUE );
    slot->setText( "" );
    
    pulldown = new QComboBox(this, name);
    pulldown->hide();
    QToolTip::add(  this, "Drop object from WL or calculator list, or type in value" );
    connect(  slot, SIGNAL( validated_number( const QString & ) ), qTab,  SLOT( SetOperand( const QString & ) ) );
    connect(  slot, SIGNAL( validated_alias( const QString & ) ), qTab,  SLOT( SetOperand( const QString & ) ) );
    connect(  slot, SIGNAL( validated_TEXT( const QString & ) ), qTab,  SLOT( SetOperand( const QString & ) ) );
    connect(  slot, SIGNAL( validated_WL_OBJ( const QString & ) ), qTab,  SLOT( SetLink( const QString & ) ) );
    connect(  slot, SIGNAL( textChanged( const QString & ) ), qTab,  SLOT( updateFlags(  ) ) );
    
#if QT_VERSION >= 0x040000
    layout->addWidget(inputLabel, Qt::AlignVCenter);
    layout->addWidget(slot, Qt::AlignVCenter);
    layout->addWidget(pulldown, Qt::AlignVCenter);
    this->setLayout(layout);
    this->adjustSize();
#endif

}


void QcalcInput::setText(QString txt)
{
    slot->setText( txt );
}
 
void QcalcInput::setAnglePulldown()
{
  
  this->slot->hide();
  pulldown->insertItem( QString("degrees"), 0 );
  pulldown->insertItem( QString("radians"), 1 );
  this->pulldown->show();

}

