#include "qcalc.h"
#include "qwui.h"
#include <qprogressbar.h>

extern qwuiWindow* qsasWindow;

using namespace QSAS;

// CALC Window
//---------------


/*!
    \class QcalcWindow
    \brief The class that encapsulates the calculator main window. It is inherited from the QSAS QguWindow class.

	The main widgets included in this class are the widgets for the join tab page. In addition, it includes the QTabWidget
	that is used when adding dynamically the operation tab pages.
*/

QcalcWindow::QcalcWindow( QWidget *parent, const char *name )
  : QguWindow( parent,name ),
   _join_listview(  NULL )
{
  this->setPalette(qApp->palette());
  this->setFont(qApp->font());
  
  setAcceptDrops( true );
  setFocusPolicy( WHEEL_FOCUS );

  verbose    = false;

   _top_widget = new QWidget( this,"top_widget" );
   setCentralWidget( _top_widget );

   _top_gl = new QGridLayout( _top_widget, 3, 1 );
   _top_gl->setSpacing(1);
   _top_gl->setMargin(0);

  // splitter to manage database list view and data selection widget

   _top_splitter = new QSplitter( _top_widget );
   _top_gl->addWidget( _top_splitter, 0,0 );

  //  list view of objects to perform calculation on
  CreateObjectLisView();

  // Create the tab widget (Joining / Calculator )
  tabWidget      = new QTabWidget( _top_splitter );
  connect( tabWidget, SIGNAL( currentChanged( QWidget * ) ), _join_listview, SLOT( tabChanged( QWidget * ) ) );


  // Create the Feed-back group box
  CreateFeedBackGroupBox();

  // Button group box
#if QT_VERSION < 0x040000
   buttons_layout = new QHGroupBox( _top_widget, "buttons" );
  _top_gl->addWidget( buttons_layout, 3,0 );
  buttons_layout->setMargin(0);
  buttons_layout->setInsideSpacing(0);
  buttons_layout->setInsideMargin(1);
  QSizePolicy fixed( QSizePolicy::Fixed,QSizePolicy::Fixed );
#else
  buttons_groupbox = new QGroupBox( _top_widget, "buttons" );
  QHBoxLayout * buttons_layout=new QHBoxLayout;
  _top_gl->addWidget( buttons_groupbox, 3,0 );
  buttons_layout->setMargin(0);
  buttons_layout->setSpacing(0);
  QSizePolicy fixed( QSizePolicy::Fixed,QSizePolicy::Fixed );
  buttons_groupbox->setLayout(buttons_layout);
#endif


  // Log Area
#if QT_VERSION < 0x040000
  Q_H_GROUP_BOX* logLayout = new Q_H_GROUP_BOX( _top_widget, "log" );
  _top_gl->addWidget( logLayout, 4,0 );
  logLayout->setInsideMargin(0);

  logArea = new QTextEdit( logLayout, "textArea" );
  logArea->setMinimumHeight(125);
#else
  QGroupBox* logGroupbox = new QGroupBox( _top_widget, "log" );
  QHBoxLayout * logLayout=new QHBoxLayout;
  _top_gl->addWidget( logGroupbox, 4,0 );
  logLayout->setMargin(0);

  logArea = new QTextEdit( logGroupbox, "textArea" );
  logArea->setMinimumHeight(125);
  logLayout->addWidget(logArea);
  logGroupbox->setLayout(logLayout);
#endif


  //  buttons added in JOIN or EXPORT create windows

  // Create menus
  CreatePopUpMenus();

  messageHandler(" ", false);
  this->setMinimumSize( this->sizeHint() );

  this->fromSession = false;

  // Load the operators map
  m_mapOps.LoadAll();
  
  this->setDefaults(); // initialise values used for new items

  // insert a blank operation
  this->_join_listview->insertVirtualItem();

  // set this item to use defaults
  this->defaultAll();
  setMinimumHeight(800);
}



void QcalcWindow::CreateObjectLisView()
{
  //  list view of objects to perform join and calculation on
  _join_listview = new QcalcListView( _top_splitter, this, "Object List" );

  _join_listview->setAllColumnsShowFocus( TRUE );
  _join_listview->setRootIsDecorated( true );
  _join_listview->addColumn( "Object List", 350 );
  _join_listview->setMinimumWidth( 350 );
  _join_listview->addColumn( "Alias", 50 );

    connect(_join_listview, SIGNAL(viewWL_ObjRequest()), qsasWindow->WorkingList, SIGNAL(viewObjRequest()));
}

void QcalcWindow::keyPressEvent(QKeyEvent * e )
{  
  if(e->key() == Qt::Key_Escape){
	  if(   JoinBtn->isDefault()) {
		 JoinBtn->setDefault(false);
		 newCalcBtn->setDefault(true);
	  }
	  else if(newCalcBtn->isDefault()) {
		 newCalcBtn->setDefault(false);
		 dismissBtn->setDefault(true);
	  }

	  else if(dismissBtn->isDefault())  {
		 dismissBtn->setDefault(false);
		 JoinBtn->setDefault(true);
	  }
	  return;
  }
  if(e->key() == Qt::Key_Return ||  e->key() == Qt::Key_Enter) {
     if(   JoinBtn->isDefault()) DoCalculation();
	 else if(newCalcBtn->isDefault()) _join_listview->insertVirtualItem();
	 else if(dismissBtn->isDefault()) close(); 
     return;
  }

  // else pass on to list view 
  e->ignore(); // let selection change

}


void QcalcWindow::CreatePopUpMenus()
{

  _menu_bar->clear();
#if QT_VERSION < 0x040000
  QPopupMenu* file_menu = new QPopupMenu( this,"file_menu" );
  edit_menu = new QPopupMenu( this,"edit_menu" );
#else
  QMenu* file_menu = new QMenu( this);
  edit_menu = new QMenu( this);
#endif

  file_menu->clear();
  file_menu->insertItem("&Save", this, SLOT(save()), Qt::CTRL+Qt::Key_S);
  file_menu->insertItem("Save &As...", this, SLOT(saveAs()), Qt::CTRL+Qt::Key_A);
  file_menu->insertSeparator();
  file_menu->insertItem("&Include...", this, SLOT(include()), Qt::CTRL+Qt::Key_I);
  file_menu->insertItem("&Restore...", this, SLOT(restore()), Qt::CTRL+Qt::Key_R);
  file_menu->insertSeparator();
  file_menu->insertItem("Use Default Settings", this, SLOT(defaultAll()));
  _menu_bar->insertItem("File", file_menu );

  edit_menu->clear();

  //edit_menu->insertItem("&New Folder",    _join_listview, SLOT( makeGroup() ),         CTRL+Key_N );
  edit_menu->insertItem( "Re-&Name Object Alias",     _join_listview, SLOT( renameSelectedItemAlias() ),    Qt::CTRL+Qt::Key_N, RENAME_ALIAS_EDITMENU_MENU_ITEM  );
  edit_menu->insertSeparator();

  edit_menu->insertItem("&Copy",       _join_listview, SLOT( copy() ),      Qt::CTRL+Qt::Key_C );
  edit_menu->insertItem("Paste",       _join_listview, SLOT(paste()),                       Qt::CTRL+Qt::Key_V);
  edit_menu->insertItem("select &All", _join_listview, SLOT( selectAll() ), Qt::CTRL+Qt::Key_A );
  edit_menu->insertSeparator();

  edit_menu->insertItem("Copy Join Settings",  this,  SLOT(keepChoices()),  Qt::CTRL+Qt::Key_Left);
  edit_menu->insertItem("Apply Join Settings", this,  SLOT(applyChoices()), Qt::CTRL+Qt::Key_Right);
  edit_menu->insertSeparator();

  edit_menu->insertItem("Delete Item", _join_listview, SLOT(deleteListItems()), Qt::CTRL+Qt::Key_X);
  edit_menu->insertItem("Clear List",  _join_listview, SLOT(deleteAllItems()));
  edit_menu->insertSeparator();
	
  edit_menu->insertItem( "Move Down",      _join_listview, SLOT( moveItemDown() ) ,                Qt::CTRL+Qt::Key_Down);
  edit_menu->insertItem( "Move Up",        _join_listview, SLOT( moveItemUp() ) ,                  Qt::CTRL+Qt::Key_Up);

  
  _menu_bar->insertItem("Edit", edit_menu );

#if QT_VERSION < 0x040000
  options_menu = new QPopupMenu(this,"options_menu");
#else
  options_menu = new QMenu(this);
#endif
  options_menu->setCheckable(true);
  options_menu->clear();

  toggleID = options_menu->insertItem("Verbose", this, SLOT(verboseToggle()));
  _menu_bar->insertItem("Options", options_menu );

  _menu_bar->insertSeparator();   
#if QT_VERSION < 0x040000
  QPopupMenu * help = new QPopupMenu( this, "help_menu" );
#else
  QMenu * help = new QMenu( this );
#endif
  _menu_bar->insertItem( " Help", help );

  help->insertItem("QSAS &Manual",   this, SLOT( launchManual() ), Qt::CTRL+Qt::Key_M );
  help->insertSeparator();

  //help->insertItem("...Data Export", this, SLOT( launchManExport() ));
  help->insertItem("...Calculator",        this, SLOT( launchManJoin() ));
  help->insertSeparator();

  help->insertItem("Module Version", this, SLOT( launchVersion() ));
}


void QcalcWindow::CreateFeedBackGroupBox()
#if QT_VERSION < 0x040000
{
  QGroupBox* feedback_groupbox = new QGroupBox( 1, Qt::Horizontal, _top_widget, "feedback_groupbox");
  
  feedback_groupbox->setInsideSpacing(2);
  feedback_groupbox->setInsideMargin(4);
  feedback_groupbox->setMargin(2);
  feedback_groupbox->setTitle( "  Time Source Details  ");
   _top_gl->addWidget( feedback_groupbox, 1,0 );

  QWidget* multiJn_widget = new QWidget( feedback_groupbox, "multiJn_widget" );

  QGridLayout* multiJn_layout = new QGridLayout( multiJn_widget, 1, 2 );
  multiJn_layout->setSpacing( 1 );

  multiJn_check1 = new QCheckBox( "Multi-Join Before calculation", multiJn_widget, "multiJn_check1" );
  multiJn_layout->addWidget( multiJn_check1, 0, 0 );
  multiJn_check1->setChecked( false );

  multiJn_check2 = new QCheckBox( "Multi-Join After calculation", multiJn_widget, "multiJn_check2" );
  multiJn_layout->addWidget( multiJn_check2, 0, 1 );
  multiJn_check2->setChecked( false );
    
  QWidget* feedback_widget = new QWidget( feedback_groupbox,"feedback_widget" );

   // Time source
 
  QWidget* time_widget = new QWidget( feedback_groupbox,"time_widget" );

  QGridLayout* time_layout = new QGridLayout( time_widget, 4, 4 );
  time_layout->setSpacing( 1 );
  time_layout->setColStretch( 1, 10 );

  // Time interval label
  QLabel* TimeInterval_label = new QLabel( "Time Interval  ", time_widget, "TimeInterval_label");
  time_layout->addWidget( TimeInterval_label, 0, 0, Qt::AlignRight );

  // Time interval combo box
  _TimeIvlSelect = new QComboBox( false, time_widget, "TimeIvlSelect" );
  _TimeIvlSelect->insertItem( "from Data",   0 );
  _TimeIvlSelect->insertItem( "from Object", 1 );
  time_layout->addWidget( _TimeIvlSelect, 0, 1, Qt::AlignLeft );
  connect( _TimeIvlSelect, SIGNAL( activated( int ) ), this, SLOT( setTTivl( int ) ) );

  IvlSource_label = new QLabel( "  ", time_widget, "IvlSource_label" );
  time_layout->addWidget( IvlSource_label, 1, 0, Qt::AlignRight );

  _IvlSource = new QguWLinput( time_widget, "IvlSource" );
  QFontMetrics calc_font_mt(qApp->font());
   int char_width = calc_font_mt.width('m');

  _IvlSource->setMinimumWidth( 35*char_width );
  _IvlSource->setMaxLength( 100 );
  _IvlSource->active           = false;

  QToolTip::add( _IvlSource, "not active");
  time_layout->addWidget( _IvlSource, 1, 1, Qt::AlignLeft );

  QLabel* TimeTags_label = new QLabel("Time Tags  ",time_widget,"TimeTags_label");
  time_layout->addWidget( TimeTags_label, 2, 0, Qt::AlignRight );

  _TimeTagSelect = new QComboBox( false, time_widget, "TimeTagSelect" );
  _TimeTagSelect->insertItem( "Create Regular", 0 );
  _TimeTagSelect->insertItem( "from Object",    1 );
  _TimeTagSelect->setCurrentItem(1);

  time_layout->addWidget( _TimeTagSelect, 2, 1, Qt::AlignLeft );
  connect( _TimeTagSelect, SIGNAL(activated(int)), this, SLOT( setTTsource( int ) ) );

  TagSource_label = new QLabel( "Time Tag Object  ",time_widget,"TagSource_label" );
  time_layout->addWidget( TagSource_label, 3, 0, Qt::AlignRight );

  _TagSource = new QguWLinput( time_widget, "TagSource" );
  _TagSource->setMinimumWidth( 35*char_width );
  _TagSource->setMaxLength( 100 );
  setTTsource(1);
  
  connect( _join_listview, SIGNAL( pasteDetected() ), _TagSource,  SLOT( paste()            ));
  connect( _TagSource,     SIGNAL( validated_OK()  ),  this,       SLOT( setTimeFeedback()  ));

  QToolTip::add( _TagSource, "drag and drop any object with time tags");
  time_layout->addWidget( _TagSource, 3, 1, Qt::AlignLeft );

  connect( _join_listview, SIGNAL( pasteDetected() ),   _IvlSource,     SLOT( paste()           ) );
  connect( _join_listview, SIGNAL( raiseWindow()   ),   this,          SLOT( bringToFront()    ) );
  connect( _IvlSource,     SIGNAL( validated_OK()  ),   this,          SLOT( setTimeFeedback() ) );


  QLabel* Start_label = new QLabel("Start Time: ",time_widget,"Start_label");
  time_layout->addWidget( Start_label, 0, 2, Qt::AlignRight );

  Start_field = new QLabel("<Not yet set>",time_widget,"Start_field");
  time_layout->addWidget( Start_field, 0, 3, Qt::AlignLeft );

  QLabel* End_label = new QLabel("End Time: ",time_widget,"End_label");
  time_layout->addWidget( End_label, 1, 2, Qt::AlignRight );

  End_field = new QLabel("<Not yet set>",time_widget,"End_field");
  time_layout->addWidget( End_field, 1, 3, Qt::AlignLeft );

  QLabel* Ntags_label = new QLabel("     Number of Time Tags in Interval: ",time_widget,"Ntags_label");
  time_layout->addWidget( Ntags_label, 2, 2, Qt::AlignRight );

  Ntags_field = new QLabel("<Not yet set>",time_widget,"Ntags_field");
  time_layout->addWidget( Ntags_field, 2, 3, Qt::AlignLeft );

  QLabel* spacing_label = new QLabel("Time Tag Spacing (sec): ",time_widget,"spacing_label");
  time_layout->addWidget( spacing_label, 3, 2, Qt::AlignRight );

  spacing_field = new QLabel("<Not yet set>",time_widget,"spacing_field");
  time_layout->addWidget( spacing_field, 3, 3, Qt::AlignLeft );
}
#else
{
	QGroupBox* feedback_groupbox=new QGroupBox(_top_widget);
	QVBoxLayout * feedback_layout=new QVBoxLayout;
	feedback_layout->setMargin(QSAS_profiles->current()->GUImargins/2);
	feedback_layout->setSpacing(QSAS_profiles->current()->GUImargins);
	feedback_groupbox->setTitle( "  Time Source Details  ");
	
	// Checks for multijoin options
	
	QGridLayout* multiJn_layout = new QGridLayout;
	multiJn_layout->setSpacing(1);
	
	multiJn_check1 = new QCheckBox("Multi-Join Before calculation");
	multiJn_layout->addWidget( multiJn_check1, 0, 0 );
	multiJn_check1->setChecked( false );

	multiJn_check2 = new QCheckBox("Multi-Join After calculation");
	multiJn_layout->addWidget( multiJn_check2, 0, 1 );
	multiJn_check2->setChecked( false );
	
	feedback_layout->addLayout(multiJn_layout);
	
	// Time source
	
	QGridLayout * time_layout=new QGridLayout;
	time_layout->setSpacing( 1 );
	time_layout->setColStretch( 1, 10 );
	
	time_layout->addWidget(new QLabel("Time Interval"), 0, 0, Qt::AlignRight);
	
	_TimeIvlSelect = new QComboBox;
	_TimeIvlSelect->insertItem( "from Data",   0 );
	_TimeIvlSelect->insertItem( "from Object", 1 );
	connect( _TimeIvlSelect, SIGNAL( activated( int ) ), this, SLOT( setTTivl( int ) ) );
	time_layout->addWidget( _TimeIvlSelect, 0, 1, Qt::AlignLeft );
	
	IvlSource_label = new QLabel("  ");
	time_layout->addWidget( IvlSource_label, 1, 0, Qt::AlignRight );
	
	_IvlSource = new QguWLinput(_top_widget, "IvlSource" );
	QFontMetrics calc_font_mt(qApp->font());
	int char_width = calc_font_mt.width('m');
	_IvlSource->setMinimumWidth( 35*char_width );
	_IvlSource->setMaxLength( 100 );
	_IvlSource->active = false;
	QToolTip::add( _IvlSource, "not active");
	time_layout->addWidget( _IvlSource, 1, 1, Qt::AlignLeft );
	
	time_layout->addWidget( new QLabel("Time Tags"), 2, 0, Qt::AlignRight );
	
	_TimeTagSelect = new QComboBox;
	_TimeTagSelect->insertItem( "Create Regular", 0 );
	_TimeTagSelect->insertItem( "from Object",    1 );
	_TimeTagSelect->setCurrentItem(1);
	connect( _TimeTagSelect, SIGNAL(activated(int)), this, SLOT( setTTsource( int ) ) );
	time_layout->addWidget( _TimeTagSelect, 2, 1, Qt::AlignLeft );
	
	TagSource_label = new QLabel("Time Tag Object  ");
	time_layout->addWidget( TagSource_label, 3, 0, Qt::AlignRight );
	
	_TagSource = new QguWLinput(_top_widget, "TagSource" );
	_TagSource->setMinimumWidth( 35*char_width );
	_TagSource->setMaxLength( 100 );
	setTTsource(1);
	connect( _join_listview, SIGNAL( pasteDetected() ), _TagSource,  SLOT( paste()            ));
	connect( _TagSource,     SIGNAL( validated_OK()  ),  this,       SLOT( setTimeFeedback()  ));
	QToolTip::add( _TagSource, "drag and drop any object with time tags");
	time_layout->addWidget( _TagSource, 3, 1, Qt::AlignLeft );
	connect( _join_listview, SIGNAL( pasteDetected() ),   _IvlSource,     SLOT( paste()           ) );
	connect( _join_listview, SIGNAL( raiseWindow()   ),   this,          SLOT( bringToFront()    ) );
	connect( _IvlSource,     SIGNAL( validated_OK()  ),   this,          SLOT( setTimeFeedback() ) );
	
	time_layout->addWidget(new QLabel("Start Time:"), 0, 2, Qt::AlignRight );
	
	Start_field = new QLabel("<Not yet set>");
	time_layout->addWidget( Start_field, 0, 3, Qt::AlignLeft );

	QLabel* End_label = new QLabel("End Time: ");
	time_layout->addWidget( End_label, 1, 2, Qt::AlignRight );

	End_field = new QLabel("<Not yet set>");
	time_layout->addWidget( End_field, 1, 3, Qt::AlignLeft );

	QLabel* Ntags_label = new QLabel("     Number of Time Tags in Interval: ");
	time_layout->addWidget( Ntags_label, 2, 2, Qt::AlignRight );

	Ntags_field = new QLabel("<Not yet set>");
	time_layout->addWidget( Ntags_field, 2, 3, Qt::AlignLeft );

	QLabel* spacing_label = new QLabel("Time Tag Spacing (sec): ");
	time_layout->addWidget( spacing_label, 3, 2, Qt::AlignRight );

	spacing_field = new QLabel("<Not yet set>");
	time_layout->addWidget( spacing_field, 3, 3, Qt::AlignLeft );
	
	
	feedback_layout->addLayout(time_layout);
	
	// Done
	
	
	feedback_groupbox->setLayout(feedback_layout);
	_top_gl->addWidget(feedback_groupbox);
	
// 	Q3GroupBox* feedback_groupbox = new Q3GroupBox( 1, Qt::Horizontal, _top_widget, "feedback_groupbox");
//   
// 	feedback_groupbox->setInsideSpacing(2);
// 	feedback_groupbox->setInsideMargin(4);
// 	feedback_groupbox->setMargin(2);
// 	feedback_groupbox->setTitle( "  Time Source Details  ");
// 	_top_gl->addWidget( feedback_groupbox, 1,0 );
// 
// 	QWidget* multiJn_widget = new QWidget( feedback_groupbox, "multiJn_widget" );
// 
// 	QGridLayout* multiJn_layout = new QGridLayout( multiJn_widget, 1, 2 );
// 	multiJn_layout->setSpacing( 1 );
// 
// 	multiJn_check1 = new QCheckBox( "Multi-Join Before calculation", multiJn_widget, "multiJn_check1" );
// 	multiJn_layout->addWidget( multiJn_check1, 0, 0 );
// 	multiJn_check1->setChecked( false );
// 
// 	multiJn_check2 = new QCheckBox( "Multi-Join After calculation", multiJn_widget, "multiJn_check2" );
// 	multiJn_layout->addWidget( multiJn_check2, 0, 1 );
// 	multiJn_check2->setChecked( false );
//     
// 	QWidget* feedback_widget = new QWidget( feedback_groupbox,"feedback_widget" );
// 
//    // Time source
//  
// 	QWidget* time_widget = new QWidget( feedback_groupbox,"time_widget" );
// 
// 	QGridLayout* time_layout = new QGridLayout( time_widget, 4, 4 );
// 	time_layout->setSpacing( 1 );
// 	time_layout->setColStretch( 1, 10 );
// 
//   // Time interval label
// 	QLabel* TimeInterval_label = new QLabel( "Time Interval  ", time_widget, "TimeInterval_label");
// 	time_layout->addWidget( TimeInterval_label, 0, 0, Qt::AlignRight );
// 
//   // Time interval combo box
// 	_TimeIvlSelect = new QComboBox( false, time_widget, "TimeIvlSelect" );
// 	_TimeIvlSelect->insertItem( "from Data",   0 );
// 	_TimeIvlSelect->insertItem( "from Object", 1 );
// 	time_layout->addWidget( _TimeIvlSelect, 0, 1, Qt::AlignLeft );
// 	connect( _TimeIvlSelect, SIGNAL( activated( int ) ), this, SLOT( setTTivl( int ) ) );
// 
// 	IvlSource_label = new QLabel( "  ", time_widget, "IvlSource_label" );
// 	time_layout->addWidget( IvlSource_label, 1, 0, Qt::AlignRight );
// 
// 	_IvlSource = new QguWLinput( time_widget, "IvlSource" );
// 	QFontMetrics calc_font_mt(qApp->font());
// 	int char_width = calc_font_mt.width('m');
// 
// 	_IvlSource->setMinimumWidth( 35*char_width );
// 	_IvlSource->setMaxLength( 100 );
// 	_IvlSource->active           = false;
// 
// 	QToolTip::add( _IvlSource, "not active");
// 	time_layout->addWidget( _IvlSource, 1, 1, Qt::AlignLeft );
// 
// 	QLabel* TimeTags_label = new QLabel("Time Tags  ",time_widget,"TimeTags_label");
// 	time_layout->addWidget( TimeTags_label, 2, 0, Qt::AlignRight );
// 
// 	_TimeTagSelect = new QComboBox( false, time_widget, "TimeTagSelect" );
// 	_TimeTagSelect->insertItem( "Create Regular", 0 );
// 	_TimeTagSelect->insertItem( "from Object",    1 );
// 	_TimeTagSelect->setCurrentItem(1);
// 
// 	time_layout->addWidget( _TimeTagSelect, 2, 1, Qt::AlignLeft );
// 	connect( _TimeTagSelect, SIGNAL(activated(int)), this, SLOT( setTTsource( int ) ) );
// 
// 	TagSource_label = new QLabel( "Time Tag Object  ",time_widget,"TagSource_label" );
// 	time_layout->addWidget( TagSource_label, 3, 0, Qt::AlignRight );
// 
// 	_TagSource = new QguWLinput( time_widget, "TagSource" );
// 	_TagSource->setMinimumWidth( 35*char_width );
// 	_TagSource->setMaxLength( 100 );
// 	setTTsource(1);
//   
// 	connect( _join_listview, SIGNAL( pasteDetected() ), _TagSource,  SLOT( paste()            ));
// 	connect( _TagSource,     SIGNAL( validated_OK()  ),  this,       SLOT( setTimeFeedback()  ));
// 
// 	QToolTip::add( _TagSource, "drag and drop any object with time tags");
// 	time_layout->addWidget( _TagSource, 3, 1, Qt::AlignLeft );
// 
// 	connect( _join_listview, SIGNAL( pasteDetected() ),   _IvlSource,     SLOT( paste()           ) );
// 	connect( _join_listview, SIGNAL( raiseWindow()   ),   this,          SLOT( bringToFront()    ) );
// 	connect( _IvlSource,     SIGNAL( validated_OK()  ),   this,          SLOT( setTimeFeedback() ) );
// 
// 
// 	QLabel* Start_label = new QLabel("Start Time: ",time_widget,"Start_label");
// 	time_layout->addWidget( Start_label, 0, 2, Qt::AlignRight );
// 
// 	Start_field = new QLabel("<Not yet set>",time_widget,"Start_field");
// 	time_layout->addWidget( Start_field, 0, 3, Qt::AlignLeft );
// 
// 	QLabel* End_label = new QLabel("End Time: ",time_widget,"End_label");
// 	time_layout->addWidget( End_label, 1, 2, Qt::AlignRight );
// 
// 	End_field = new QLabel("<Not yet set>",time_widget,"End_field");
// 	time_layout->addWidget( End_field, 1, 3, Qt::AlignLeft );
// 
// 	QLabel* Ntags_label = new QLabel("     Number of Time Tags in Interval: ",time_widget,"Ntags_label");
// 	time_layout->addWidget( Ntags_label, 2, 2, Qt::AlignRight );
// 
// 	Ntags_field = new QLabel("<Not yet set>",time_widget,"Ntags_field");
// 	time_layout->addWidget( Ntags_field, 2, 3, Qt::AlignLeft );
// 
// 	QLabel* spacing_label = new QLabel("Time Tag Spacing (sec): ",time_widget,"spacing_label");
// 	time_layout->addWidget( spacing_label, 3, 2, Qt::AlignRight );
// 
// 	spacing_field = new QLabel("<Not yet set>",time_widget,"spacing_field");
// 	time_layout->addWidget( spacing_field, 3, 3, Qt::AlignLeft );
}
#endif




QcalcWindow::~QcalcWindow()
{
 	delete _IvlSource;
 	delete _TagSource;
 	m_mapOps.clear();
}

