/*-*- c++ -*-
// ------------------------------------------------------------------
//
// ISSI QSAS PlugIn : Generic Residue Analysis of Conservation Laws
// 
//
// Inputs :
// 	bfield		- vector time series with magnetic field
//	velocity	- vector time series with plasma velocity
//	density	        - scalar time series with plasma density
//      ptensor         - 6 scalar time series with the pressure tensor entries
//
// Outputs
//
// SEH, 5 Sep, 2002
//      31 May 2006, added MVA dB/dt
// AJA  24 Aug 2006, added defesive programming (input, unit checks etc.)
// SEH  08 Aug 2008, removed QT dependencies, replaced Qdos::invert with invert function
//
// ------------------------------------------------------------------*/
 
#include "GenericMinVar.h"
#include "GenericResidue.h"


QdRMatrix_var str2Vec(const char* s);

extern "C" QplugReturnStatus  GenericResidue  ( QplugArgList *);
using namespace QSAS;


QplugReturnStatus  GenericResidue  ( QplugArgList *call_list )
{ 
  long    nestsize, nseq=0,triage_cnt=0;
  int i,j,k,constrained=0, min_nest=0, nb=0;
  double deltat, avg_width;
  char SC_id[127] = "",txtbuf[127];

  bool inSup[10];
  bool doType[10]; 
  char* inSyms[10]={"B","dBdt","V","N","E","Q","P","",""};
  char* doSyms[10]={"MVAB","MVAdBdt","HT","MFR","MMR","MLMR","MER","MTER","MVAV","COM"};
  int  joinSeq[10];

  for (int i=0; i<10; i++) inSup[i] = doType[i] = false;

  QdRMatrixSeq_var  B;                                       // B = Magnetic field series
  QdRMatrixSeq_var  dB;                                      // dB = time derivative of magnetic field series
  QdRMatrixSeq_var  V;                                       // V = Velocity series
  QdRScalarSeq_var  N;                                       // N = Plasma density
  QdRMatrixSeq_var  E;                                       // E = E-field vector series
  QdRMatrixSeq_var  H;                                       // H = Heat flux vector series
  QdRMatrixSeq_var  P;                                       // P = Pressure tensor
  QdRScalarSeq_var  PS;                                      // PS = .. or pressure as scalar

  QdTimeSeq_var ttJoined;                                    // tt of joined V, N, B are from the pressure tensor
  QdTimeSeq_var tt_out;                                      // tt of masked values

  QdRMatrixSeq_var  Erpar, Erperp;                           // Residual electric field

//   QTime t;
//   t.start();


 // --- start by checking inputs 0 : Bfield -----------------

 QdRMatrixSeq_var check_B = QdRMatrixSeq_var::narrow(QarConvertUnitsTo((*call_list)[0], "1.0e-9>T"));   // get B
 if (!check_B.is_nil()) {
   inSup[0] = true;   
   nb = check_B->sequence_size();                             // get length of B 
   ttJoined = get_timetags((*call_list)[0]);                  // dafault : tt of joineds are from B;
   if (ttJoined.is_nil()) {
     QplugAppendTextDisplay("Bad or missing time tags in input data\n");
     return QPLUG_FAILURE;
   }
   QuIsTSRegular(ttJoined,&deltat);
   avg_width = 1.5 * deltat;
   printf("Sampling time = %f.\n",(double) deltat); 
   fflush(stdout);
   nseq+=1;
 }


    
   


 // ---- check for dB/dt, take from B if input == 1 ------------

 bool dBdt_supplied=false;
 QdRMatrixSeq_var check_dB = QdRMatrixSeq_var::narrow((*call_list)[1]);        // check if dB/dt supplied as TS...
 if (!check_dB.is_nil()) {
   nseq +=1;
 } else {
   QdRScalar_var check_dB_replace = QdRScalar_var::narrow((*call_list)[1]);      // if not..  check if dB/dt supplied as integer
   dBdt_supplied = !check_dB.is_nil();
   bool dBdt_replace = !check_dB_replace.is_nil() && ((int)(*check_dB_replace) == 1);  // if '1' => calc dB/dt from B
   if (!dBdt_supplied && deltat > 0.0 && dBdt_replace) {   
     QplugAppendTextDisplay("Calculating dB / dt from B field input\n");                      
     check_dB =  new QdRMatrixSeq(nb,RMatrix(Dim(3)));
     for (int i=1; i<nb-1; i++) {
       for (int j=0; j<3; j++) 
	 check_dB[i][j] = ( (double)(*check_B)[i+1][j] - (double)(*check_B)[i-1][j] ) / (2.0*deltat);
     }
     for (int j=0; j<3; j++) {
       check_dB[0][j] = (double)(*check_dB)[1][j];
       check_dB[nb-1][j] = (double)(*check_dB)[nb-2][j];
     }
     set_timetags((QdObject_var) check_dB, ttJoined);
     QdString_var SICxref = new QdString("1.0e-9>T s^-1");
     check_dB->change_xref("SI_conversion", SICxref);
     nseq +=1;
     inSup[1] = true;        
   }       
 }
 
 printf("nseq = %d\n",nseq);
 fflush(stdout);

 // ----- check V ------------

 QdRMatrixSeq_var check_V = QdRMatrixSeq_var::narrow(QarConvertUnitsTo((*call_list)[2], "1.0e3>m s^-1"));   // get V
 if (!check_V.is_nil()) {
   inSup[2] = true;   
   ttJoined = get_timetags((*call_list)[2]);                  // if V supplied, tt of joineds are from V;
   if (ttJoined.is_nil()) {
     QplugAppendTextDisplay("Bad or missing time tags in input data\n");
     return QPLUG_FAILURE;
   }
   QuIsTSRegular(ttJoined,&deltat);
   avg_width = 1.5 * deltat;
   nseq +=1;
 } else QplugAppendTextDisplay("WARNING: cannot confirm V units  or no V given");



 // ----- check N ------------

 QdRScalarSeq_var check_N = QdRScalarSeq_var::narrow(QarConvertUnitsTo((*call_list)[3], "1.0e6>m^-3"));   // get N
 if (!check_N.is_nil()) {
   inSup[3] = true;   
   nseq +=1;
 } else QplugAppendTextDisplay("WARNING: cannot confirm that N units are 1.0e6m^-3 or no N given");		


 // ----- check E ------------

 QdRMatrixSeq_var check_E = QdRMatrixSeq_var::narrow(QarConvertUnitsTo((*call_list)[4], "1.0e3>m s^-1"));   // get E
 if (!check_E.is_nil()) {
   nseq +=1;
   inSup[4] = true;   
 } else {
   if(check_E.is_nil() || check_E->sequence_size() == 1) {
     check_E = QdRMatrixSeq_var::narrow(QarConvertUnitsTo((*call_list)[4], "1.0e-6>m s^-1 T")); // try units from VxB
     if(check_E.is_nil() || check_E->sequence_size() == 1) {
       QplugAppendTextDisplay("WARNING: cannot confirn E units are either 1.0e3>V m^-1  or 1.0e-6>m s^-1 T \n");
     } else {
       nseq +=1;
       inSup[4] = true;   
     }
   }
 }


 // ----- check Q - heat flux vector ------------

 QdRMatrixSeq_var check_Q = QdRMatrixSeq_var::narrow((*call_list)[5]);
 if (!check_Q.is_nil()) {
   nseq +=1;
   inSup[5] = true;   
 } else QplugAppendTextDisplay("WARNING: cannot confirm Q units are 1.0e6m^-3 or no Q given");		


 // ----- check P pressure ------------

 QdRScalarSeq_var check_P = QdRScalarSeq_var::narrow((*call_list)[6]);   // get P scalar
 if (!check_P.is_nil()) {
   nseq +=1;
   inSup[6] = true;   
 } else QplugAppendTextDisplay("WARNING: cannot confirm that P units are eV m^-6 or no P given");		


 // ----- Matrix of what to do - depends on supplied inputs -- 

 doType[0] = inSup[0];                                                               // MVAB
 doType[1] = inSup[1];                                                               // MVAdBdt
 doType[2] = inSup[0] && inSup[2];                                                   // HT needs B and (V or E)
 doType[3] = inSup[0] && inSup[2];                                                   // MFR needs B and V
 doType[4] = inSup[2] && inSup[3];                                                   // MMR needs N and V
 doType[5] = inSup[0] && inSup[2] && inSup[3] && inSup[6];                           // MLMR : N,V,N,P
 doType[6] = inSup[0] && inSup[2] && inSup[3] && inSup[6];                           // MER : N,V,N,P
 doType[7] = inSup[0] && inSup[2] && inSup[3] && inSup[4]  && inSup[5] && inSup[6];  // MTER : N,V,N,E,Q,P
 doType[8] = inSup[2];                                                               // MVAV requires V;

 for (int i=0; i<10; i++) {
   sprintf(txtbuf,"insup[%d] : %d\n",i,(int)(inSup[i] == 1)); 
   QplugAppendTextDisplay(txtbuf);
 }

 sprintf(txtbuf,"Joining %d sequences\n\nDoing:",nseq);
   QplugAppendTextDisplay(txtbuf);

 for (int i=0; i<10; i++) {
   if(doType[i]) {
     sprintf(txtbuf,"%s\n",doSyms[i]); 
     QplugAppendTextDisplay(txtbuf);
   }
 }


 // ---- now, join the data -----

 QdObjectSeq_var ttIn    = new QdObjectSeq(nseq);
 QdObjectSeq_var toJoin  = new QdObjectSeq(nseq);
 QdObjectSeq_var joined  = new QdObjectSeq(nseq);
 QdObjectSeq_var restSig = new QdObjectSeq(nseq);
 QdObjectSeq_var joinMethod = new QdObjectSeq(nseq);
 QdObjectSeq_var gapValue   = new QdObjectSeq(nseq);
 QdObjectSeq_var gapHandler = new QdObjectSeq(nseq);

 int seq=0;
 if (inSup[0]) {
   toJoin[seq] = check_B;
   ttIn[seq] = get_timetags(check_B);
   ttJoined = get_timetags(check_B); 
   joinSeq[0]=seq;
   seq++;
 }
 if (inSup[1]) {
   toJoin[seq] = check_dB;
   ttIn[seq] = get_timetags(check_dB);
   if (!inSup[0]) ttJoined = get_timetags(check_dB); 
   joinSeq[1]=seq;
   seq++;
 }
 if (inSup[2]) {
   toJoin[seq] = check_V;
   ttIn[seq] = get_timetags(check_V);
   ttJoined = get_timetags(check_V); 
   joinSeq[2]=seq;
   seq++;
 }
 if (inSup[3]) {
   toJoin[seq] = check_N;
   ttIn[seq] = get_timetags(check_N);
   joinSeq[3]=seq;
   seq++;
 }
 if (inSup[4]) {
   toJoin[seq] = check_E;
   ttIn[seq] = get_timetags(check_E);
   joinSeq[4]=seq;
   seq++;
 }
 if (inSup[5]) {
   toJoin[seq] = check_Q;
   ttIn[seq] = get_timetags(check_Q);
   joinSeq[5]=seq;
   seq++;
 }
 if (inSup[6]) {
   toJoin[seq] = check_P;
   ttIn[seq] = get_timetags(check_P);
   joinSeq[6]=seq;
   seq++;
 }


 for (i=0; i<nseq; i++) {
   joined[i]  = toJoin[i]->create();
   joinMethod[i] = new QdString(LINEAR);
   gapValue[i] = new QdRScalar(avg_width);
   gapHandler[i] = new QdString(LINEAR);
 }

  try {
   multijoin(ttIn, toJoin, ttJoined, joined,
      joinMethod, gapValue, gapHandler, triage_cnt);
 }
 catch(Exception &e) {
   e.print_msg();
   return QPLUG_FAILURE;
 }

 printf("Join OK..\n");
 fflush(stdout);

 QdTimeInterval_var start_end = get_timeinterval((* call_list)[7]);
 if(start_end.is_nil()){
     QplugAppendTextDisplay("Error : Time interval not recognised ");
     return (QPLUG_FAILURE);
 }
 
 SequenceMask msk(ttJoined->get_mask(QdTimeSeq::InRange(start_end->start(), start_end->end())));
 tt_out = new QdTimeSeq();
 ttJoined->get_subsequence(*tt_out, msk);
 if(tt_out.is_nil() || tt_out->sequence_size() <= 0 ){
     QplugAppendTextDisplay("Error : No time tags in range ");
     return (QPLUG_FAILURE);
 }

 if (inSup[0]) {
   printf("joined B..\n");
   fflush(stdout);
   QdRMatrixSeq_var joined_B = QdRMatrixSeq_var::narrow(joined[joinSeq[0]]);
   QdRMatrixSeq_var masked_B = new QdRMatrixSeq();
   joined_B->get_subsequence(*masked_B,msk);
   if(masked_B.is_nil() || masked_B->sequence_size() <= 0){
     QplugAppendTextDisplay("Error : No B field in range ");
     return (QPLUG_FAILURE);
   }
   masked_B->copy_xrefs_from((* call_list)[0] );
   B = QdRMatrixSeq_var::narrow(QarConvertUnitsTo(masked_B, "1.0e-9>T")); // ensure units are in those assumed
   if(B.is_nil() || B->sequence_size() == 1) {
     QplugAppendTextDisplay("WARNING: Cannot confirm B units are 1.0e-9>T\n");
     B = (QdRMatrixSeq_var) masked_B;
   }    
   nestsize = B->sequence_size();
   strncpy(SC_id,QuGetAttrText("SC_id",(*call_list)[0]),1);		//get SC ID
 }

 if (inSup[1]) {
   printf("joined ddtB..\n");
   fflush(stdout);
   QdRMatrixSeq_var joined_dB = QdRMatrixSeq_var::narrow(joined[joinSeq[1]]);
   QdRMatrixSeq_var masked_dB = new QdRMatrixSeq();
   joined_dB->get_subsequence(*masked_dB,msk);
   if(masked_dB.is_nil() || masked_dB->sequence_size() <= 0){
     QplugAppendTextDisplay("Error : No dB/dt data in range ");
     return (QPLUG_FAILURE);
   }
   if(dBdt_supplied) masked_dB->copy_xrefs_from((* call_list)[1] );
   else masked_dB->copy_xrefs_from(check_dB );
   
   dB = QdRMatrixSeq_var::narrow(QarConvertUnitsTo(masked_dB, "1.0e-9>T s^-1")); // ensure units are in those assumed
   if(dB.is_nil() || dB->sequence_size() == 1) {
     QplugAppendTextDisplay("WARNING: Cannot confirm dB units are 1.0e-9>T s^-1\n");
     dB = (QdRMatrixSeq_var) masked_dB;
   } 
   strncpy(SC_id,QuGetAttrText("SC_id",(*call_list)[1]),1);		//get SC ID
   nestsize = dB->sequence_size();
 }

 if (inSup[2]) {
   printf("joined V..\n");
   fflush(stdout);
   QdRMatrixSeq_var joined_V = QdRMatrixSeq_var::narrow(joined[joinSeq[2]]);
   QdRMatrixSeq_var masked_V = new QdRMatrixSeq();
   joined_V->get_subsequence(*masked_V,msk);
   if(masked_V.is_nil() || masked_V->sequence_size() <= 0){
     QplugAppendTextDisplay("Error : No V data in range ");
     return (QPLUG_FAILURE);
   } 
   masked_V->copy_xrefs_from((* call_list)[2] );
   V = QdRMatrixSeq_var::narrow(QarConvertUnitsTo(masked_V, "1.0e3>m s^-1")); // ensure units are in those assumed
   if(V.is_nil() || V->sequence_size() == 1) {
     QplugAppendTextDisplay("WARNING: Cannot confirm N units are 1.0e6>m^-3\n");
     V = (QdRMatrixSeq_var) masked_V;
   }     
   strncpy(SC_id,QuGetAttrText("SC_id",(*call_list)[2]),1);		//get SC ID
   nestsize = V->sequence_size();
 }

 if (inSup[3]) {
   printf("joined N..\n");
   fflush(stdout);
   QdRScalarSeq_var joined_N = QdRScalarSeq_var::narrow(joined[joinSeq[3]]);
   QdRScalarSeq_var masked_N = new QdRScalarSeq();
   joined_N->get_subsequence(*masked_N,msk);
   if(masked_N.is_nil() || masked_N->sequence_size() <= 0){
     QplugAppendTextDisplay("Error : No N data in range ");
     return (QPLUG_FAILURE);
   } 
   masked_N->copy_xrefs_from((* call_list)[3] );
   N = QdRScalarSeq_var::narrow(QarConvertUnitsTo(masked_N, "1.0e6>m^-3")); // ensure units are in those assumed
   if(N.is_nil() || N->sequence_size() == 1) {
     QplugAppendTextDisplay("WARNING: Cannot confirm N units are 1.0e6>m^-3\n");
     N = (QdRScalarSeq_var) masked_N;
   }  
 }

 if (inSup[4]) {
   QdRMatrixSeq_var joined_E = QdRMatrixSeq_var::narrow(joined[joinSeq[4]]);
   QdRMatrixSeq_var masked_E = new QdRMatrixSeq();
   joined_E->get_subsequence(*masked_E,msk);
   if(masked_E.is_nil() || masked_E->sequence_size() <= 0){
     QplugAppendTextDisplay("Error : No E data in range ");
     return (QPLUG_FAILURE);
   }
   masked_E->copy_xrefs_from((* call_list)[4] );
   E = QdRMatrixSeq_var::narrow(QarConvertUnitsTo(masked_E, "1.0e3>V m^-1")); // ensure units are in those assumed
   if(E.is_nil() || E->sequence_size() == 1) {
     E = QdRMatrixSeq_var::narrow(QarConvertUnitsTo(masked_E, "1.0e-6>m s^-1 T")); // try VxB
     if(E.is_nil() || E->sequence_size() == 1) {
       QplugAppendTextDisplay("WARNING: cannot confirn E units are either 1.0e3>V m^-1  or 1.0e-6>m s^-1 T \n");
       E = (QdRMatrixSeq_var) masked_E;
     }
   }  
 }
  
 if (inSup[5]) {
   QdRMatrixSeq_var joined_H = QdRMatrixSeq_var::narrow(joined[joinSeq[5]]);
   QdRMatrixSeq_var masked_H = new QdRMatrixSeq();
   joined_H->get_subsequence(*masked_H,msk);
   if(masked_H.is_nil() || masked_H->sequence_size() <= 0){
     QplugAppendTextDisplay("Error : No Heat Flux in range ");
     return (QPLUG_FAILURE);
   }
   masked_H->copy_xrefs_from((* call_list)[5] );
   H = QdRMatrixSeq_var::narrow(QarConvertUnitsTo(masked_H, "1.0>J m^-2 s^-1")); // ensure units are in those assumed
   if(H.is_nil() || H->sequence_size() == 1) {
     QplugAppendTextDisplay("WARNING: Cannot confirm H units are 1.0>J m^-2 s^-1\n");
     H = (QdRMatrixSeq_var)  masked_H;
   }  
 }

 
 if (inSup[6]) {
   QdRMatrixSeq_var joined_P = QdRMatrixSeq_var::narrow(joined[joinSeq[6]]);                // assume P is a tensor
   if (joined_P.is_nil()) {                                                        // ..if not, then scalar pressure
     QdRScalarSeq_var joined_PS = QdRScalarSeq_var::narrow(joined[joinSeq[6]]);
     if (joined_PS.is_nil()) {                                                     // ..if neither tensor or scalar -> failure
       QplugAppendTextDisplay("Error : P is neither scalar or tensor ");
       return (QPLUG_FAILURE);
     }
     QdRScalarSeq_var masked_PS = new QdRScalarSeq();                              // treat as scalar for the moment...
     joined_PS->get_subsequence(*masked_PS,msk); 
     if(masked_PS.is_nil() || masked_PS->sequence_size() <= 0){
       QplugAppendTextDisplay("Error : No Scalar pressure in range ");
       return (QPLUG_FAILURE);
     }
     
     masked_PS->copy_xrefs_from((* call_list)[6] );
     PS = QdRScalarSeq_var::narrow(QarConvertUnitsTo(masked_PS, "1.0e12>J m^-3")); // ensure units are in those assumed
     if(PS.is_nil() || PS->sequence_size() == 1) {
       QplugAppendTextDisplay("WARNING: Cannot confirm Pressure units are 1.0e12>J m^-3\n");
       PS = (QdRScalarSeq_var)  masked_PS;
     }  
     P = new QdRMatrixSeq(nestsize,RMatrix(Dim(3,3),0.0));                         // ..construct empty symmetric P matrix 
     printf("P is scalar - filling diagonal....\n");
	 fflush(stdout);
     for (k=0; k<nestsize; k++){
       for (i=0; i<3; i++) 
	 for (j=0; j<3; j++) 
	   (*P)[k][i][i] = (double)(*PS)[k];                                       // .. and fill the trace of P with scalar p
     }
   } 
   else {
     QdRMatrixSeq_var masked_P = new QdRMatrixSeq();
     joined_P->get_subsequence(*masked_P,msk);
     if(masked_P.is_nil() || masked_P->sequence_size() <= 0){
       QplugAppendTextDisplay("Error : No pressure tensor in range ");
       return (QPLUG_FAILURE);
     }
     masked_P->copy_xrefs_from((* call_list)[6] );
     P = QdRMatrixSeq_var::narrow(QarConvertUnitsTo(masked_P, "1.0e12>J m^-3")); // ensure units are in those assumed
     if(P.is_nil() || P->sequence_size() == 1) {
       QplugAppendTextDisplay("WARNING: Cannot confirm P units are 1.0e12>J m^-3\n");
       P = (QdRMatrixSeq_var)  masked_P;
     }    
     printf("P is tensor....\n");
	 fflush(stdout);
   } 
 }


 // ---- Get minimum nest size (if 0 - single nest size) ------------------

 QdObject_var in_7 = (* call_list)[8];
 QdRScalar_var minnest_do = QdRScalar_var::narrow(in_7);
 if( minnest_do.is_nil()) {
   min_nest = 0;
 } else  {
   min_nest  = ((int) *minnest_do);
   if (min_nest != 0 && min_nest <= 3) min_nest = 3;
 }
 if (min_nest <= nestsize) min_nest=0;  // do single nest if insufficeient samples.

 // ---- Get type MVA (constrained or normal) ------------------

 QdObject_var in_8 = (* call_list)[9];
 QdRScalar_var mvatype_do = QdRScalar_var::narrow((*call_list)[9]);
 double alpha=1.0;   // walenslope
 if( mvatype_do.is_nil() ) {
     QdString_var c_string = QdString_var::narrow((*call_list)[9]);
     if (!c_string.is_nil()) {
       QdRMatrix_var c_vector = new QdRMatrix(Dim(3),1.00);
       QString s = c_string->c_str();
       c_vector = str2Vec(s);
       constrained = (int) (*c_vector)[0];
       alpha  = (double) (*c_vector)[1];
     } else {
       return QPLUG_FAILURE;
     }
 } else {
   constrained = ((int) *mvatype_do);
 }
 alpha = 1.00;

 // --- certain constrains not possible without specific inputs :
 if ( (!inSup[0] && (constrained == 1 || constrained >=3))  ||  (!inSup[2] && constrained == 2) ) {
   constrained = 0;
   QplugAppendTextDisplay("WARNING : Insufficient inputs for selected constrain\n");
   
 }
 printf("Constrain = %d  %f\n",constrained, alpha);
 fflush(stdout);

 // ---- Get weights - 4 numbers with Q-ratios ------

 double sum_weights=0.0;
 QdRMatrix_var Weights  = new QdRMatrix(RMatrix(Dim(8), 1.0)); // initialise weights to all unity
 QdString_var W_string = QdString_var::narrow((*call_list)[10]);
 if (!W_string.is_nil()) {
    QString s = W_string->c_str();
    Weights = str2Vec( W_string->c_str());
 }
 for (i=0; i<8; i++) sum_weights += (double)(*Weights)[i];

 
 // ---- check for sensible amount of data -----

 if (nestsize > 1000) {
  QplugAppendTextDisplay("\n == More than 1000 datapoints. Try reducing time interval ==\n");
  return(QPLUG_FAILURE);
 }


 // ---- Get Q_matrices for the various conservation laws ------------------

 if  (min_nest <= 0) {
   printf("Single nest analysis..\n");
   fflush(stdout);
   if (!SingleAnalysis(V,B,dB,P,E,N,H,tt_out,Weights,constrained,alpha,SC_id,doType,nestsize,call_list))
   {
	printf("Single nest analysis OK\n");
	fflush(stdout);
   }
 }
 else {
   printf("Nested analysis - min nest size = %d\n",min_nest); fflush(stdout);
   if (!NestedAnalysis(V,B,dB,P,E,N,H,tt_out,Weights,constrained,alpha,min_nest,SC_id,doType,nestsize))
   {
	printf("Nested analysis OK\n");
	fflush(stdout);
   }
 }

 // QString s2 = "\nTime elapsed : " + QString::number(t.elapsed()/1000.0) +"\n\n";
 // printf("T elapsed = %f\n",t.elapsed()/1000.0);
 // QplugAppendTextDisplay(s2);

 return (QPLUG_SUCCESS);
}







// -------------------------------------------------------------------
/*!
  \brief Parses the a text and converts text to a Vector
  \return  QSAS matrix with the parsed values of the string
  \param s - a string on the form '[2.00 -23.0  9.0]'
*/
//
// SEH, 08 Aug 2008, Non-QT version
//
// -------------------------------------------------------------------
QdRMatrix_var str2Vec(const char *s)
{
   QdRMatrix_var v = new QdRMatrix(RMatrix(Dim(8), 1.0));  // NB initialize to 1 to get Walenslope=1.0 as default
   string str (s);
   char *cstr, *p;
   cstr = new char [str.size()+1];
   strcpy (cstr, str.c_str());
   int i=0;
   // cstr now contains a c-string copy of str

   p=strtok (cstr," ");
   while (p!=NULL)
   {
     (*v)[i++] = atof(p);
     cout << p << endl;
     p=strtok(NULL," ");
   }

  delete[] cstr;  
  return(v);
}
