/***********************************************************************
*           See Accompanying licence file QSAS_licence                 *
***********************************************************************/
/*************************************************************************
* PROJECT: Cluster UK CDHF
*
* COMPONENT:    Plug In library
* 
* MODULE:	ClusterConfig.cc
*
* LANGUAGE:	C++
*
* FUNCTION:     QSAS Cluster Config (Qtetra) Plugin wrapper
*
* PURPOSE:      Show/Manipulate Cluster 4 s/c positions
* 
* DESCRIPTION:  .
*
* ARGUMENTS:
*    IN:      QdCDOList  list of plug in objects containing
*
*             sc1, sc2, sc3, sc4      Constant vector or Vector TS of 4s/c
*                                     Typically sc_drn deltas from AUX
*             time_int    Centre time interval for crossing
*
*   OUT:
*             Nothing; separate gui launched
*
* GLOBALS:    none
*
* RETURN:
*    TYPE:    QplugReturnStatus
*    VALUES:
*
*
* FUNCTIONS USED: built on Qtetra, and bits copied/modified from QBNCoords
*
* HISTORY: March 2002 Steve Schwartz 
*
***********************************************************************/

// #include "Qtetra_gui.h"
#include "qplug_if.h"
#include "qrtn.h"
#include "QCCWindow.h"
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>

#include "qdutil.h"
#include "Qdos.h"
#include "qar.h"

#define MAXSTRLEN 500 

// extern Qtetra_options *qtet_widget;
// extern Qtetra_dialog *qtetWindow;

extern QCCWindow * qccWindow;

using namespace QSAS;  

// local prototypes
extern "C" QplugReturnStatus QClusterConfig(QplugArgList *);
QplugReturnStatus QCFPosition(QrVector *vecp,
			      QdObject_var pos_in,
	 QdTimeInterval_var time_int);
QplugReturnStatus QCFConstPosition(QrVector *vecp, 
				   QdRMatrix_var pos_in, 
       QdTimeInterval_var time_int);
QplugReturnStatus QCFCentrePosition(QrVector *vecp, 
				    QdRMatrixSeq_var pos_in_seq, 
	QdTimeInterval_var time_int);
double SIConv_factor(QdObject_var in1_ptr);
void QrVector_diag_print(QrVector *vecp);

extern int QrGetFrameFromAttr(QdObject_var);
  
/**********************************************************************/
QplugReturnStatus QClusterConfig(  QplugArgList *call_list){
	/**********************************************************************/

// define some C variables

// call-list arguments
 
	QdObject_var         cin[4];       // 0,1,2,3
	QdTimeInterval_var   time_int;     // 4

// define Qdos data objects

	QrVector  c[4];

	QplugReturnStatus ret_status;

// ...first 4 items are input position vectors/vector TS 
//... last one is time interval

	int isc;
	for (isc = 0; isc < 4; isc++){
		cin[isc] = QdObject_var::narrow((*call_list)[isc]);
		if (cin[isc].is_nil()){
			QplugAppendTextDisplay("QClusterConfig>> Error getting position object\n");
			return QPLUG_FAILURE;
		}
		char * attr = QuGetAttrText(SC_ID,cin[isc]);
		int scid = atoi(attr) - 1;
		if ( scid < 0 || scid != isc){
			QplugAppendTextDisplay ("WARNING, unknown or WRONG Spacecraft number\n");
			QuMsgBar("WARNING: Check Plugin messages", true);
		}
	}

	time_int = get_timeinterval((*call_list)[4]);
	if( time_int.is_nil() ) {
		QplugAppendTextDisplay ("QClusterConfig>> Time Interval object invalid.\n");
		return QPLUG_FAILURE;
	}

	
	for (isc = 0; isc < 4; isc++)
	{

  // first check if this is a position vector
		bool warn=false;
		if (!QuIsThreeVector(cin[isc])) warn = true;
		char *attr=QuGetAttrText(SI_CONVERSION,cin[isc]);
		if( !QarSameBaseSI((const char *)attr, "1>m" ) ) warn = true;
		if(warn){
			QplugAppendTextDisplay(
					       "CAUTION, spacecraft positions do not appear to be vector distances\n");
		}
  
		ret_status = QCFPosition(&c[isc],cin[isc],time_int);
		if (ret_status != QPLUG_SUCCESS){
			return QPLUG_FAILURE;
		}
	}
//  c[i] now has a sequence with 1 element, the xyz position
// of the ith spacecraft with one timetag. 
	
// 	QString s;
// 	for (isc = 0; isc < 4; isc++){
// 		qtet_widget->sc_details[isc]->posx->setText(s.setNum(c[isc].ptr[0][0]));
// 		qtet_widget->sc_details[isc]->posy->setText(s.setNum(c[isc].ptr[0][1]));
// 		qtet_widget->sc_details[isc]->posz->setText(s.setNum(c[isc].ptr[0][2]));
// 	}
	QString qtitlepre = "Cluster Config ";
	QString qtitle  = qtitlepre.append(c[0].t[0].iso_srep().c_str());
	
	const char *frameptr = QuGetFrame( c[--isc].ptr);

	char *attr;

// we now have frame sub-str 
	if (c[isc].ptr->xref_exists(UNITS))attr = QuGetAttrText(UNITS,c[isc].ptr);
	else if (c[isc].ptr->xref_exists(LABLAXIS))attr = QuGetAttrText(UNITS,c[isc].ptr);
	else attr = "()";

	char units_str[MAXSTRLEN];
	sprintf(units_str,"%s (%s)",attr,frameptr);
	
	QString units(units_str);
	qccWindow->setParameters(c[0], c[1], c[2], c[3], qtitle, units);
	qccWindow->show();
	
	
	return QPLUG_SUCCESS;

}

/**********************************************************************/
QplugReturnStatus QCFPosition(QrVector *vecp,
			      QdObject_var pos_in,
	 QdTimeInterval_var time_int){
		 /**********************************************************************/

		 QdRMatrixSeq_var pos_in_seq;
		 QdRMatrix_var    pos_in_const;
		 QplugReturnStatus ret_status;

// sort out position. Input can either be a time series or
// a constant vector, so this a bit complicated. Separate routines
// return in vecp a sequence of 1 cartesian vector whose timetag
// is the centre of the time interval time_int, and whose frame is
// that of pos_in. 

		 pos_in_seq       = QdRMatrixSeq_var::narrow(pos_in);
		 pos_in_const     = QdRMatrix_var::narrow(pos_in);
		 if( (!pos_in_seq.is_nil()) && (!pos_in_const.is_nil()) ){
			 QplugAppendTextDisplay("QClusterConfig>> pos_in both const and sequence!?\n");
			 return QPLUG_FAILURE;
		 }
		 if( pos_in_seq.is_nil() && pos_in_const.is_nil()){
			 QplugAppendTextDisplay("QClusterConfig>> pos_in neither const nor sequence!?\n");
			 return QPLUG_FAILURE;
		 }
		 if( !pos_in_seq.is_nil() ) {
			 QplugAppendTextDisplay ("QclusterConfig>> Vector Sequence. Interpolating...\n");
			 ret_status = QCFCentrePosition(vecp,pos_in_seq,time_int);
			 if (ret_status != QPLUG_SUCCESS){
				 QplugAppendTextDisplay("QClusterConfig>> Failed to calculate crossing position\n");
				 return QPLUG_FAILURE;
			 }
		 } else if (!pos_in.is_nil()){
			 ret_status = QCFConstPosition(vecp, pos_in_const, time_int);
			 if (ret_status != QPLUG_SUCCESS){
				 QplugAppendTextDisplay("QClusterConfig>> Failed to calc const. position\n");
				 return QPLUG_FAILURE;
			 }
		 } else {
			 QplugAppendTextDisplay("QClusterConfig>> Confused about pos_in object\n");
			 return QPLUG_FAILURE;
		 }

		 return QPLUG_SUCCESS;
	 }

	 /**********************************************************************/
	 QplugReturnStatus  QCFConstPosition(QrVector *vecp, 
					     QdRMatrix_var pos_in, 
	  QdTimeInterval_var time_int){
		  /**********************************************************************/

		  QplugReturnStatus ret_status;
		  const char * xyz_comp[] = {"x", "y", "z"};
		  int i;

// grab constant vector and fill vecp.

		  if (!QuIsThreeVector(pos_in)){
			  QplugAppendTextDisplay("QClusterConfig>> Const pos_in not a 3Vector\n");
			  return QPLUG_FAILURE;
		  }

		  QdTimeSeq_var tt = new QdTimeSeq(1, 0.5*(time_int->start()+time_int->end()));
		  vecp->ptr = new QdRMatrixSeq(1,RMatrix(Dim(3)));
		  (vecp->ptr)->copy_xrefs_from(pos_in);
		  QdRMatrixSeq_var pos_seq = new QdRMatrixSeq(1, (*pos_in) );
		  pos_seq->copy_xrefs_from(pos_in);
		  for (i=0; i<3; i++){
			  QdRScalarSeq_var comp = 
					  QdRScalarSeq_var::narrow(QuGetVectorComponent(pos_seq,xyz_comp[i]));
			  if (comp.is_nil()){
				  QplugAppendTextDisplay("BNCoords>> Const pos comp narrow failed\n");
				  return QPLUG_FAILURE;
			  }
			  (vecp->ptr)[0][i] = comp[0];
		  }
		  vecp->n_recs = 1;
		  vecp->t = tt;

		  vecp->in_frame = QrGetFrameFromAttr(pos_in); // to complete vecp

		  return QPLUG_SUCCESS;
	  }


	  /**********************************************************************/
	  QplugReturnStatus  QCFCentrePosition(QrVector *vecp, 
					       QdRMatrixSeq_var pos_in_seq, 
	    QdTimeInterval_var time_int){
		    /**********************************************************************/

		    int i,nttin,ntrue,il,ih;
		    QplugReturnStatus ret_status;
		    QdTime tc;
		    double cl,ch, frac;
		    const char * xyz_comp[] = {"x", "y", "z"};

		    if (!QuIsThreeVector(pos_in_seq)){
			    QplugAppendTextDisplay("QClusterConfig>> Sequence pos_in not a 3Vector\n");
			    return QPLUG_FAILURE;
		    }

// prepare vecp 

		    tc = 0.5*(time_int->start()+time_int->end());
		    QdTimeSeq_var tt = new QdTimeSeq(1,tc );
		    vecp->ptr = new QdRMatrixSeq(1,RMatrix(Dim(3)));
		    (vecp->ptr)->copy_xrefs_from(pos_in_seq);
		    vecp->t = tt;
		    vecp->n_recs = 1;
		    vecp->in_frame = QrGetFrameFromAttr(pos_in_seq);

// linearly interpolate crossing position from bracketting time tags
// If any problems, default to 1st vector in pos_in_seq
// There are probably much simpler ways to do this. Note, though, that
// Qrtn routines only give correct answers if given x,y,z representation

		    QdTimeSeq_var tt_in = get_timetags(pos_in_seq);
		    if (tt_in.is_nil()){
			    QplugAppendTextDisplay(
					    "QClusterConfig>> No timetags for input pos seq. Taking 1st pt\n");
			    il = ih = 0;
		    } else {
			    nttin = tt_in->size();
			    if (nttin <= 0){
				    QplugAppendTextDisplay("QClusterConfig>> Input seq 0 length\n");
				    return QPLUG_FAILURE;
			    } else if (nttin == 1){
				    il = ih = 0;
			    } else {
				    SequenceMask msk(tt_in->get_mask( QdTimeSeq::LE( tc)));
				    ntrue = msk.num_true();
				    if (ntrue <= 0) il = ih = 0;
				    else {
					    il = ntrue -1;
					    ih =  (ntrue < nttin) ?   il + 1 : il; 
				    }  
			    }
		    }

		    if(ih == il) frac = 0.0;   // safe
		    else frac = (tc-tt_in[il])/(tt_in[ih] - tt_in[il]); // calc lin interpol factor

/* printf("tl: %s\ntc: %s\nth: %s\nfrac = %g\n",
		    tt_in[il].iso_srep().c_str(),
		    tc.iso_srep().c_str(),
		    tt_in[ih].iso_srep().c_str(),
		    frac);
*/

		    for (i=0; i<3; i++){                          // fill crossing position
			    QdRScalarSeq_var comp = 
					    QdRScalarSeq_var::narrow(QuGetVectorComponent(pos_in_seq,xyz_comp[i]));
			    if (comp.is_nil()){
				    QplugAppendTextDisplay("BNCoords>> Centre Pos comp narrow failed\n");
				    return QPLUG_FAILURE;
			    }
			    if (ih==il) {
				    vecp->ptr[0][i] = (double)comp[il];
			    } else {
				    cl = (double)comp[il];
				    ch = (double)comp[ih];
				    vecp->ptr[0][i] = cl + frac*(ch-cl);
			    }
		    }



		    return QPLUG_SUCCESS;
	    }

	    /***********************************************************************/
	    double SIConv_factor(QdObject_var in1_ptr){ 
		    /***********************************************************************/

		    char * siconv1;
		    double value1;
		    char * ptr1;
		    static char * null_unit = "1 > ()";

		    char * txt1 = QuGetAttrText("SI_conversion", in1_ptr);
  
		    if(strcmp(txt1, "") == 0) {
			    siconv1 = new char[strlen(null_unit)+1] ;
			    strcpy(siconv1, null_unit);
		    } else {
			    siconv1 = new char[strlen(txt1)+1] ;
			    strcpy(siconv1, txt1);  
		    }
  
		    if( (ptr1 = strtok(siconv1, ">")) == NULL) {
			    delete[] siconv1;
			    return 0.0;
		    }
		    value1 = (double) atof(ptr1);
 
		    delete[] siconv1;
  
		    return value1;
  
	    } // end SIConv_factor

	    /**************************************************************************/
	    void QrVector_diag_print(QrVector *vecp){
		    printf("IsPositionVector = %d, in_frame = %d, out_frame = %d, n_recs = %d\n",
			   vecp->IsPositionVector,vecp->in_frame,vecp->out_frame, vecp->n_recs);
		    printf("First Vector: (%9.3g, %9.3g, %9.3g) at %s\n",
			   (double)vecp->ptr[0][0], (double)vecp->ptr[0][1], (double)vecp->ptr[0][2],
			    vecp->t[0].iso_srep().c_str());
				fflush(stdout);
		    return;
	    }



