//  qslot.cc
//
// slot Utility Widgets
//
// Aug 2007: Tony Allen
//

/// \file qgu.cc
///  Implementation file for the main GUI classes that are used within QSAS which are: \n
///  \arg \c QguListItem \n
///  \arg \c QguListView \n
///  \arg \c QguWLinput  \n
///  \arg \c QguWindow   \n \n
///  Those classes specialise respectively the QT3 built-in classes as listed below\n
///  \arg \c Q_LIST_VIEW_ITEM \n
///  \arg \c QListView     \n
///  \arg \c QLineEdit     \n
///  \arg \c QMainWindow   \n 
///  \author Tony Allen    \n
///  \version 2.1         \n
///  \date    2007

#include <qcolor.h>
#include <qpalette.h>
#include <qglobal.h>
#include <qpaintdevice.h>
#include <qfont.h>

#if QT_VERSION < 0x040000
#include "qgu.h"
#else
#include "qgu_qt4.h"
#endif

#include "qslot.h"
#include "qwui.h"
#include "TimeUtils.h"

#include <stdio.h>

extern qwuiWindow* qsasWindow;

/// \var  extern vector<QguListItem *> dragList
/// \brief Global variable that contains the list of items being dragged
///
extern vector<QguListItem *> dragList;

/// \var  extern QguListView * dragFromList
/// \brief Global variable that defines the list view from which the items are being dragged
///
extern QguListView * dragFromList;


QColor col_White("#ffffff");
QColor col_Light_Grey("#eeeeee");
QColor col_Grey("#8c8c8c");
QColor col_Black("#000000");
QColor col_Green("#228b22");
QColor col_Red("#ee0000");
QColor col_DarkRed("#990000");
QColor col_Yellow("#eeee00");
QColor col_Amber("#bbaa33");
QColor col_LightYellow("#ffffaa");
QColor col_Blue("#000099");


// QguWLinput class
//----------------------------

/// \fn QguWLinput::QguWLinput(QWidget *parent, const char *name)
/// \brief Constructor of QguWLinput class
/// \param     parent  of type QWidget      
/// \param     name    of type const char *  
/// \exception None
/// \return    None
///
QguWLinput::QguWLinput(QWidget *parent, const char *name)
   :  Q3_H_GROUP_BOX(parent,name)
{
   if  ( QuSlotDebug() ){
		printf(" QguWLinput::QguWLinput %ld of %s \n", this, name);	
// 		QDateTime  qdatetime ; 
// 		qdatetime = QDateTime::currentDateTime();
// 		printf(" ***** SLOT CONSTRUCTOR ******  %s\n", qdatetime.toString( "dd.MM.yyyy-hh:mm:ss.zzz" ).latin1());
   } 

   QHBoxLayout * layout=new QHBoxLayout(this);
#if QT_VERSION >= 0x040000
   this->setLayout(layout);
#endif

  WLI_palette = new QPalette(qApp->palette());
  QFont slot_font = QFont(qApp->font());
  QFontMetrics slot_font_mt(slot_font); 
  int font_width = slot_font_mt.width("n");
  int font_height = slot_font_mt.height();
  this->setFont(slot_font);
  reduce3Dto1 = false;
  reduce3DtoScalar = false;


  canEmitCmpSel = true;
  active = true;       // turn this slot on
  aliasList = NULL;
  acceptNumber = false;
  acceptInteger = false;
  acceptObject = false;
  acceptObjectPlus = false;
  acceptObjectMinus = false;
  acceptObjectPlus_TSonly = false;
  acceptScalarObject = false;
  acceptScalarObject_TSonly = false;
  acceptVectorObject = false;
  acceptVectorObject_TSonly = false;
  acceptArrayObject = false;
  acceptArrayObject_TSonly = false;
  acceptTimeObject = false;
  acceptTTagObject = false;
  acceptTIvlObject = false;  
  acceptDuration = false;  
  acceptAlias = false;
  acceptAliasToScalar = false;
  acceptAliasToVector = false;
  acceptText = false;
  acceptTime = false;
  acceptTimeSlice = false;
  acceptTimeSubset = false;
  acceptTIvlSeqObject = false;
  refresh = false;
  beenEdited = false;
  ObjectName_nil = QString("");
  ObjectName = ObjectName_nil;
  objType = QString("");
  
  lineBox = new QguLineEdit(this, "InputLineBox");
  lineBox->WLinput = this;
  this->setEdited(false); 
  layout->addWidget(lineBox); 
 
  dropDownArray = new QToolButton( Qt::DownArrow, this, "");
  dropDownArray->setFixedSize(25,21);
  userChoice = true;
  layout->addWidget(dropDownArray);
	  
  // initialise the dropDown menu possibilities
    
  defaultPopup = 0;
  defaultPopupSlice = 0;
  TrangePopup = 0;
  TivlPopup = 0;
  TivlPopupPlus = 0;
  TivlFromEvtTablePopup = 0;
  TivlSlicePopup = 0;
  vectorPopup = 0;
  vectorPopupPlus = 0;
  vectorPopupSlice = 0;
  vectorPopupPlusSlice = 0;
  arrayPopup = 0;
  arrayPopupPlus = 0;

  
//   this->setInsideSpacing(0);
//   this->setInsideMargin(0);
  
  layout->setMargin(0);
  layout->setSpacing(0);
  
  dropDownArray->hide();
  connect(lineBox, SIGNAL(textChanged(const QString &)), this, SLOT(sendTextChanged(const QString &)));
  connect(lineBox, SIGNAL(hadDropEvent()), this, SIGNAL(hadDropEvent()));
  connect(this, SIGNAL(textChanged(const QString &)), this, SLOT(validateInput(const QString &)));
  connect(qsasWindow->WorkingList, SIGNAL(listModified()), this, SLOT(reValidate()));
//   this->setFrameStyle(QFrame::NoFrame);
  this->resize(font_width*23,(int)(font_height*1.5)); // large enough for a time string
  this->setMinimumWidth(font_width*23);
  
}


/// \fn QguWLinput::~QguWLinput()
/// \brief Destructor of QguWLinput class
/// \exception None
/// \return    None
///
QguWLinput::~QguWLinput()
{
  delete WLI_palette;
  delete lineBox;
}

// --------------------------
Q_POPUP_MENU * QguWLinput::getDefaultPopup()
{
  if(this->defaultPopup == 0){
#if QT_VERSION < 0x040000
     defaultPopup = new QPopupMenu(this, "defaultPopup");
#else
     defaultPopup = new QMenu(this);
#endif
     defaultPopup->insertItem("This Object", this, SLOT(setDropdownText_dobj()));
     defaultPopup->insertSeparator();
     defaultPopup->insertItem("Metadata", this, SLOT(openSelectMetadata()));
  }
  this->defaultPopup->hide();
  return this->defaultPopup;
}
// --------------------------
Q_POPUP_MENU * QguWLinput::getDefaultPopupSlice()
{
  if(this->defaultPopupSlice == 0){
#if QT_VERSION < 0x040000
     defaultPopupSlice = new QPopupMenu(this, "defaultPopupSlice");
#else
     defaultPopupSlice = new QMenu(this);
#endif
     defaultPopupSlice->insertItem("This Object", this, SLOT(setDropdownText_dobj()));
     defaultPopupSlice->insertSeparator();
     defaultPopupSlice->insertItem(" Time Slice ", this, SLOT(openSliceS()));
     defaultPopupSlice->insertSeparator();
     defaultPopupSlice->insertItem("Metadata", this, SLOT(openSelectMetadata()));
  }
  this->defaultPopupSlice->hide();
  return this->defaultPopupSlice;
}
// --------------------------
Q_POPUP_MENU * QguWLinput::getTrangePopup()
{
  if(this->TrangePopup == 0){
#if QT_VERSION < 0x040000
     TrangePopup = new QPopupMenu(this, "TrangePopup");
#else
     TrangePopup = new QMenu(this);
#endif
     TrangePopup->insertItem("This Object", this, SLOT(setDropdownText_dobj()));
     TrangePopup->insertSeparator();
     TrangePopup->insertItem("Subset...", this, SLOT(openSliceT()));
     TrangePopup->insertSeparator();
     TrangePopup->insertItem("Metadata", this, SLOT(openSelectMetadata()));
  }
  this->TrangePopup->hide();
  return this->TrangePopup;
}
// --------------------------
Q_POPUP_MENU * QguWLinput::getTivlPopup()
{
  if(this->TivlPopup == 0){
#if QT_VERSION < 0x040000
	 TivlPopup = new QPopupMenu(this, "TivlPopup");
#else
         TivlPopup = new QMenu(this);
#endif
     TivlPopup->insertItem("Start Time", this, SLOT(setDropdownTstart()));
     TivlPopup->insertItem("Centre Time", this, SLOT(setDropdownTcentre()));
     TivlPopup->insertItem("End Time", this, SLOT(setDropdownTend()));
     TivlPopup->insertSeparator();
     TivlPopup->insertItem("Time at Record", this, SLOT(openSliceT()));
     TivlPopup->insertSeparator();
     TivlPopup->insertItem("Metadata", this, SLOT(openSelectMetadata()));
  }
  this->TivlPopup->hide();
  return this->TivlPopup;
}
// --------------------------
Q_POPUP_MENU * QguWLinput::getTivlFromEvtTablePopup()
{
    if(this->TivlFromEvtTablePopup == 0){
#if QT_VERSION < 0x040000
        TivlFromEvtTablePopup = new QPopupMenu(this, "TivlFromEvtTablePopup");
#else
        TivlFromEvtTablePopup = new QMenu(this);
#endif
        TivlFromEvtTablePopup->insertItem("Select Event", this, SLOT(selectTivlEvent()));
    }
    this->TivlFromEvtTablePopup->hide();
    return this->TivlFromEvtTablePopup;
}
// --------------------------
Q_POPUP_MENU * QguWLinput::getTivlSeqSlicePopup()
{
    if(this->TivlSlicePopup == 0){
#if QT_VERSION < 0x040000
        TivlSlicePopup = new QPopupMenu(this, "TivlSeqSlicePopup");
#else
        TivlSlicePopup = new QMenu(this);
#endif
        TivlSlicePopup->insertItem("This Object", this, SLOT(setDropdownText_dobj()));
        TivlSlicePopup->insertSeparator();
        TivlSlicePopup->insertItem("Select Subset", this, SLOT(selectTivlSeqSubset()));
        TivlSlicePopup->insertItem("Select Event", this, SLOT(selectTivlEvent()));

    }
    this->TivlSlicePopup->hide();
    return this->TivlSlicePopup;
}
 // --------------------------
Q_POPUP_MENU * QguWLinput::getTivlPopupPlus()
{
  if(this->TivlPopupPlus == 0){
#if QT_VERSION < 0x040000
	 TivlPopupPlus = new QPopupMenu(this, "TivlPopupPlus");
#else
         TivlPopupPlus = new QMenu(this);
#endif
     TivlPopupPlus->insertItem("This Object", this, SLOT(setDropdownText_dobj()));
     TivlPopupPlus->insertItem("Start Time", this, SLOT(setDropdownTstart()));
     TivlPopupPlus->insertItem("Centre Time", this, SLOT(setDropdownTcentre()));
     TivlPopupPlus->insertItem("End Time", this, SLOT(setDropdownTend()));
     TivlPopupPlus->insertSeparator();
     TivlPopupPlus->insertItem("Time at Record", this, SLOT(openSliceT()));
     TivlPopupPlus->insertSeparator();
     TivlPopupPlus->insertItem("Metadata", this, SLOT(openSelectMetadata()));
  }
  this->TivlPopupPlus->hide();
  return this->TivlPopupPlus;
} 
 // --------------------------
Q_POPUP_MENU * QguWLinput::getVectorPopup()
{
  if(this->vectorPopup == 0){
#if QT_VERSION < 0x040000
	 vectorPopup = new QPopupMenu(this, "vectorPopup");
#else
         vectorPopup = new QMenu(this);
#endif
     vectorPopup->insertItem(" r (mag) ", this, SLOT(setDropdownText_r()));
     vectorPopup->insertItem(" x ", this, SLOT(setDropdownText_x()));
     vectorPopup->insertItem(" y ", this, SLOT(setDropdownText_y()));
     vectorPopup->insertItem(" z ", this, SLOT(setDropdownText_z()));
     vectorPopup->insertItem(" Lat (deg) ", this, SLOT(setDropdownText_ld()));
     vectorPopup->insertItem(" theta (deg) ", this, SLOT(setDropdownText_td()));
     vectorPopup->insertItem(" phi (deg) ", this, SLOT(setDropdownText_pd()));
     vectorPopup->insertItem(" Lat (rad) ", this, SLOT(setDropdownText_lr()));
     vectorPopup->insertItem(" theta (rad) ", this, SLOT(setDropdownText_tr()));
     vectorPopup->insertItem(" phi (rad) ", this, SLOT(setDropdownText_pr()));
     vectorPopup->insertSeparator();
     vectorPopup->insertItem("Metadata", this, SLOT(openSelectMetadata()));
  }
  this->vectorPopup->hide();
  return this->vectorPopup;
} 
 // --------------------------
Q_POPUP_MENU * QguWLinput::getVectorPopupPlus()
{
  if(this->vectorPopupPlus == 0){
#if QT_VERSION < 0x040000
     vectorPopupPlus = new QPopupMenu(this, "vectorPopupPlus");
#else
     vectorPopupPlus = new QMenu(this);
#endif
     vectorPopupPlus->insertItem("This Object", this, SLOT(setDropdownText_dobj()));
     vectorPopupPlus->insertItem(" r (mag) ", this, SLOT(setDropdownText_r()));
     vectorPopupPlus->insertItem(" x ", this, SLOT(setDropdownText_x()));
     vectorPopupPlus->insertItem(" y ", this, SLOT(setDropdownText_y()));
     vectorPopupPlus->insertItem(" z ", this, SLOT(setDropdownText_z()));
     vectorPopupPlus->insertItem(" Lat (deg) ", this, SLOT(setDropdownText_ld()));
     vectorPopupPlus->insertItem(" theta (deg) ", this, SLOT(setDropdownText_td()));
     vectorPopupPlus->insertItem(" phi (deg) ", this, SLOT(setDropdownText_pd()));
     vectorPopupPlus->insertItem(" Lat (rad) ", this, SLOT(setDropdownText_lr()));
     vectorPopupPlus->insertItem(" theta (rad) ", this, SLOT(setDropdownText_tr()));
     vectorPopupPlus->insertItem(" phi (rad) ", this, SLOT(setDropdownText_pr()));
     vectorPopupPlus->insertSeparator();
     vectorPopupPlus->insertItem("Metadata", this, SLOT(openSelectMetadata()));
  }
  this->vectorPopupPlus->hide();
  return this->vectorPopupPlus;
} 
 // --------------------------
Q_POPUP_MENU * QguWLinput::getVectorPopupSlice()
{
  if(this->vectorPopupSlice == 0){
#if QT_VERSION < 0x040000
	 vectorPopupSlice = new QPopupMenu(this, "vectorPopupSlice");
#else
         vectorPopupSlice = new QMenu(this);
#endif
     vectorPopupSlice->insertItem(" r (mag) ", this, SLOT(setDropdownText_r()));
     vectorPopupSlice->insertItem(" x ", this, SLOT(setDropdownText_x()));
     vectorPopupSlice->insertItem(" y ", this, SLOT(setDropdownText_y()));
     vectorPopupSlice->insertItem(" z ", this, SLOT(setDropdownText_z()));
     vectorPopupSlice->insertItem(" Lat (deg) ", this, SLOT(setDropdownText_ld()));
     vectorPopupSlice->insertItem(" theta (deg) ", this, SLOT(setDropdownText_td()));
     vectorPopupSlice->insertItem(" phi (deg) ", this, SLOT(setDropdownText_pd()));
     vectorPopupSlice->insertItem(" Lat (rad) ", this, SLOT(setDropdownText_lr()));
     vectorPopupSlice->insertItem(" theta (rad) ", this, SLOT(setDropdownText_tr()));
	 vectorPopupSlice->insertItem(" phi (rad) ", this, SLOT(setDropdownText_pr()));
     vectorPopupSlice->insertSeparator();
     vectorPopupSlice->insertItem(" Time and Component Slice ", this, SLOT(openSliceV()));
     vectorPopupSlice->insertItem(" Time and Array Slice ", this, SLOT(openArrayReduce()));
	 vectorPopupSlice->insertSeparator();
     vectorPopupSlice->insertItem("Metadata", this, SLOT(openSelectMetadata()));
  }
  this->vectorPopupSlice->hide();
  return this->vectorPopupSlice;
} 
 // --------------------------
Q_POPUP_MENU * QguWLinput::getVectorPopupPlusSlice()
{
  if(this->vectorPopupPlusSlice == 0){
#if QT_VERSION < 0x040000
     vectorPopupPlusSlice = new QPopupMenu(this, "vectorPopupPlusSlice");
#else
     vectorPopupPlusSlice = new QMenu(this);
#endif
     vectorPopupPlusSlice->insertItem("This Object", this, SLOT(setDropdownText_dobj()));
     vectorPopupPlusSlice->insertItem(" r (mag) ", this, SLOT(setDropdownText_r()));
     vectorPopupPlusSlice->insertItem(" x ", this, SLOT(setDropdownText_x()));
     vectorPopupPlusSlice->insertItem(" y ", this, SLOT(setDropdownText_y()));
     vectorPopupPlusSlice->insertItem(" z ", this, SLOT(setDropdownText_z()));
     vectorPopupPlusSlice->insertItem(" Lat (deg) ", this, SLOT(setDropdownText_ld()));
     vectorPopupPlusSlice->insertItem(" theta (deg) ", this, SLOT(setDropdownText_td()));
     vectorPopupPlusSlice->insertItem(" phi (deg) ", this, SLOT(setDropdownText_pd()));
     vectorPopupPlusSlice->insertItem(" Lat (rad) ", this, SLOT(setDropdownText_lr()));
     vectorPopupPlusSlice->insertItem(" theta (rad) ", this, SLOT(setDropdownText_tr()));
     vectorPopupPlusSlice->insertItem(" phi (rad) ", this, SLOT(setDropdownText_pr()));
     vectorPopupPlusSlice->insertSeparator();
     vectorPopupPlusSlice->insertItem(" Time and Component Slice ", this, SLOT(openSliceVP()));
     vectorPopupPlusSlice->insertItem(" Time and Array Slice ", this, SLOT(openArrayReduce()));
     vectorPopupPlusSlice->insertSeparator();
     vectorPopupPlusSlice->insertItem("Metadata", this, SLOT(openSelectMetadata()));
  }
  this->vectorPopupPlusSlice->hide();
  return this->vectorPopupPlusSlice;
} 
 // --------------------------
Q_POPUP_MENU * QguWLinput::getArrayPopup()
{
  if(this->arrayPopup == 0){
#if QT_VERSION < 0x040000
     arrayPopup = new QPopupMenu(this, "arrayPopup");
#else
     arrayPopup = new QMenu(this);
#endif
     arrayPopup->insertItem("Reduce Array Dimensions", this, SLOT(openArrayReduce()));
     arrayPopup->insertSeparator();
     arrayPopup->insertItem("Metadata", this, SLOT(openSelectMetadata()));
  }
  this->arrayPopup->hide();
  return this->arrayPopup;
} 
 // --------------------------
Q_POPUP_MENU * QguWLinput::getArrayPopupPlus()
{
  if(this->arrayPopupPlus == 0){
#if QT_VERSION < 0x040000
     arrayPopupPlus = new QPopupMenu(this, "arrayPopupPlus");
#else
     arrayPopupPlus = new QMenu(this);
#endif
     arrayPopupPlus->insertItem("This Object", this, SLOT(setDropdownText_dobj()));
     arrayPopupPlus->insertSeparator();
     arrayPopupPlus->insertItem("Reduce Array Dimensions", this, SLOT(openArrayReduce()));
     arrayPopupPlus->insertSeparator();
     arrayPopupPlus->insertItem("Metadata", this, SLOT(openSelectMetadata()));
  }
  this->arrayPopupPlus->hide();
  return this->arrayPopupPlus;
} 

 // --------------------------
 
bool QguWLinput::hasFocus(){
   return lineBox->hasFocus();
}

 // --------------------------
 
void QguWLinput::clearFocus(){
   return lineBox->clearFocus();
}

// --------------------------
QguComp_e QguWLinput::getComponent(){

   QguComp_e component;
   QString StrCmp = QString( this->text().latin1() );

   int compEndAt = StrCmp.find(" [");
   int sliceEndAt = StrCmp.find(" <");
   if(compEndAt != -1)
   {
   		StrCmp.remove(0,compEndAt); // get " [ x ]" etc
		if (StrCmp.contains( R_STR_CMP))       component = R_COMP;
		else if (StrCmp.contains( X_STR_CMP))  component = X_COMP;
		else if (StrCmp.contains( Y_STR_CMP))  component = Y_COMP;
		else if (StrCmp.contains( Z_STR_CMP))  component = Z_COMP;
		else if (StrCmp.contains( Ld_STR_CMP)) component = LAT_DEG_COMP;
		else if (StrCmp.contains( Lr_STR_CMP)) component = LAT_RAD_COMP;
		else if (StrCmp.contains( Td_STR_CMP)) component = THETA_DEG_COMP;
		else if (StrCmp.contains( Tr_STR_CMP)) component = THETA_RAD_COMP;
		else if (StrCmp.contains( Pd_STR_CMP)) component = PHI_DEG_COMP;
		else if (StrCmp.contains( Pr_STR_CMP)) component = PHI_RAD_COMP;
		else if (StrCmp.contains(T_STR_START)) component = T_START_COMP;
		else if (StrCmp.contains(T_STR_CENTRE)) component = T_CENTRE_COMP;
		else if (StrCmp.contains( T_STR_END))   component = T_END_COMP;
		else if (StrCmp.contains("[=>") > 0) component = XREF_COMP;
		else component = ARRAY_COMP;
   }
   else if(sliceEndAt !=-1)
     component=SLICE_COMP;
   else
   {
     component = WL_OBJ;
   }

	return component;
}


void QguWLinput::openSliceT(){

  if(QuSlotDebug())
    std::cout << "QguWLinput::openSliceT "  << endl;

  QString objName = this->textObject();

  TimeSliceWindow * tSliceWindow = new TimeSliceWindow(objName, "T", this, "tSliceWindow", acceptTimeSubset, acceptTimeSlice, false);
  tSliceWindow->setPalette(qApp->palette());

  tSliceWindow->parentSlot = this;

  tSliceWindow->show();
  tSliceWindow->setToLastChoice();
  
}

void QguWLinput::openSliceS(){

  if(QuSlotDebug())
    std::cout << "QguWLinput::openSliceS "  << endl;

  QString objName = this->textObject();

  TimeSliceWindow * tSliceWindow = new TimeSliceWindow(objName, "S", this, "tSliceWindow", acceptTimeSubset, acceptTimeSlice, false);
  tSliceWindow->setPalette(qApp->palette());

  tSliceWindow->parentSlot = this;

  tSliceWindow->show();
  tSliceWindow->setToLastChoice();
  
}

void QguWLinput::openSliceV(){

  if(QuSlotDebug())
    std::cout << "QguWLinput::openSliceV "  << endl;

  QString objName = this->textObject();

  TimeSliceWindow * tSliceWindow = new TimeSliceWindow(objName, "V", this, "tSliceWindow", acceptTimeSubset, acceptTimeSlice, false);
  tSliceWindow->setPalette(qApp->palette());

  tSliceWindow->parentSlot = this;

  tSliceWindow->show();
  tSliceWindow->setToLastChoice();
  
}

void QguWLinput::openSliceVP(){

  if(QuSlotDebug())
    std::cout << "QguWLinput::openSliceVP "  << endl;

  QString objName = this->textObject();

  TimeSliceWindow * tSliceWindow = new TimeSliceWindow(objName, "V+", this, "tSliceWindow", acceptTimeSubset, acceptTimeSlice, false);
  tSliceWindow->setPalette(qApp->palette());

  tSliceWindow->parentSlot = this;

  tSliceWindow->show();
  tSliceWindow->setToLastChoice();
  
}

void QguWLinput::setDropdownText_r(){

  if(QuSlotDebug())
    std::cout << "QguWLinput::setDropdownText_r " << R_STR_CMP << endl;

  this->lineBox->setText(QString(ObjectName.latin1())+QString(R_STR_CMP));
  this->lineBox->setCursorPosition(ObjectName.length());
  this->lineBox->setFocus();

  this->setColour("component");
  if(this->canEmitCmpSel) emit this->componentSelected(R_STR_CMP, "VECTOR");
}

void QguWLinput::setDropdownText_x(){

  if(QuSlotDebug())
    std::cout << "QguWLinput::setDropdownText_x " << X_STR_CMP << endl;

  this->lineBox->setText(QString(ObjectName.latin1())+QString(X_STR_CMP));
  this->lineBox->setCursorPosition(ObjectName.length());
  this->lineBox->setFocus();
  this->setColour("component");
  if(this->canEmitCmpSel) emit this->componentSelected(X_STR_CMP, "VECTOR");
}

void QguWLinput::setDropdownText_y(){
  
  if(QuSlotDebug())
    std::cout << "QguWLinput::setDropdownText_y " << Y_STR_CMP << endl;

  this->lineBox->setText(QString(ObjectName.latin1())+QString(Y_STR_CMP));
  this->lineBox->setCursorPosition(ObjectName.length());
  this->lineBox->setFocus();
  this->setColour("component");
  if(this->canEmitCmpSel) emit this->componentSelected(Y_STR_CMP, "VECTOR");
}

void QguWLinput::setDropdownText_z(){
  
  if(QuSlotDebug())
    std::cout << "QguWLinput::setDropdownText_z " << Z_STR_CMP << endl;

  this->lineBox->setText(QString(ObjectName.latin1())+QString(Z_STR_CMP));
  this->lineBox->setCursorPosition(ObjectName.length());
  this->lineBox->setFocus();
  this->setColour("component");
  if(this->canEmitCmpSel) emit this->componentSelected(Z_STR_CMP, "VECTOR");
}

void QguWLinput::setDropdownText_ld(){
  
  if(QuSlotDebug())
    std::cout << "QguWLinput::setDropdownText_ld " << Ld_STR_CMP << endl;

  this->lineBox->setText(QString(ObjectName.latin1())+QString(Ld_STR_CMP));
  this->lineBox->setCursorPosition(ObjectName.length());
  this->lineBox->setFocus();
  this->setColour("component");
  if(this->canEmitCmpSel) emit this->componentSelected(Ld_STR_CMP, "VECTOR");
}

void QguWLinput::setDropdownText_td(){
  
  if(QuSlotDebug())
    std::cout << "QguWLinput::setDropdownText_td " << Td_STR_CMP << endl;

  this->lineBox->setText(QString(ObjectName.latin1())+QString(Td_STR_CMP));
  this->lineBox->setCursorPosition(ObjectName.length());
  this->lineBox->setFocus();
  this->setColour("component");
  if(this->canEmitCmpSel) emit this->componentSelected(Td_STR_CMP, "VECTOR");
}

void QguWLinput::setDropdownText_pd(){
  
  if(QuSlotDebug())
    std::cout << "QguWLinput::setDropdownText_pd " << Pd_STR_CMP << endl;

  this->lineBox->setText(QString(ObjectName.latin1())+QString(Pd_STR_CMP));
  this->lineBox->setCursorPosition(ObjectName.length());
  this->lineBox->setFocus();
  this->setColour("component");
  if(this->canEmitCmpSel) emit this->componentSelected(Pd_STR_CMP, "VECTOR");
}

void QguWLinput::setDropdownText_lr(){
  
  if(QuSlotDebug())
    std::cout << "QguWLinput::setDropdownText_lr " << Lr_STR_CMP << endl;

  this->lineBox->setText(QString(ObjectName.latin1())+QString(Lr_STR_CMP));
  this->lineBox->setCursorPosition(ObjectName.length());
  this->lineBox->setFocus();
  this->setColour("component");
  if(this->canEmitCmpSel) emit this->componentSelected(Lr_STR_CMP, "VECTOR");
}

void QguWLinput::setDropdownText_tr(){
  
  if(QuSlotDebug())
    std::cout << "QguWLinput::setDropdownText_tr " << Tr_STR_CMP << endl;

  this->lineBox->setText(QString(ObjectName.latin1())+QString(Tr_STR_CMP));
  this->lineBox->setCursorPosition(ObjectName.length());
  this->lineBox->setFocus();
  this->setColour("component");
  if(this->canEmitCmpSel) emit this->componentSelected(Tr_STR_CMP, "VECTOR");
}

void QguWLinput::setDropdownText_pr(){
  
  if(QuSlotDebug())
    std::cout << "QguWLinput::setDropdownText_pr " << Pr_STR_CMP << endl;

  this->lineBox->setText(QString(ObjectName.latin1())+QString(Pr_STR_CMP));
  this->lineBox->setCursorPosition(ObjectName.length());
  this->lineBox->setFocus();
  this->setColour("component");
  if(this->canEmitCmpSel) emit this->componentSelected(Pr_STR_CMP, "VECTOR");
}

void QguWLinput::setDropdownText_dobj(){
  
  if(QuSlotDebug())
    std::cout << "QguWLinput::setDropdownText_dobj " << "this obj" << endl;

  this->lineBox->setText(QString(ObjectName.latin1()));
  this->reValidate();
  this->lineBox->setCursorPosition(ObjectName.length());
  this->lineBox->setFocus();
  if(this->canEmitCmpSel) emit this->componentSelected("", "VECTOR");
}

void QguWLinput::setDropdownTstart(){
  
  if(QuSlotDebug())
    std::cout << "QguWLinput::setDropdownTstart " << T_STR_START << endl;

  this->lineBox->setText(QString(ObjectName.latin1())+QString(T_STR_START));
  this->lineBox->setCursorPosition(ObjectName.length());
  this->lineBox->setFocus();
  this->setColour("component");
  if (userChoice)  emit this->TimeChosen();
  else userChoice = false; // reset
  if(this->canEmitCmpSel) emit this->componentSelected(T_STR_START, "TIME");
}

void QguWLinput::setDropdownTcentre(){
  
  if(QuSlotDebug())
    std::cout << "QguWLinput::setDropdownTcentre " << T_STR_CENTRE << endl;

  this->lineBox->setText(QString(ObjectName.latin1())+QString(T_STR_CENTRE));
  this->lineBox->setCursorPosition(ObjectName.length());
  this->lineBox->setFocus();
  this->setColour("component");
  emit this->TimeChosen();
  if(this->canEmitCmpSel) emit this->componentSelected(T_STR_CENTRE, "TIME");
}
void QguWLinput::setDropdownTend(){
  
  if(QuSlotDebug())
    std::cout << "QguWLinput::setDropdownTend " << T_STR_END << endl;

  this->lineBox->setText(QString(ObjectName.latin1())+QString(T_STR_END));
  this->lineBox->setCursorPosition(ObjectName.length());
  this->lineBox->setFocus();
  this->setColour("component");
  emit this->TimeChosen();
  if(this->canEmitCmpSel) emit this->componentSelected(T_STR_END, "TIME");
}

void QguWLinput::openSelectMetadata(){

  QString objName = this->textObject();

  if(this->acceptAlias){

	// defensive test that list exists
	QguListView *qview = this->aliasList;
	if(qview == NULL) return;
	QdObject_var dobj;
	QguAliasType aliasType =qview->getAliasObject(objName, dobj);

	if( !dobj.is_nil() )
	{
          selectMetadataWindow * metaWindow =
	    new selectMetadataWindow(dobj, this, "metaWindow");
       	  metaWindow->parentSlot = this;
	  metaWindow->show();
    	}
	else if(aliasType==ALIAS_NOT_FOUND)
	{
	  QuMsgBar("Could not resolve object Alias", true);
	}
	else if(aliasType==ALIAS_VIRTUAL)
	{
    	  // cannot return metadata for virtual object
    	  QMessageBox mb("Get Metadata", "No metadata available for Virtual Object",
           QMessageBox::Warning, QMessageBox::Ok, QMessageBox::NoButton,
	   QMessageBox::NoButton );
   	  mb.exec();
	}
  }
  else{
    QdObject_var dobj = QuGetObjByName(objName);
    if( !dobj.is_nil() )
    {
       selectMetadataWindow * metaWindow =
         new selectMetadataWindow(dobj, this, "metaWindow");
       metaWindow->parentSlot = this;
       metaWindow->show();
    }
  }

}


void QguWLinput::openArrayReduce(){

  if(QuSlotDebug())
    std::cout << "QguWLinput::openArrayReduce () " << this << endl;

  QString objName = this->textObject();
  Dim_var mdim;

  if(this->acceptAlias)
  {
     // defensive test that list exists
     QguListView *qview = this->aliasList;
     if(qview == NULL) return;
     qview->getAliasArrDim(objName,mdim);
  }
  else{
    QdObject_var dobj = QuGetObjByName(objName);
    if(dobj.is_nil() ){
      cout << "unresolved object: notify csc-support" << endl;
	 setColour("not_ok");
	 return;
    }
    QdRMatrixSeq_var ms_dobj = QdRMatrixSeq_var::narrow(dobj);
    if (ms_dobj.is_nil() ){
        QdRMatrix_var m_dobj = QdRMatrix_var::narrow(dobj);
	   if (m_dobj.is_nil() ) {
	     cout << "Cannot resolve matrix object - notify csc_upport" << endl;
	     setColour("not_ok");
	     return;
	   }
        mdim = new Dim(m_dobj->dimdata()); // only handles sequences of arrays of same dimensions
    }
    else{
      mdim = new Dim((ms_dobj[0]).dimdata()); // only handles sequences of arrays of same dimensions
    }
  }
  if(mdim.is_nil()) {
	 QuMsgBar("Cannot get dimension of object", true);
     return;
  }

  ArrayReduceWindow * arrWindow = new ArrayReduceWindow(objName, *mdim, this,
    "ArrayWindow", acceptTimeSubset, acceptTimeSlice, false);
  arrWindow->setPalette(qApp->palette());

  arrWindow->parentSlot = this;
  arrWindow->show();
  arrWindow->setToLastChoice();
  if(aliasList) aliasList->selectedObjDim = mdim; // keep here for multiple selection propagation

}

void QguWLinput::setEnabled(bool enabled){
   lineBox->setEnabled(enabled);
//   lineBox->clear();
   if( enabled) setColour("ok");
   else setColour("inactive");
 }
 
 
/// \fn QguWLinput::setReadOnly()
/// \brief passes flag to setReadOnly of contained QLineEdit
/// \exception None
/// \return    None
///  
void QguWLinput::setReadOnly(bool readOnly)
{
  lineBox->setReadOnly(readOnly);
}

/// \fn QguWLinput::clear()
/// \brief passes clear to contained QLineEdit
/// \exception None
/// \return    None
///
void QguWLinput::clear()
{
  lineBox->clear();
}

// \fn QguWLinput::sendTextChanged(const QString &)
/// \brief emits textChanged signal to trigger validation. Traps changes made by validator itself
/// \exception None
/// \return    None
///  
void QguWLinput::sendTextChanged(const QString & txt)
{
   setEdited(false);
   if ( !(this->lineBox->hasFocus()) ) return;  // don't issue signal if text not changed by user

   if (refresh == true){  // prevent infinite validation loop
     refresh = false;
     return;
   }

   setEdited(true);
   emit textChanged(txt);
}


void QguWLinput::emitCmpSelected(const char * cmpTxt, const char *cmpType){
  
  if(QuSlotDebug())
    std::cout << "QguWLinput::emitCmpSelected: cmpTxt=" << cmpTxt << " cmpType=" << cmpType << endl;
    
	// skip if subsample only ("< >" not "[ ]") in cmpTxt
	//if ( strstr(cmpTxt, "[") == NULL ) return;

	emit componentSelected(cmpTxt, cmpType);
}


/// \fn void QguWLinput::reValidate()
/// \brief Method that validates the content of the line edit object
/// \param     None
/// \exception None
/// \return    None
///  
void QguWLinput::reValidate()
{
 
if ( !(this->active) ) return;

 if(QuSlotDebug())
 {
	printf("QguWLinput::reValidate %ld %s %s\n", this, this->name(), this->text().latin1());
	fflush(stdout);
 }
 
   // allows validation of input from signal that doesn't know slot text
   validateInput(text());

}

void QguWLinput::resetColour(){

	 this->setPalette(*WLI_palette);
	 this->lineBox->setPalette(*WLI_palette);
     this->dropDownArray->setPalette(*WLI_palette); 
}

void QguWLinput::selectTivlEvent(){

    if(QuSlotDebug())
        std::cout << "QguWLinput::selectTivlEvent "  << endl;

    QString objName = this->textObject();

    SelectTivlWindow * selectWindow = new SelectTivlWindow(objName, "T", this, "tSliceWindow", false);
    selectWindow->setPalette(qApp->palette());

    selectWindow->parentSlot = this;

    selectWindow->show();
}

void QguWLinput::selectTivlSeqSubset(){
    if(QuSlotDebug())
        std::cout << "QguWLinput::selectTivlSeqSubset "  << endl;

    QString objName = this->textObject();

    SelectTivlSubsetWindow * selectWindow = new SelectTivlSubsetWindow(objName, "T", this, "tSliceWindow", false);
    selectWindow->setPalette(qApp->palette());

    selectWindow->parentSlot = this;

    selectWindow->show();
}

/// \fn void QguWLinput::setColour(const QString validity)
/// \brief Method that sets the colour of the line edit object on the basis of the validity state of the content of this object
/// \param     validity of type const QString
/// \exception None
/// \return    None
///
void QguWLinput::setColour(const QString& validity){

	if(QuSlotDebug()) std::cout << "QguWLinput::setColour (validity) " << this << "( "<< validity.latin1() << " )" <<endl;

#if QT_VERSION < 0x040000
  
	QColorGroup g = QColorGroup(WLI_palette->active());
	QColorGroup in_g = QColorGroup(WLI_palette->inactive());
	g.setColor(QColorGroup::Button, g.light());
	in_g.setColor(QColorGroup::Button, g.light());
	in_g.setColor(QColorGroup::Text, g.mid());
	in_g.setColor(QColorGroup::Base, g.midlight());


// palettes need to be changed BEFORE signals are emitted, to guarantee that
// any fields subsequently changed by slots connected to these signals,
// are not undone on return to this function

  if( validity == "ok"){
	
    g.setColor(QColorGroup::Base, col_White);
    g.setColor(QColorGroup::Text, col_Black);
    WLI_palette->setActive(g);
    WLI_palette->setInactive(in_g);
	this->setPalette(*WLI_palette);
    this->lineBox->setPalette(*WLI_palette);
    this->dropDownArray->setPalette(*WLI_palette);
	emit validated_OK();
  }
  else if( validity == "inactive"){
    g.setColor(QColorGroup::Base, col_White);
    g.setColor(QColorGroup::Text, col_Grey);
    WLI_palette->setActive(g);
    WLI_palette->setInactive(g);
	this->setPalette(*WLI_palette);
    this->lineBox->setPalette(*WLI_palette);
    this->dropDownArray->setPalette(*WLI_palette);
    emit validated_INACTIVE();
  }
  else if( validity == "not_ok"){
    g.setColor(QColorGroup::Base, col_Red);
    g.setColor(QColorGroup::Text, col_Black);
	in_g.setColor(QColorGroup::Base, col_DarkRed);
    WLI_palette->setActive(g);
    WLI_palette->setInactive(in_g);
	this->setPalette(*WLI_palette);
    this->lineBox->setPalette(*WLI_palette);
    this->dropDownArray->setPalette(*WLI_palette);
	emit validated_NOT_OK();
  }
  else if( validity == "incomplete"){
    g.setColor(QColorGroup::Base,col_Amber);
    g.setColor(QColorGroup::Text,col_Black);
    in_g.setColor(QColorGroup::Base,col_Yellow);
    WLI_palette->setActive(g);
    WLI_palette->setInactive(in_g);
	this->setPalette(*WLI_palette);
    this->lineBox->setPalette(*WLI_palette);
    this->dropDownArray->setPalette(*WLI_palette);
	emit validated_WARN();
  }
  else if( validity == "component"){
    g.setColor(QColorGroup::Base,  col_Yellow);
    g.setColor(QColorGroup::Text, col_Blue);
    in_g.setColor(QColorGroup::Base,  col_LightYellow);
	WLI_palette->setActive(g);
    WLI_palette->setInactive(in_g);
	this->setPalette(*WLI_palette);
    this->lineBox->setPalette(*WLI_palette);
    this->dropDownArray->setPalette(*WLI_palette);
    emit validated_OK();
  }

#else
// QT4

	WLI_palette->setColor(QPalette::Active, QPalette::Button, WLI_palette->light());
	WLI_palette->setColor(QPalette::Inactive, QPalette::Button, WLI_palette->light());
	WLI_palette->setColor(QPalette::Inactive, QPalette::Text, WLI_palette->mid());
	WLI_palette->setColor(QPalette::Inactive, QPalette::Base, WLI_palette->midlight());

// palettes need to be changed BEFORE signals are emitted, to guarantee that
// any fields subsequently changed by slots connected to these signals,
// are not undone on return to this function

  if( validity == "ok"){
	
	WLI_palette->setColor(QPalette::Active, QPalette::Base, col_White);
	WLI_palette->setColor(QPalette::Active, QPalette::Text, col_Black);
	WLI_palette->setColor(QPalette::Inactive, QPalette::Base, col_White);
	WLI_palette->setColor(QPalette::Inactive, QPalette::Text, col_Black);

	this->setPalette(*WLI_palette);
    this->lineBox->setPalette(*WLI_palette);
    this->dropDownArray->setPalette(*WLI_palette);
	emit validated_OK();
  }
  else if( validity == "inactive"){

	WLI_palette->setColor(QPalette::Active, QPalette::Base, col_White);
	WLI_palette->setColor(QPalette::Active, QPalette::Text, col_Grey);
	WLI_palette->setColor(QPalette::Inactive, QPalette::Base, col_White);
	WLI_palette->setColor(QPalette::Inactive, QPalette::Text, col_Grey);

	this->setPalette(*WLI_palette);
    this->lineBox->setPalette(*WLI_palette);
    this->dropDownArray->setPalette(*WLI_palette);
    emit validated_INACTIVE();
  }
  else if( validity == "not_ok"){

	WLI_palette->setColor(QPalette::Active, QPalette::Base, col_Red);
	WLI_palette->setColor(QPalette::Active, QPalette::Text, col_Grey);
	WLI_palette->setColor(QPalette::Inactive, QPalette::Base, col_DarkRed);
	WLI_palette->setColor(QPalette::Inactive, QPalette::Text, col_Grey);

	this->setPalette(*WLI_palette);
    this->lineBox->setPalette(*WLI_palette);
    this->dropDownArray->setPalette(*WLI_palette);
	emit validated_NOT_OK();
  }
  else if( validity == "incomplete"){

	WLI_palette->setColor(QPalette::Active, QPalette::Base, col_Amber);
	WLI_palette->setColor(QPalette::Active, QPalette::Text, col_Black);
	WLI_palette->setColor(QPalette::Inactive, QPalette::Base, col_Yellow);
	WLI_palette->setColor(QPalette::Inactive, QPalette::Text, col_Black);

	this->setPalette(*WLI_palette);
    this->lineBox->setPalette(*WLI_palette);
    this->dropDownArray->setPalette(*WLI_palette);
	emit validated_WARN();
  }
  else if( validity == "component"){

	WLI_palette->setColor(QPalette::Active, QPalette::Base, col_Yellow);
	WLI_palette->setColor(QPalette::Active, QPalette::Text, col_Blue);
	WLI_palette->setColor(QPalette::Inactive, QPalette::Base, col_LightYellow);
	WLI_palette->setColor(QPalette::Inactive, QPalette::Text, col_Blue);

	this->setPalette(*WLI_palette);
    this->lineBox->setPalette(*WLI_palette);
    this->dropDownArray->setPalette(*WLI_palette);
    emit validated_OK();
  }

#endif
 

  return;
}


/// \fn void QguWLinput::validateInput(const QString &txt)
/// \brief Method that validates the input string txt
/// \param     txt of type const QString
/// \exception None
/// \return    None
///
void QguWLinput::validateInput(const QString &txt){
 if(QuSlotDebug())
 {
	printf("QguWLinput::validateInput %ld %s\n", this, txt.latin1());
	fflush(stdout);
 }

  // This slot type will change colour according to whether the text is valid, not valid or incomplete
  // It will also store the ObjectName (input in slot) so that the object is not lost when changing
  // text feedback from pull down

   if(!(this->active)){
     // grey out if inactive
      ObjectName = ObjectName_nil;
      objType = QString("");
      setColour("inactive");
      return;
   }
   
   
	const bool scalarOnly = true;
	const bool anything = false;
	bool OK;
	bool OKint;
	QguListView *qview;
	QdObject_var dobj;

	QguComp_e component = this->getComponent();

   bool ObjIsVec = false;
   ObjectName = QString(baseObjectText(txt).latin1()); // deep copy

   if(!acceptText &&  component == UNKNOWN_COMP )
   {
   	setColour("incomplete");
	return;
   }


   if( txt.isNull() || txt.isEmpty()) {
      ObjectName = ObjectName_nil;
      objType = QString("");
      emit validated_EMPTY();
      setColour("ok");
      if ( dropDownArray->isVisibleTo(this)) dropDownArray->hide();
      return;
   }

   QString noWhite = txt.stripWhiteSpace();
   if(noWhite.isEmpty()) {
      ObjectName = ObjectName_nil;
      objType = QString("");
      emit validated_EMPTY();
      setColour("ok");
      if ( dropDownArray->isVisibleTo(this)) dropDownArray->hide();
      return;
   }

   txt.toDouble(&OK); // OK = accept numeric value for some slots
   txt.toInt(&OKint); // OK = accept numeric value for some slots

   if(acceptObject     || acceptObjectPlus || acceptScalarObject        || acceptVectorObject        || acceptArrayObject  ||
       acceptObjectMinus || acceptTTagObject || acceptTIvlObject || acceptTIvlSeqObject || acceptTimeObject          ||
	 acceptScalarObject_TSonly            || acceptVectorObject_TSonly || acceptArrayObject_TSonly  || acceptObjectPlus_TSonly) {
       // Does this resolve to a folder
       QguListItem *WLItem = qsasWindow->WorkingList->getItemByName(ObjectName);
	   
       if(WLItem != NULL && WLItem->getEntryType() != ListTypeEntry){
          ObjectName = ObjectName_nil;
          objType = QString("");
          setColour("incomplete");
	     if ( dropDownArray->isVisibleTo(this)) dropDownArray->hide();
	     return;
       }
       // is there an object on the WL
       dobj = QuGetObjByName(ObjectName);
       bool useTimeSlicer = false;

       if( !(dobj.is_nil()) ) {
		// we are working with an object on the WL
		emit validated_WL_OBJ(ObjectName);
		if((acceptTimeSlice || acceptTimeSubset) && dobj->sequence_size() > 1 ) useTimeSlicer = true;

		if( QuIsThreeVector(dobj) )  ObjIsVec = true;
		 

          objType = QString("object");

	  // ensure TSonly types have time tags
          if(acceptScalarObject_TSonly || acceptVectorObject_TSonly || acceptArrayObject_TSonly || acceptObjectPlus_TSonly) {
	        if( !dobj->xref_exists(TIME_TAGS)) {
	           setColour("not_ok");
	           if ( dropDownArray->isVisibleTo(this)) dropDownArray->hide();
	           return;
             }
	     }

          if (acceptObject) {
	     // This is the generic slot to get back the object you drop in under all circumstances
	     // To get the dimension reduction slots turn on SPECIFIC type required or use acceptObjectsPlus
			setColour("ok");
	        if(useTimeSlicer) dropDownArray->setPopup(this->getDefaultPopupSlice());
			else dropDownArray->setPopup(this->getDefaultPopup());
			dropDownArray->hide();
	        if ( !(dropDownArray->isVisibleTo(this))) {
                    dropDownArray->show();
                    dropDownArray->setPopupDelay(30);
             }
             return;
          }

          if (acceptObjectPlus || acceptObjectPlus_TSonly) {
             // This is the generic slot to get back the object you drop plus
             // possible objects from dimension reduction
	     	setDropdownActive(dobj, anything, ObjIsVec, useTimeSlicer);
	     	return;
          }

          if (acceptObjectMinus ) {
             // This is the generic slot to get back
             // possible objects from dimension reduction
	     	setDropdownActive(dobj, true, ObjIsVec, useTimeSlicer); // notSelf flag set true ( = old flag toScalar)
	     	return;
          }

          // otherwise specific type needed
		 if (acceptTTagObject) {
	        QdTimeSeq_var nobj = get_timetags(dobj);
             if( !(nobj.is_nil()) ){
	          setColour("ok");
	          if ( dropDownArray->isVisibleTo(this)) dropDownArray->hide();
	          return;
             }
          }

          if (acceptTimeObject) {

		    QdTime_var tobj = QdTime_var::narrow(dobj);
		    if(!tobj.is_nil()){
	          setColour("ok");
	          if ( dropDownArray->isVisibleTo(this)) dropDownArray->hide();
	          return;
		    }

	        // this is more general than a TTagObject as it accepts intervals or objects with time tags
	        QdTimeInterval_var nobj = get_timeinterval(dobj);
			if( !(nobj.is_nil()) ){
		      if(acceptTIvlObject){
				// also accept interval itself as this test is ahead of acceptTIvlObject
				setColour("ok");
				dropDownArray->setPopup(getTivlPopupPlus());
				dropDownArray->hide();
				if ( !(dropDownArray->isVisibleTo(this))) {
                        dropDownArray->show();
                        dropDownArray->setPopupDelay(30);
               	}
			  }
			  else{
				dropDownArray->setPopup(getTivlPopup());
				dropDownArray->hide();
				if ( !(dropDownArray->isVisibleTo(this))) {
                        dropDownArray->show();
                        dropDownArray->setPopupDelay(30);
               	}
				if(component == WL_OBJ) {
				   userChoice = false;
				   setDropdownTstart();
				}
			  }
	          return;
			}
          }

          if (acceptTIvlObject) {
			// this is more general than a TTagObject as it accepts intervals or objects with time tags
			QdTimeInterval_var nobj = get_timeinterval(dobj);
             if( !(nobj.is_nil()) ){
			   QdTimeSeq_var ttobj = get_timetags(dobj);
			   if( !(ttobj.is_nil()) ){

			      dropDownArray->setPopup(getTrangePopup());
         	      if ( !(dropDownArray->isVisibleTo(this))) {
                        dropDownArray->show();
                        dropDownArray->setPopupDelay(30);
               	  }
			   }
			   else{
	             setColour("ok");
	             if ( dropDownArray->isVisibleTo(this)) dropDownArray->hide();
			   }
			   setColour("ok");
			   return;
             }

              QdTimeIntervalSeq_var _nobj=QdTimeIntervalSeq_var::narrow(dobj);
              if(!_nobj.is_nil())
              {
                  dropDownArray->setPopup(getTivlFromEvtTablePopup());
                      if ( !(dropDownArray->isVisibleTo(this))) {
                          dropDownArray->show();
                          dropDownArray->setPopupDelay(30);
                      }
                      
                      switch(component)
                      {
                          case WL_OBJ:
                              setColour("incomplete");
                              break;
                          
                          case SLICE_COMP:
                              setColour("ok");
                              break;
                              
                          default:setColour("not_ok");
                            break;
                      }
                      return;
              }
              else
              {
              }

          }
    
          if (acceptTIvlSeqObject) {
              cout << "QguWLinput::validateInput: acceptTIvlSeqObject To complete" << endl; //TODO
//               QdTimeIntervalSeq_var nobj=QdTimeIntervalSeq_var::narrow(dobj);
//               if(!nobj.is_nil())
//               {
//                   dropDownArray->setPopup(getTivlFromEvtTablePopup());
//                   if ( !(dropDownArray->isVisibleTo(this))) {
//                       dropDownArray->show();
//                       dropDownArray->setPopupDelay(30);
//                   }
//                           setColour("ok");
//                   return;
//               }
//               else
//               {
//               }

          }
          
         

          if (acceptScalarObject || acceptScalarObject_TSonly ) {

	        QdRScalar_var nobj_1 = QdRScalar_var::narrow(dobj);

             if( !(nobj_1.is_nil()) ){
	           setColour("ok");
	           if(useTimeSlicer) dropDownArray->setPopup(getDefaultPopupSlice());
			   else dropDownArray->setPopup(getDefaultPopup());
			   dropDownArray->hide();
	           if ( !(dropDownArray->isVisibleTo(this))) {
                    dropDownArray->show();
                    dropDownArray->setPopupDelay(30);
                }
	           return;
             }

             QdRScalarSeq_var nobj_s = QdRScalarSeq_var::narrow(dobj);
             if( !(nobj_s.is_nil()) ){
	           setColour("ok");
	           if(useTimeSlicer) dropDownArray->setPopup(getDefaultPopupSlice());
	           else dropDownArray->setPopup(getDefaultPopup());
			 
			 dropDownArray->hide();
	           if ( !(dropDownArray->isVisibleTo(this))) {
                    dropDownArray->show();
                    dropDownArray->setPopupDelay(30);
                }
	           return;
             }

	        // try to reduce an array or vector

	        if( setDropdownActive(dobj, scalarOnly, ObjIsVec, useTimeSlicer) ) {
	           // can subsample array type
	           return;
	        }
	        // else was not a scalar or array, fall through
          }

          if (acceptVectorObject || acceptVectorObject_TSonly) {
             if( ObjIsVec ){
	           setColour("ok");
	           if(useTimeSlicer) dropDownArray->setPopup(getDefaultPopupSlice());
			   else dropDownArray->setPopup(getDefaultPopup());
			   dropDownArray->hide();
	           if ( !(dropDownArray->isVisibleTo(this))) {
                    dropDownArray->show();
                    dropDownArray->setPopupDelay(30);
                }
	           return;
            }
          }

	     if (acceptArrayObject || acceptArrayObject_TSonly) {

	       QdRMatrix_var nobj_1 = QdRMatrix_var::narrow(dobj);
		   if( !(nobj_1.is_nil()) ){
              if(setDropdownActive(nobj_1, anything, ObjIsVec, useTimeSlicer)){
	              setColour("ok");
	              return;
		      }
		    }

	       QdRMatrixSeq_var nobj_s = QdRMatrixSeq_var::narrow(dobj);
	       if( !(nobj_s.is_nil()) ){
	         if(setDropdownActive(nobj_s, anything, ObjIsVec, useTimeSlicer)){
	           setColour("ok");
	           return;
		     }
           }
        }

     } // end if on dobj.is_nil()

   } // end object type tests
   else if(acceptAlias ){

	// defensive test that list exists
	QguListView *qview = this->aliasList;
        if(qview == NULL) return;

	QguAliasObjType aliasObjType=qview->getAliasObjectType(ObjectName);
	if(aliasObjType!=ALIAS_IS_UNKNOWN)
	{
	  objType=QString("alias");
          // handle individual alias types
	  // Note alias may be virtual, so rely on getIsVector() and
	  // getIsArray()  constructed from inputs
	  // dobj itself may not exist yet

	  if (acceptAliasToScalar)
	  {
	    if(aliasObjType==ALIAS_IS_VECTOR)
	    {
	      dropDownArray->setPopup(getVectorPopup());
	      if(component == WL_OBJ)
	      {
	        this->canEmitCmpSel = false;
			setDropdownText_r(); // WL object is not valid here
			this->canEmitCmpSel = true;
	      }
              dropDownArray->hide();
	      if ( !(dropDownArray->isVisibleTo(this)))
	      {
                dropDownArray->show();
                dropDownArray->setPopupDelay(30);
              }
	    }
	    else if(aliasObjType==ALIAS_IS_ARRAY)
	    {
               // array sub-sampling
	       // if reduction already made this may now be a scalar really
	       Dim_var vDim = this->getSlotArrDim();
	       if (vDim.is_nil())
	         setColour("incomplete"); // original array in slot
	       else if ( vDim->rank() > 0)
		 setColour("incomplete");  // still has dimensions
	       else
		 setColour("component");  // reduced to scalar

	       dropDownArray->setPopup(getArrayPopup());
               dropDownArray->hide();
	       if ( !(dropDownArray->isVisibleTo(this)))
	       {
                 dropDownArray->show();
                 dropDownArray->setPopupDelay(30);
               }
	     }
	     else
	     {
	       // arithmetic not handling text and other data types >>>
	       // update for plugin interface
	       dropDownArray->setPopup(getDefaultPopup());
               dropDownArray->hide();
	       if ( !(dropDownArray->isVisibleTo(this)))
	       {
                 dropDownArray->show();
                 dropDownArray->setPopupDelay(30);
	       }
	       setText(ObjectName);
	       // reset name in case input virtual obj has gone from vector to scalar
	       setColour("ok");
             }

	     emit validated_alias(ObjectName);
	     return;
	  }
          else if (acceptAliasToVector)
	  {
	    if(aliasObjType==ALIAS_IS_VECTOR)
	    {
	      setColour("ok");
	      emit validated_alias(ObjectName);
	    }
            if ( dropDownArray->isVisibleTo(this)) dropDownArray->hide();
	       return;
	  }
          else
	  {  // generic acceptAlias
	    if(aliasObjType==ALIAS_IS_VECTOR)
	    {
              dropDownArray->setPopup(getVectorPopupPlus());
              dropDownArray->hide();
	      if ( !(dropDownArray->isVisibleTo(this)))
	      {
                dropDownArray->show();
                dropDownArray->setPopupDelay(30);
              }
	      if (component != WL_OBJ) setColour("component");
	      else setColour("ok");
            }
	    else if(aliasObjType==ALIAS_IS_ARRAY)
	    {
	      // general case of any object or array object, so
	      // dobj may be acceptable
	      setColour("ok"); // original array in slot
	      dropDownArray->setPopup(getArrayPopupPlus());
              dropDownArray->hide();
	      if ( !(dropDownArray->isVisibleTo(this)))
	      {
                dropDownArray->show();
             	dropDownArray->setPopupDelay(30);
              }
	      if (component != WL_OBJ)
	        setColour("component");
	    }
	    else
	    {
	      dropDownArray->hide();
	      if(ObjectName != txt) setText(ObjectName);
	      // reset if component no longer allowed
	      setColour("ok");
	    }

	    emit validated_alias(ObjectName);
	    return;

         }
       } // endif on qItem

   }  // end if on alias
   else{
       // not an object or alias to object
       ObjectName = ObjectName_nil;
   }

   if (acceptTime) {
      QDOS::Time _timeValue;
      if(txt.length() < 23 ) {
	      if(valid_iso_time_str(string(txt.latin1()),_timeValue))
	      {
			objType = QString("timeString");
			setColour("ok");
	      }
	      else
			setColour("incomplete");
	      return;
 	 }
      if( valid_iso_time_str(string(txt.latin1()), _timeValue) ){
	 	objType = QString("timeString");
	 	int cursorPosn = this->lineBox->cursorPosition();
	 	setColour("ok");
         	setText(_timeValue.iso().c_str());
	 	setCursorPosition(cursorPosn);
	 	if ( dropDownArray->isVisibleTo(this)) 
		{
			dropDownArray->hide();
         	adjustSize();
		}
	 	return;
      }
   }
   if (acceptNumber && OK) {
	 objType = QString("number");
	 emit validated_number(txt);
	 setColour("ok");
	 if ( dropDownArray->isVisibleTo(this)) 
	 {
		dropDownArray->hide();
		adjustSize();
	 }
	 return;
   }
   if (acceptDuration && OK) {
	 objType = QString("number");
	 emit validated_number(txt);
	 setColour("ok");
	 if ( dropDownArray->isVisibleTo(this)) 
	 {
		dropDownArray->hide();
		adjustSize();
	 }
	 return;
   }
   if (acceptInteger && OKint) {
	 objType = QString("integer");
	 setColour("ok");
	 if ( dropDownArray->isVisibleTo(this))
	 {
		dropDownArray->hide();
		adjustSize();
	 }
	 return;
   }

   if (acceptText) {

       // always accept input
      objType = QString("text");
	 setColour("ok");
	 if ( dropDownArray->isVisibleTo(this)) 
	 {
		dropDownArray->hide();
        adjustSize();
	 }
      emit validated_TEXT(txt);
      return;
    }
    // anything else is bad
    objType = QString("");
    setColour("not_ok");
    if ( dropDownArray->isVisibleTo(this))
	{
		dropDownArray->hide();
		adjustSize();
	}
	
    return;

}

/// \fn void QguWLinput::setDropdownActive(QdObject_var dobj, bool toScalar, bool isVec, bool useTimeSlicer)
/// \brief This method shows/hides the dropDown arrow if dimension reduction is possible
///
/// \param     dobj of type QdObject_var
/// \param     toScalar flag of type bool
/// \param     isVec flag of type bool
/// \param     useTimeSlicer flag of type bool
/// \exception None
/// \return    None
///
bool QguWLinput::setDropdownActive(QdObject_var dobj, const bool notSelf, const bool isVec, bool useTimeSlicer)
{
   if(QuArrayDims(dobj) > 0 ){

	 QguComp_e component = this->getComponent(); 

     if(isVec)
     {  
        // specific case of reduction to scalar so dobj is not itself acceptable
        if(notSelf){
           refresh = true; // prevent slot from validating this text
           if(useTimeSlicer) dropDownArray->setPopup(getVectorPopupSlice());
	      else dropDownArray->setPopup(getVectorPopup());

		   if(component == WL_OBJ) {
		      this->canEmitCmpSel = false;
		      setDropdownText_r();
		      this->canEmitCmpSel = true;
		   } 
		   else setColour("component");

        }
        else{

           // general case of any object or array object, so dobj may be acceptable
           if(component == WL_OBJ) setColour("ok"); 
		   else setColour("component");
           if(useTimeSlicer) dropDownArray->setPopup(getVectorPopupPlusSlice());
		   else dropDownArray->setPopup(getVectorPopupPlus());
        }
     
     } // end if on 3-vector
     else{

       // array sub-sampling
         if(notSelf){
            refresh = true; // prevent slot from validating this text
            if(component == ARRAY_COMP ) setColour("component");
			else setColour("incomplete");
            dropDownArray->setPopup(getArrayPopup());
        }
        else{

           // general case of any object or array object, so dobj may be acceptable
           
            if(component == ARRAY_COMP ) setColour("component");
			else setColour("ok");

            dropDownArray->setPopup(getArrayPopupPlus());
        }
       
     }
     // show selected dropdown
     dropDownArray->hide();
     if ( !(dropDownArray->isVisibleTo(this))) {

        dropDownArray->show();
        dropDownArray->setPopupDelay(90);
     }
     return true;
  }
  else{
           //Specific tools for event tables
      QdTimeIntervalSeq_var temp_obj=QdTimeIntervalSeq_var::narrow(dobj);
      if(!temp_obj.is_nil())
      {
        dropDownArray->setPopup(getTivlSeqSlicePopup());
        if ( !(dropDownArray->isVisibleTo(this))) {
            dropDownArray->show();
            dropDownArray->setPopupDelay(30);
        }
        setColour("ok");
        return false;
      }
     // object is not an array
     if(useTimeSlicer) dropDownArray->setPopup(getDefaultPopupSlice());
	 else dropDownArray->setPopup(getDefaultPopup()); 
     dropDownArray->hide();
	if ( !(dropDownArray->isVisibleTo(this))) {
        dropDownArray->show();
        dropDownArray->setPopupDelay(30);
     }
     setColour("ok");
     return false;
  }
}


/// \fn QdObject_var QguWLinput::getSlotObject()
/// \brief Method that gets the QdObject from the WL on the basis of the name that is held within the line edit text
///
/// Will create an object from input text (not on WL, so volatile) if slot accepts non-object input.
/// Will return a QdObject_var containing a dimension reduced object if pulldown is used to select dimension reduction
/// \param     None
/// \exception None
/// \return    Returns QdObject_var, nil if conversion fails
///
QdObject_var QguWLinput::getSlotObject()
{
  if ( QuSlotDebug() )
  {
	printf( " QguWLinput::getSlotObject of type  '%s' for '%s'\n",  objType.latin1(), text().latin1());
	fflush(stdout);
  }

  bool OK;
  QdObject_var slotObj;

  // always returns a QdObject_var pointer.
  // Returns nil pointer on failure (test on obj.is_nil() ) and must be narrowed to object type required
  // validation sets the objType flag, so these are mutually exclusive

  if(objType.isEmpty() ){
		// fast return on empty slot
		return nilObject;
  }
  
  else if(objType == QString("text")){
    QdString_var obj = new QdString(text().latin1());
    slotObj = obj;
  }
  
  else if(objType == QString("number")){   
    QdRScalar_var obj = new QdRScalar(text().toDouble(&OK));
    if(OK) slotObj = obj;
    else slotObj = nilObject;
  }
  
  else if(objType == QString("integer")){ 
    QdRScalar_var obj = new QdRScalar(text().toInt(&OK));
    if(OK) slotObj = obj;
    else slotObj = nilObject;
  }
  
  else if(objType == QString("timeString")){
    QdTime_var obj = new QdTime(text().latin1());
    if(obj.is_nil()) slotObj = nilObject;
    else slotObj = obj;
  }

  else if(objType == QString("object") ){
	bool isTime = isTimeSelection();
    if(acceptTIvlObject && !isTime) slotObj = QuGetIntervalByNameAndComp(this->text());
	else                 slotObj = QuGetObjByNameAndComp(this->text());
  }
  else if( objType == QString("alias") ){

	// cannot use utility call QuGetObjByNameAndComp() as base object is not on WL
	// Slicing and subsetting on time is also not supported for virtuial objects

    QguComp_e component = this->getComponent();

    QdObject_var baseObj;
    if(aliasList == NULL) return nilObject;

    baseObj=aliasList->getAliasResultObject(ObjectName);

    if(component == WL_OBJ){
        slotObj = baseObj;
    }
    else if(component == ARRAY_COMP){
        dimTripleStack dimStack = this->parseArrayReduce();
        slotObj = QuGetArrayComponent(baseObj, dimStack );
    }
    else if(component == R_COMP){
        slotObj = QuGetVectorComponent(baseObj, "r(mag)" );
    }
    else if(component == X_COMP){
        slotObj = QuGetVectorComponent(baseObj, "x" );
    }
    else if(component == Y_COMP){
        slotObj = QuGetVectorComponent(baseObj, "y" );
    }
    else if(component == Z_COMP){
        slotObj = QuGetVectorComponent(baseObj, "z" );
    }
    else if(component == LAT_DEG_COMP){
        slotObj = QuGetVectorComponent(baseObj, "Lat(deg)" );
    }
    else if(component == THETA_DEG_COMP){
        slotObj = QuGetVectorComponent(baseObj, "theta(deg)" );
    }
    else if(component == PHI_DEG_COMP){
        slotObj = QuGetVectorComponent(baseObj, "phi(deg)" );
    }
    else if(component == LAT_RAD_COMP){
        slotObj = QuGetVectorComponent(baseObj, "Lat(rad)" );
    }
    else if(component == THETA_RAD_COMP){
        slotObj = QuGetVectorComponent(baseObj, "theta(rad)" );
    }
    else if(component == PHI_RAD_COMP){
        slotObj = QuGetVectorComponent(baseObj, "phi(rad)" );
    }
    else if(component == T_START_COMP){
        slotObj = QuGetTimeComponent(baseObj, "Tstart" );
    }
    else if(component == T_CENTRE_COMP){
        slotObj = QuGetTimeComponent(baseObj, "Tcentre" );
    }
    else if(component == T_END_COMP){
        slotObj = QuGetTimeComponent(baseObj, "Tend" );
    }
    else if(component == XREF_COMP){
        QString xref_name = getXrefName(text().latin1());
	   try{
          QdObject_var xrefObj = baseObj->get_xref(xref_name.latin1());
          slotObj = xrefObj;
	   }
	   catch(Exception &e){
                QuMsgBar("Object Xref not found: ", xref_name.latin1(), true);   
           slotObj = nilObject;
	   }
    }
    else{
        slotObj = nilObject;
    }
  }
  
  else{
   // objType not recognised
    slotObj = nilObject;
  }

  return slotObj; 
}

bool QguWLinput::isTimeSelection()
{
	QString txt = this->text();
	if(txt.contains(T_STR_START)) return true;
	if(txt.contains(T_STR_CENTRE)) return true;
	if(txt.contains(T_STR_END)) return true;
	return false;

}
/// \fn double QguWLinput::value()
/// \brief Method that returns the double value in the input slot, if acceptNumber or acceptTime is true 
/// \param     None
/// \exception None
/// \return    double
///
double QguWLinput::value()
{
	if (acceptNumber)
	{			
		return atof(this->lineBox->text());
	}
	else if (acceptTime)
	{
		Time slotTime;
		try{
			slotTime.set_from_iso(this->lineBox->text().latin1());
		}
		catch(...){return 0;}

		return  slotTime.get_time_rep();
	}
	else return 0;
}

/// \fn void QguWLinput::setDouble()
/// \brief Method that sets the slot text as time string from double if acceptTime is true 
/// \param     None
/// \exception None
/// \return    none
///
void QguWLinput::setDouble(double value)
{
	if (acceptTime)
	{
		Time slotTime;
		slotTime.set_time_rep(value);
		this->lineBox->setText(slotTime.iso_srep().c_str());
	}
	if(acceptNumber)
	{
		this->lineBox->setText(double_to_string(value).c_str());
	}
	return;
}
 
// -------------------------------------------

Dim_var QguWLinput::getSlotArrDim( )
{
	Dim_var vDim;

	// construct dim for selection in ArrayReduce via text() in slot

	dimTripleStack arrTriple = this->parseArrayReduce();
	if(arrTriple.size() == 0)  return vDim;  // no reduction done, return is_nil() object
		
	vector <int> dimList;
	for(int i=0; i<arrTriple.size(); i++){
		if(arrTriple[i][0] == 0 ) {
			// this dim not collapsed
			int elems = arrTriple[i][2] - arrTriple[i][1] + 1; // start > end not yet supported
			dimList.push_back(elems);
		}
	}
	vDim = new Dim(dimList);
	return vDim; 
}


/// \fn void QguWLinput::setText(QString txt)
/// \brief Method that sets the text in the input slot
/// \param     txt of type QString
/// \exception None
/// \return    None
///
void QguWLinput::setText(QString txt)
{
  this->lineBox->setText(txt);
  return;

}

/// \fn void QguWLinput::setTextKeepComp(QString txt)
/// \brief Method that sets the text in the input slot but retains component choice if valid
/// \param     txt of type QString
/// \exception None
/// \return    None
///
void QguWLinput::setTextKeepComp(const QString &txt){
	
	if ( QuSlotDebug() )
	{
		printf( " QguWLinput::setTextKeepComp for  '%s'\n",  txt.latin1());
		fflush(stdout);
	}

	bool compOK = true;
	QString oldText = QString(this->text().latin1()); // deep copy
	int cptr = oldText.find("[");
	int tptr = oldText.find("<");

	// find any tokens to replace
	QString newText = QString(txt.latin1()); // deep copy of dropped name
	QuReplaceTokens(newText, oldText);
	
	if (cptr == -1 && tptr == -1 ){
	   // Empty slot or no component and subset.
	   // Set new object and return
	   if(reduce3Dto1)
	   {
			QString defComp = QuGetDefaultComp(newText, QSAS_profiles->current()->Array3D);
			this->lineBox->setText(defComp);
			this->setEdited(true);
			this->validateInput(defComp);
			return;
	   }
	   if(reduce3DtoScalar)
	   {
			QString defComp = QuGetDefaultComp(newText, QSAS_profiles->current()->Array3DtoScalar);
			this->lineBox->setText(defComp);
			this->setEdited(true);
			this->validateInput(defComp);
			return;
	   }
	   
	   this->lineBox->setText(newText);
	   this->setEdited(true);
	   this->validateInput(newText);
	   return;
	}

	int ptr = oldText.find("[");
	if (ptr == -1) 	oldText = "";
	else {
	  oldText.remove(0, ptr-1);
	  ptr = oldText.find("]");
	  oldText.truncate(ptr+1);
	}

    QdObject_var dobj = QuGetObjByName(newText);

	bool vectorDrop = false;
	compOK = compatibleComp(dobj, oldText, vectorDrop);

	int maxSeq = 0;
	if( !dobj.is_nil() ) maxSeq = dobj->sequence_size();

	// now replace with components if still valid
	if( compOK ) newText +=  oldText;
	else
	{
		// use default component
		if(reduce3Dto1)
		{
			newText = QuGetDefaultComp(newText, QSAS_profiles->current()->Array3D);
		}
		else if(reduce3DtoScalar)
		{
			newText = QuGetDefaultComp(newText, QSAS_profiles->current()->Array3DtoScalar);
		}
		else if(vectorDrop)
		{
			if(!acceptVectorObject && !acceptVectorObject_TSonly) newText += R_STR_CMP;
		}
	
	}	

	this->lineBox->setText(newText);
	this->setEdited(true);
	this->validateInput(newText);
	emit dropDetected(txt);

}

// ------------------------------------------------------
 
bool QguWLinput::compatibleComp(QdObject_var dobj, const QString &comp, bool &nowVector){
	const char *cmp = comp.latin1();
	bool vectorComp=false;
	nowVector = false;

	if(strcmp(cmp, R_STR_CMP) == 0)  vectorComp = true;
	if(strcmp(cmp, X_STR_CMP) == 0)  vectorComp = true;
	if(strcmp(cmp, Y_STR_CMP) == 0)  vectorComp = true;
	if(strcmp(cmp, Z_STR_CMP) == 0)  vectorComp = true;
	if(strcmp(cmp, Ld_STR_CMP) == 0) vectorComp = true;
	if(strcmp(cmp, Lr_STR_CMP) == 0) vectorComp = true;
	if(strcmp(cmp, Td_STR_CMP) == 0) vectorComp = true;
	if(strcmp(cmp, Tr_STR_CMP) == 0) vectorComp = true;
	if(strcmp(cmp, Pd_STR_CMP) == 0) vectorComp = true;
	if(strcmp(cmp, Pr_STR_CMP) == 0) vectorComp = true;	

	// most things take time comp
	if(strcmp(cmp, T_STR_START) == 0) return true;
	if(strcmp(cmp, T_STR_CENTRE) == 0) return true;
	if(strcmp(cmp, T_STR_END) == 0) return true;
	int dim_size = QuArrayDims(dobj);
	
	if (dim_size == 0) return false;

	if( QuIsThreeVector(dobj) )
	{
		nowVector = true;
		// look only for Vector comp types
		if(vectorComp) return true;
		return false;
	}

	// must be matrix data 
	if(vectorComp) return false;

	return true;
	
}


// ------------------------------------------------------

void QguWLinput::setAliasKeepComp(const QString &txt){

	QString newTxt = QString(this->text().latin1()); // deep copy
	if (newTxt.isEmpty() ){
	   // set new object and return - should not happen
	   this->lineBox->setText(txt);
	   return;
	}

	int ptr = newTxt.find(" [");
	if (ptr != -1) 	
	{
	   newTxt.remove(0, ptr);
	   newTxt = txt + newTxt;
	}
    else  newTxt = txt;
   
	
	this->lineBox->setText(newTxt);
	return;
}


/// \fn QString QguWLinput::text()
/// \brief Method that returns the text in the input slot
/// \param     None
/// \exception None
/// \return    text of type QString
///
QString QguWLinput::text()
{

  return this->lineBox->text();

}

/// \fn QString QguWLinput::textObject()
/// \brief Method that returns the text in the input slot with component text removed
/// \param     None
/// \exception None
/// \return    text of type QString
///
QString QguWLinput::textObject()
{
   QString baseTxt = QString( this->lineBox->text().latin1() );
   return baseObjectText(baseTxt);
}

/// \fn void QguWLinput::setCursorPosition()
/// \brief Method that sets the cursor position in the underlying lineEdit
/// \param     posn of type int
/// \exception None
/// \return    None
void QguWLinput::setCursorPosition(int posn)
{
   this->lineBox->setCursorPosition(posn);
   return;

}


/// \fn void QguWLinput::setAlignment(int flag)
/// \brief Method that sets the text alignment in the underlying lineEdit
/// \param     flag of type int
/// \exception None
/// \return    None
#if QT_VERSION < 0x040000
void QguWLinput::setAlignment(int flag)
#else
void QguWLinput::setAlignment(Qt::AlignmentFlag flag)
#endif
{
   // Takes int flag of form QT:AlignLeft of QT:AlignAuto
    this->lineBox->setAlignment(flag);

    return;

}


/// \fn void QguWLinput::setMaxLength(int len)
/// \brief Method that sets the max text length in the underlying lineEdit
/// \param     len of type int
/// \exception None
/// \return    None
void QguWLinput::setMaxLength(int len)
{

   this->lineBox->setMaxLength(len);

   return;

}


/// \fn bool QguWLinput::edited()
/// \brief Method that returns the edited flag from this slot
/// \param     None
/// \exception None
/// \return    edited of type bool
bool QguWLinput::edited()
{

   return this->beenEdited;

}



/// \fn void QguWLinput::setEdited(bool hasBeenEdited)
/// \brief Method that sets the edited flag from this slot
/// \param     hasBeenEdited of type bool
/// \exception None
/// \return    None
void QguWLinput::setEdited(bool hasBeenEdited)
{
   this->beenEdited = hasBeenEdited;
   return;

}

/// \fn void QguWLinput::paste()
/// \brief Slot that allows paste into contained QguLineEdit
/// \param     None
/// \exception None
/// \return    none

void QguWLinput::paste()
{
   if ( QuSlotDebug() )
	printf( "QguWLinput::paste this = %ld \n", this );

      this->lineBox->paste();
 }

/// \fn vector <dimTriple *> QguWLinput::parseArrayReduce()
/// \brief Method that parses text in QguLineEdit and finds the array reduction requested
/// \param     None
/// \exception None
/// \return    vector <dimTriple *>

dimTripleStack QguWLinput::parseArrayReduce()
{
   if ( QuSlotDebug() )
	printf( "QguWLinput::parseArrayReduce this = %s \n", this->text().latin1() );

   dimTripleStack arrayTriples;

   QString compTxt = QString( this->lineBox->text().latin1() );
   int ptr = compTxt.find(" [");
   if(ptr == -1) return arrayTriples;  // no component section found, arrayTriples is empty vector

   compTxt.remove(0, ptr+2); // remove " ["

   return getDimTripleStack(compTxt);

 }

/// \fn vector <dimTriple *> QguWLinput::parseTimeSlice()
/// \brief Method that parses text in QguLineEdit and finds the array reduction requested
/// \param     None
/// \exception None
/// \return    vector <dimTriple *>

dimTriple QguWLinput::parseTimeSlice()
{
   if ( QuSlotDebug() )
	printf( "QguWLinput::parseTimeSlice this = %s \n", this->text().latin1() );

   QString cmpTxt = QString(this->lineBox->text().latin1());
   return getSliceTriple(cmpTxt);
 }


/// \fn Vostream& operator<<(Vostream& v_os, QguWLinput& slot)
/// \brief Method that writes the content of slot object (object type/QdObject) to the text stream v_os
/// \param     v_os of type Vostream&
/// \param     slot of type QguWLinput&
/// \exception None
/// \return    None
///
Vostream& operator<<(Vostream& v_os, QguWLinput& slot)
{
   v_os << slot.text().latin1();

   if (slot.with == 1){
     QdObject_var obj = QuGetObjByName( slot.text().latin1() );
     if( !(obj.is_nil()) ){
       v_os << slot.with;
       string type=obj->object_type();
       v_os << type;
       v_os << obj;
     }
     else { 
       slot.with = 0;
       v_os << slot.with;
     }
   }
   else v_os << slot.with;
   
   return v_os;
}


/// \fn Vistream& operator>>(Vistream& v_is, QguWLinput& slot)
/// \brief Method that reads the content of slot object (object type/QdObject) from the text stream v_is
/// \param     v_is of type Vistream&    
/// \param     slot of type QguWLinput& 
/// \exception None
/// \return    Returns Vistream& object
///
Vistream& operator>>(Vistream& v_is, QguWLinput& slot)
{
   string field_str;
   v_is >>  field_str;
   
   v_is >> slot.with;
   if (slot.with == 1){

       string obj_type;
       v_is >> obj_type;
       QdObject_var obj=QdObject::create(obj_type);
       v_is >> obj;
       Name name=field_str;
       name.delete_component(0);
       field_str=name.srep();
     
       QString ObjectName = QwuiPutOnWL(obj,field_str.c_str(),"Data");
       slot.setText(ObjectName);
   }   
   else
       slot.setText(QString(field_str.c_str()));

   return v_is;
}

// QguLIneEdit Class

/// \fn void QguLineEdit::QguLineEdit(QWidget *parent, const char *name)
/// \brief QguLineEdit constructor
///
/// \param    parent of type QWidget*
/// \param    name of type const char*
/// \exception None
/// \return    None
///
QguLineEdit::QguLineEdit(QWidget *parent, const char *name)
      :QLineEdit(parent,name)
{
      // simple class to allow us to overload drop and paste
}

QguLineEdit::~QguLineEdit(){

 }



/// \fn void QguLineEdit::dropEvent(QDropEvent *e)
/// \brief This event handler is called when the drag is dropped on this widget. It has been re-implemented from QWidget
///
/// Sets and validates dropped text
/// \param     e of type QDropEvent *
/// \exception None
/// \return    None
///
void QguLineEdit::dropEvent(QDropEvent *e)
{
	if(QuSlotDebug())
	{
		printf("QguLineEdit::dropEvent %ld \n", this);
		fflush(stdout);
	}

	if( !(WLinput->active) ) return;

	QguListItem *localItem;
	Q_LIST_VIEW_ITEM *qItem;
	QString txt = "";

	QApplication::setOverrideCursor( Qt::waitCursor );
#if QT_VERSION < 0x040000
	qApp->processEvents(2000);
#else
	qApp->processEvents(QEventLoop::AllEvents, 2000);
#endif

	if( WLinput->acceptAlias && WLinput->aliasList != NULL)
	{
#if QT_VERSION < 0x040000
		if ( QTextDrag::decode( e, txt ) )
#else
		if (e->mimeData()->hasText())
#endif
		{
#if QT_VERSION >= 0x040000
			txt=e->mimeData()->text();
#endif
			// Check if this item is included within the local object list
			qItem = WLinput->aliasList->findItem( txt, Alias_);

			if ( qItem == NULL )
			{
				// Item does not exist within the current list of objects
				// Check if this object is coming from the external working list
				QguListItem* qWLitem = qsasWindow->WorkingList->getItemByName( txt );
				if ( qWLitem )
				{
					WLinput->aliasList->insertEntries( NULL, qWLitem );
					Q_LIST_VIEW_ITEM* item = WLinput->aliasList->findItem(qWLitem->nameContext(), Obj_Name_);
					if ( item ) txt = item->text(Alias_).latin1();
				}

			}

			e->acceptAction( false );
			QString newText = QString( txt.latin1() );
			WLinput->setTextKeepComp( newText );
		}
	}
	else
	{
		// Regular slot,  use valid ObjectName in feedback
#if QT_VERSION < 0x040000
		if ( QTextDrag::decode(e, txt) )
#else
		if (e->mimeData()->hasText())
#endif
		{
#if QT_VERSION >= 0x040000
			txt=e->mimeData()->text();
#endif
			if(dragFromList != NULL)
			{
				localItem = dragFromList->getItemByName(txt);
				if(localItem != NULL )
				{
					e->acceptAction(false);
					QString newText = localItem->getWLcontext();
					WLinput->refresh =false;
					WLinput->setTextKeepComp(newText);
				}
				// may be drag from list as alias
				if (dragFromList->columns() > 1)
				{
					localItem = (QguListItem*)(dragFromList->findItem( txt, Alias_));
					if(localItem != NULL )
					{
						e->acceptAction(false);
						QString newText = localItem->getWLcontext();
						WLinput->refresh =false;
						WLinput->setTextKeepComp(newText);
					}
				}
			}
		}
	}
	QApplication::restoreOverrideCursor();
#if QT_VERSION < 0x040000
	qApp->processEvents(2000);
#else
	qApp->processEvents(QEventLoop::AllEvents, 2000);
#endif
	this->setFocus();
	emit hadDropEvent();
}



/// \fn void QguLineEdit::paste()
/// \brief Method that pastes the content of the clipboard into the current line edit object and validates it.
/// \param     None
/// \exception None
/// \return    None
///
void QguLineEdit::paste(){

 if(QuSlotDebug())
 {
	printf("QguLineEdit::paste %ld \n", this);
	fflush(stdout);
 }

 if( !(WLinput->active) ) return;

    // do you mean me?
  if(!(this->hasFocus())){
    return;
  }

   QString txt;
   this->clear();
#if QT_VERSION < 0x040000
   if ( QTextDrag::decode(QApplication::clipboard()->data(), txt) )
#else
   if(QApplication::clipboard()->mimeData()->hasText())
#endif
   {

#if QT_VERSION >= 0x040000
	   txt=QApplication::clipboard()->mimeData()->text();
#endif
      if(WLinput->acceptAlias){

         // special case for alias

         Q_LIST_VIEW_ITEM* qItem = NULL;

	 if ( !txt.isEmpty() && WLinput->aliasList != NULL )
	 {
	   qItem = WLinput->aliasList->findItem( txt, Obj_Name_ );
	   if ( qItem )
	   {
	     txt = QString( qItem->text(Alias_).latin1() );
	   }
	   else
	   {
	     // Look for the object in working list
	     QguListItem* qWLitem = qsasWindow->WorkingList->getItemByName( txt );
             if (  qWLitem )
	     {
	       // Insert object within the list view
	      QguListItem* dropAtItem = ( QguListItem *)
	        WLinput->aliasList->selectedItem();
	      WLinput->aliasList->insertEntries( dropAtItem, qWLitem );

	      Q_LIST_VIEW_ITEM* item =
	        WLinput->aliasList->findItem( qWLitem->nameContext(), Obj_Name_ );
	      if ( item )
	        txt = QString( item->text(Alias_).latin1() );
	     }
	   }
         }
      } // end of alias options

      WLinput->setTextKeepComp( txt );
      this->setFocus();

   }

}

// qgu streams moved here so that QwuiPutOnWL() is available

/// \fn Vostream& operator<<(Vostream& v_os, QguListItem& item)
/// \brief Method that writes the content of QguListItem object to the output text stream v_os.
/// \param     v_os  of type Vostream& 
/// \param     item  of type QguListItem *
/// \exception None
/// \return    None
///  
Vostream& operator<<(Vostream& v_os, QguListItem& item)
{
  v_os << item.with;
  v_os << item.text(Obj_Name_).latin1();
  v_os << item.Lock;
  v_os << item.entryType;
  if(item.entryType==ListTypeEntry)
  {
     if (item.with == 1){
       v_os <<item.WLcontext.latin1();
       QdObject_var obj=item.QguGetObject();
       string type=obj->object_type();
       v_os << type;
       v_os << obj;
     }
  }
  return v_os;
}


/// \fn Vistream& operator>>(Vistream& v_is, QguListItem& item)
/// \brief Method reads the content of QguListItem object from the input text stream v_is.
/// \param     v_is of type Vistream& 
/// \param     item of type QguListItem * 
/// \exception None
/// \return    None
///  
Vistream& operator>>(Vistream& v_is, QguListItem& item)
{
  int with;
  v_is >> with;
  
  string text;
  v_is >> text;
  item.setText(Obj_Name_,text.c_str());
  v_is >> item.Lock;
  int i;
  v_is >> i;
  item.entryType=(QguListType)i;
  //v_is >> (int)item.entryType;
  if(item.entryType==ListTypeEntry)
  {
     
     if(with == 1){
       string name_str;
       v_is >> name_str;
       string obj_type;
       v_is >> obj_type;
       QdObject_var obj=QdObject::create(obj_type);
       v_is >> obj;
       Name name=name_str;
       name.delete_component(0);
       name_str=name.srep();
     
       item.WLcontext = QwuiPutOnWL(obj,name_str.c_str(),"Data");
     }
  }
  return v_is;
}

