
#include "qie_caa.h"
#include "TimeUtils.h"
#include "qdutil.h"

// global block size
int blockSize = 50;

// -----------------------------------------------------------------------------------------
//    QieFile CLASS 
//    -------------------


/// \fn QieFile::QieFile()
/// \brief Constructor of the QieFile class
/// \param     none
/// \exception None
/// \return    None
///  

QieFile::QieFile()

{

  fp = NULL;
  name = NULL;
  is_gz = false;
  includedFile = NULL;
  fileType =0;
  data_until = std::string("");
  start_after = std::string("");
  
   // set defaults
  end_of_line = '\n';
  commentMark = '!';
  recNumberingOn = false;
  attr_delim = ',';
  data_delim = ',';

}

// ------------------------

QieFile::~QieFile()

{

  if(includedFile != NULL) delete includedFile;
  if(fp != NULL) this->closeFile();
  if( name != NULL) free (name);  
  
  if(fileType == _CDF)  CDFlib(CLOSE_, CDF_, NULL_);

}

// ---------------------------

void QieFile::openIncludedFile(const char * fullName, const char *mode){

	std::string incFileName;
	if(includedFile == NULL) {
		// create QieFile object with constructor before attaching file to it
		includedFile = new QieFile();
		includedFile->end_of_line = end_of_line;
		includedFile->commentMark = commentMark;
		includedFile->fileType = fileType;
		
		// use full path in file or same path as parent
		fp=fopen(fullName, "r"); // try name as absolute path
     	if( fp == NULL ) {
			// try with relative path
			incFileName = this->name;
			int ptr = incFileName.rfind("/");
			incFileName.replace(ptr+1, incFileName.length()-ptr-1, fullName);
     	}
     	else{
			// it exists, close and let openFile() do it properly
			fclose(fp);
			incFileName = fullName;
     	}

		includedFile->openFile(incFileName.c_str(), mode);
		if(includedFile->fp == NULL){
			// open failed
			
			QieMsg ("ERROR", "Cannot open data file... \n", incFileName.c_str(), "\n");
			delete includedFile;
			includedFile = NULL;
		}
	}
	else {
		// descend recursively until we find empty pointer 
		// (which will be in current open file)
		includedFile->openIncludedFile(fullName, mode);
	}
 
}
 
// --------------------------------

void QieFile::openFile(const char * fullName, const char *mode, bool overwrite){
	
	// get safe name expanding ~ etc. This uses a static char array, so copy it
	std::string safeName = QieSafePath(&(fullName[0]));
	this->name = (char *) malloc(safeName.length()*sizeof(char) +1);
	strcpy(this->name, safeName.c_str());
	
	if(strcmp(mode,"r") == 0) {
		// open file for reading, allow for gz format
 	    if(strstr(this->name, ".cef.gz") != 0 || strstr(this->name, ".CEF.GZ") != 0){
	    	  this->is_gz = true;
	       char * unzipname = (char *) malloc(strlen(this->name) + 13);
	       strcpy(unzipname, "gzip -cd ");
	       strcat(unzipname, this->name);

            fp=popen(unzipname, mode);
	       free(unzipname);
         }
         else{
	       // open regular file
       	  fp=fopen(this->name, mode);
         }
	}
  	else if( !overwrite){ // replaces '-' option in open call

     	/* try reading to see if file of that name exists */
     	fp=fopen(this->name, "r");
     	if( fp == NULL ) {
       		fp=fopen(this->name, mode);
     	}
     	else{
       		/* it exists, overwrite not allowed, bale out */
       		fclose(fp);
       		fp = NULL;
     	}
   	}
   	else{
     	/* open file with mode supplied */
     	fp=fopen(this->name, mode);
   }
     //  pointer this->fp is NULL if open fails
}

// -----------------------------------------------------
// 
// returns true if something was read OK otherwise false
// Leaves fp ready to read next record
 
bool QieFile::skipNextRecord(){
	
	char c;
     // descend include hierarchy
	if (includedFile != NULL) {
	   if( includedFile->skipNextRecord() ) return true;
	   else {
	       // end of file reached
	       includedFile->closeFile();
		   delete includedFile;
		   includedFile = NULL;
	   }
	}
	
	// read from this File ptr
	if(fp == NULL) return false;
	c=fgetc(fp);
	
	// ignore commented line
	while(c == '!') {
		c=skipRestOfLine();	
		return skipNextRecord();
	}
	
    // skip empty lines
	while (c == ' '  || c == '\t' || c == '\n' || c == '\r') {
		if(c == EOF) return false;  // ran out of data without finding end of line
		c=fgetc(fp);
	}
	
    // start of non-commented, non-empty line
	while (c != end_of_line ){
		if(c == EOF) return false;  // ran out of data without finding end of line
		c=fgetc(fp);	
    }
	return true;
}


// ------------------------------------------------------

 string QieFile::readRecord(){
	string record="";
	
	// concatenates lines of input up to end of line marker
	// to get full cef record or header line
	// returns false if end of file reached or no file to read

     // descend include hierarchy
	if (includedFile != NULL) {
		record = includedFile->readRecord();
		if( record == "" ) {
	       // end of file reached
	       includedFile->closeFile();
		   delete includedFile;
		   includedFile = NULL;
		}
		else {
			return record;	
		}
	}
	
	// read from this File ptr
	if(fp == NULL) return ""; // will be is_nil()

	string ptr;
	char endChar;
	while( (endChar = readLine(ptr)) != EOF){

		if(data_until != std::string("")){
			if( ptr.find(data_until) != -1 ) return ""; // end of input
		}

		// actually add this line to the string
		record += ptr; 

		if(endChar == end_of_line) return record;
    }
	
	// ran out of data without finding end of line
	return "";
}


// ------------------------------------------------------

char QieFile::readLine(string &line_ptr){
	
	string emptyBlock(blockSize, '\0');
	line_ptr = emptyBlock;
	int posn=0;
	char c = assembleLine(line_ptr, posn);
	
	line_ptr = line_ptr.substr(0, posn); // lose nulls on end
	
	return c; 
 }
 

// ------------------------------------------------------

char QieFile::assembleLine(string &line_ptr, int &posn){
	
	// Reads input lines up to '\n' and concatenates on continuation marker
	// Used internally by readRecord() and header reads
    
	// it strips white space from start and adds one space at the end
	// to make concatenation safe.
        string emptyBlock(blockSize, '\0');
	
	
	int available= blockSize;
	
	
	char c = fgetc(fp);
	// remove leading white space and empty lines
	while(c == ' ' || c == '\t' || c == '\n'|| c == '\r') c = fgetc(fp);
	
	// skip comment/blank lines 
	if(  c == commentMark) {
			c = skipRestOfLine();
			return assembleLine(line_ptr, posn); // try next line
	}
	
	while(c != '\n' &&  c != EOF){
	
		if(   c == '\r') {
			// skip Windows cr 
			c=fgetc(fp);
		}
		else if(  c == commentMark) {
			// skip comment after valid line 
			c = skipRestOfLine();
			return c; 
		}
		else if(  c == '\\' ) {
			// skip continuation marker, skip rest of line and add next
			c =skipRestOfLine();
			return assembleLine(line_ptr, posn); // try next line
		}
		else if(c == '\"') {
			// take quoted text verbatim
			addChar(posn, available, line_ptr, emptyBlock, c);
			c=fgetc(fp);
			while( c != '\"'){
				if( c == EOF ) {
					// unclosed quote
					// string may be rest of file, so truncate
					char cutText[301];
					strncpy(&(cutText[0]), line_ptr.c_str(), 100);
					QieMsg("Error",  "Unclosed quotes after... \n", &(cutText[0]), "\n");
					return EOF; 
				}
				addChar(posn, available, line_ptr, emptyBlock, c);
				c=fgetc(fp);
			}
			addChar(posn, available, line_ptr, emptyBlock, c);
			c=fgetc(fp);
		}	
		else if(c == end_of_line) {
			return end_of_line;
		}
		else{
			// add this character and test next
			addChar(posn, available, line_ptr, emptyBlock, c);
			c=fgetc(fp);
		}					
	}
	
	return c; // normally '\n'
 }
 

// -------------------------------------------------------

void QieFile::addChar(int & posn, int &available, string &line_ptr, string &emptyBlock, char c){

	if (posn >= available) {
		line_ptr += emptyBlock;
		available += emptyBlock.size();
	}
	
	line_ptr[posn] = c;
	posn++;
	
	return;
}

// -------------------------------------------------------

int QieFile::skipRestOfLine(){
	// ignore input up to the end of line (or file if sooner)
	
	// look for version stamp and file type identifier first
	// as these flags are escaped with !@
	int c = fgetc(fp);
	if(c == '@'){
		string localLine="";
		while(c != '\n' ){
			if( c != '\r') localLine += c;
			c = fgetc(fp);
		}
		const char *linePtr = localLine.c_str();
		if ( strncmp(linePtr, ".qft", 4) == 0){
         fileType = _QFT;
         QieMsg("WARNING", "version identifier found, qft ");
       }
       else if ( strncmp(linePtr, ".qfd", 4) == 0){
         fileType = _QFD;
         QieMsg("WARNING", "version identifier found, qfd ");
       }
       else if ( strncmp(linePtr, ".cef2", 4) == 0 ){
         fileType = _CAA;
       }
       else if ( strncmp(linePtr, ".cef", 4) == 0 ){
         fileType = _CEF;
         QieMsg("WARNING", "version identifier found,  cef 1 ");
       }
	   
	}
	while(c != '\n' &&  c != EOF) {
		c = fgetc(fp);
	}
	return c;
}

// -------------------------------------------------------

void QieFile::closeFile(){
	
	if(fileType == _CDF) CDFlib(CLOSE_, CDF_, NULL_);  // safe against multiple closes
	
	if (fp == NULL) return; // trap multiple close on flat files
	
	if(is_gz) {
		// close compressed file
 		pclose(fp);
	}
	else {
		// close regular file
 		fclose(fp);
	}
 	fp = NULL;
}

// -------------------------------------------------------

void QieFile::rewindFile(){

	if (fp == NULL) return; // trap multiple close
	if(is_gz) {
		// close compressed file
 		pclose(fp);
		this->openFile(name,"r");
	}
	else {
		// close regular file
 		rewind(fp);
	}
}

// -------------------------------------------------------

void QieFile::setEOL(char eol){
	end_of_line = eol;
	if (includedFile != NULL) includedFile->setEOL(eol);
}

// -------------------------------------------------------

void QieFile::setCMark(char cmark){
	commentMark = cmark;
	if (includedFile != NULL) includedFile->setCMark(cmark);
}

// -------------------------------------------------------

void QieFile::setFileType(int ftype){
	fileType = ftype;
	if (includedFile != NULL) includedFile->setFileType(ftype);
}

// -------------------------------------------------------

void QieFile::setDataUntil(std::string txt){
	data_until = txt.c_str();
	if (includedFile != NULL) includedFile->setDataUntil(txt);
}

// -------------------------------------------------------

void QieFile::setStartAfter(std::string txt){
	start_after = txt.c_str();
	if (includedFile != NULL) includedFile->setStartAfter(txt);
}

// -------------------------------------------------------

void QieFile::setAttrDelim(char c){
	attr_delim = c;
	if (includedFile != NULL) includedFile->setAttrDelim(c);
}

// -------------------------------------------------------

void QieFile::setDataDelim(char c){
	data_delim = c;
	if (includedFile != NULL) includedFile->setDataDelim(c);
}

// -------------------------------------------------------

void QieFile::setRecNumbering(bool recNumOn){
	recNumberingOn = recNumOn;
	if (includedFile != NULL) includedFile->setRecNumbering(recNumOn);
}


// ------------------------------------------------------------------------------------
//    QieData CLASS 
//    -------------------


/// \fn QieData::QieData()
/// \brief Constructor of the QieData class
/// \param     none
/// \exception None
/// \return    None
///  

QieData::QieData()
{
  // start with empty variable and global lists
  QieVars.clear();
  QieGlobals.clear();
  initDebug();
  recordNumEntries = 0;
  time_var_num = -1;
}

QieData::~QieData()
{
  ; 
}

bool QieData::readFlatFile(const char * fullName, const char * headerName)
{
	// utility to query file (open and read header) then get data for whole file
	if( !queryFile( fullName,  headerName)) return false;
	
	// reads in flat records and counts max rec number
        int maxNumRecs = readDataRecords();
	
	
	return true;
	
}

// ----------------------------------------------------------

int QieData::countDataRecords()
{
  // this method is to be called after the header has been read  
  // without rewinding the file (fp is positioned at first record
  
  int nrecs = 0;
  while(inFile.skipNextRecord() ) nrecs++;
  
  return nrecs;
  
}

// ----------------------------------------------------------

bool QieData::queryFile(const char * fullName, const char * headerName)
{
    	if(fullName == NULL) return false;
	if(strcmp(fullName, "") == 0) return false;
	
	// initialise guess of file type if not set by calling routine
	if(inFile.fileType == 0) {
		const char *fullNameInUpper = QieToUpper(fullName);
		if( strstr(fullNameInUpper, ".CEF") != NULL ) inFile.fileType = _CAA; // CEF1 must be set by calling fn or in file
		else if( strstr(fullNameInUpper, ".CDF") != NULL) inFile.fileType = _CDF;
		else if( strstr(fullNameInUpper, ".QFT") != NULL) inFile.fileType = _QFT;
		else if( strstr(fullNameInUpper, ".QFD") != NULL) inFile.fileType = _QFD;
		else inFile.fileType = _CAA; // default
	}
	bool header_OK;
	
	// branch to cdf reading options
	if(inFile.fileType == _CDF) return queryCDFfile(fullName);
	
	// default is CAA syntax cef file
	// first read detached header if provided
	if(headerName != NULL){
		inFile.openFile(headerName, "r");
		if(inFile.fp != NULL){
			header_OK = readHeader();
			inFile.closeFile();	
			inFile.openFile(fullName, "r");
		}
		else{
			inFile.openFile(fullName, "r");
			header_OK = readHeader();
		}
	}
	else{	
		inFile.openFile(fullName, "r");
		if(inFile.fp==NULL) return false;
		header_OK = readHeader();
	}

	// if header was not read, look for header in same directory as file
	if( header_OK ) {
		// Add attributes from helper file
		getAttributeHelper(fullName);
		return true;
	}
	else {
		inFile.closeFile();	
		inFile.openFile(fullName, "r"); //reset ready to read data
		
		// read detached header, so attribute helper file is not loaded
		vector <std::string> hdrList = QieFindHelperFiles(fullName, ".qfh");
		const char *hdrName;
		// cannot tell which headers are applicable, so only use first found
		if(hdrList.size() > 0) hdrName = (hdrList[0]).c_str(); 
		else hdrName = "";
		
		if ( !QieStrIsNULL(hdrName) ){
			inFile.closeFile(); // close data file and open detached header
			inFile.openFile(hdrName, "r");
			if (inFile.fp != NULL) {
				header_OK = readHeader();
				inFile.closeFile();	
				if( !header_OK ){
					QieMsg ("ERROR", "Cannot open header\n", hdrName);
					return false;
				}
				inFile.openFile(fullName, "r");
				return true;
			}
		}
		else{
			QieMsg ("ERROR", "Cannot locate a valid header\n");
			inFile.closeFile(); // close data file
			return false;
		}	
	}
	
	// This leaves file open and positioned ready to read data	
	
}

// ----------------------------------------------------------

void QieData::getAttributeHelper(const char * fullName)
{
	vector <std::string> helperList = QieFindHelperFiles(fullName, ".qat");

	for (int i=0; i<helperList.size(); i++){
		const char *helperName = (helperList[i]).c_str();

		if ( !QieStrIsNULL(helperName) ){
			attrFile.openFile(helperName, "r");
			if (attrFile.fp != NULL) {
				QieMsg ("INFO:", helperName, "being read for additional attribute information\n"); 
				readAttrHelper();
				attrFile.closeFile();
			}
		}
	}
	return;
} 

// ----------------------------------------------------------

void QieData::readAttrHelper()
{
	char endOfRecord = '\n';
    
 	// CAA defaults may not be written in the file, so set in advance
	attrFile.attr_delim = ',';
	attrFile.data_delim = ',';
	
	// note: readLine skips comment lines 
	std::string nextLine;
	while(attrFile.readLine(nextLine) != EOF){
		
		if( !parseLine(nextLine) ){
			// detect (usually) data lines with no header or Start_data
			if( isdigit(parameter[0]) ) break; // non-fatal error
			else QieMsg( "WARNING", "unidentified input line ", nextLine.c_str());
			continue;
		}
		
		
		if(strncmp(QieToUpper(parameter.c_str()), "START_VARIABLE", 10) == 0) {

			int vnum = findVarNumByName('V', value.c_str());
			// keep reading even if var is not in file being read to skip var block
			
			// read attribute info for this variable
			while(attrFile.readLine(nextLine) != EOF){
				if( !parseLine(nextLine) ){
					QieMsg( "WARNING", "unidentified input line ", nextLine.c_str());
					continue;
				}
				if(strncmp(QieToUpper(parameter.c_str()), "END_VARIABLE", 10) == 0) break;
				if(vnum >= 0){
					 /* read attribute */
					QdObject_var nextAttr;
					QieVariable *V = &(QieVars[vnum]);
					try{
						nextAttr = this->QieText_to_Attr(*V, value); // V needed for type and dimensions
					}
					catch(Exception &e){
						QieMsg( "WARNING", "failed to make attribute from: ", value.c_str());
					}
		 			if( !nextAttr.is_nil() ){
						// note pushing back on end of attribute stack
						// will cause this value to REPLACE that in the file if present
						V->attrList.push_back(nextAttr);
						V->attrNameList.push_back(string(parameter.c_str()));
					}
					
				}
			}
		}
		
		
		if(strncmp(QieToUpper(parameter.c_str()), "COMMENT_MARKER", 14) == 0) {
			attrFile.setCMark(value[0]);
			continue;
		}
		
		if(strncmp(QieToUpper(parameter.c_str()), "ATTRIBUTE_DELIMITER", 19) == 0) {
			attrFile.setAttrDelim(QieGetDelim(value));
			continue;
		}
		if(strncmp(QieToUpper(parameter.c_str()), "DATA_DELIMITER", 14) == 0) {
			attrFile.setDataDelim(QieGetDelim(value));
			continue;
		}	

	} // end while
 
}

// ----------------------------------------------------------

bool QieData::readHeader()
{
	int j;
	int i;
	long n;
	char endOfRecord = '\n';
    // initialise some traps
	year_offset = -1;  /* acts as trap if YY format used */
	strictMonotonic = false;
	yp_last=-2000000000;

 		// CAA defaults may not be written in the file, so set in advance
	inFile.attr_delim = ',';
	inFile.data_delim = ',';
	
	// note: readLine skips comment lines 
	std::string nextLine;
	while(inFile.readLine(nextLine) != EOF){
		if(inFile.start_after != std::string("") && nextLine.find(inFile.start_after)) {
			QieMsg( "INFO", "Start reading data after line =", nextLine.c_str());
			break; // located start of data, stop reading header
		}
		
		if( !parseLine(nextLine) ){
			// detect (usually) data lines with no header or Start_data
			//if( isdigit(parameter[0]) ) break; // non-fatal error
		   // else QieMsg( "WARNING", "unidentified input line ", nextLine.c_str());
			//continue;
			return false;
		}
		if(strcmp(QieToUpper(parameter.c_str()), "FILE_FORMAT_VERSION") == 0 && 
		   strstr(QieToUpper(value.c_str()), "CEF-2") != NULL) {
				inFile.setFileType(_CAA);
				continue;
		}
	  
		if(strncmp(QieToUpper(parameter.c_str()), "START_DATA", 10) == 0) {
			QieMsg("INFO", value.c_str(), "records expected");
			break;
		}
		
		if(strncmp(QieToUpper(parameter.c_str()), "DATA_UNTIL", 10) == 0) {
			if(inFile.fileType == 0) inFile.setFileType(_CAA);
			inFile.setDataUntil(QieStripQuotes(value));
			break;
		}
		if(strncmp(QieToUpper(parameter.c_str()), "START_VARIABLE", 10) == 0) {
			if( this->readVariable(value.c_str()) ) continue;
			else return false;
		}
		if(strncmp(QieToUpper(parameter.c_str()), "START_META", 10) == 0) {
			if( this->readGlobal(value.c_str()) ) continue;
			else return false;
		}
		if(strncmp(QieToUpper(parameter.c_str()), "YY_OFFSET", 9) == 0) {
			year_offset = atol(value.c_str()); // global year_offset 
			if(year_offset == 0 && value.find('0') == -1 ) {
				QieMsg("ERROR", "Bad year offset = ", value.c_str());
				return false;
			}
			continue;
		}
		if(strncmp(QieToUpper(parameter.c_str()), "YEAR_INCREASING", 17) == 0) {
			if(value[0] == 'y' || value[0] == 'Y')
			strictMonotonic = true;
			continue;
		}
		if(strncmp(QieToUpper(parameter.c_str()), "FILE_TYPE", 9) == 0) {
			if(value[0] == 't' || value[0] == 'T' ){
				if(inFile.fileType != _CAA) inFile.setFileType(_QFT);
			}
			else if(value[0] == 'd' || value[0] == 'D' ){
				if(inFile.fileType != _CAA) inFile.setFileType(_QFD);
			}
			continue;
		}
		if(strncmp(QieToUpper(parameter.c_str()), "COMMENT_MARKER", 14) == 0) {
			inFile.setCMark(value[0]);
			continue;
		}
		if(strncmp(QieToUpper(parameter.c_str()), "START_AFTER", 11) == 0) {
			inFile.setStartAfter(QieStripQuotes(value));
			continue;
		}
		if(strncmp(QieToUpper(parameter.c_str()), "ATTRIBUTE_DELIMITER", 19) == 0) {
			inFile.setAttrDelim(QieGetDelim(value));
			continue;
		}
		if(strncmp(QieToUpper(parameter.c_str()), "DATA_DELIMITER", 14) == 0) {
			inFile.setDataDelim(QieGetDelim(value));
			continue;
		}	
		if(strncmp(QieToUpper(parameter.c_str()), "END_OF_RECORD_MARKER", 20) == 0) {
			if (value[0] != ' ') {
				std::string delimValue = QieStripQuotes(value);
				if( delimValue[0] == '\n' || delimValue[0] == '\0') 
					endOfRecord = '\n';
				else 
					endOfRecord = QieGetDelim(delimValue);
			}
			continue;

		}
		if(strncmp(QieToUpper(parameter.c_str()), "RECORD_NUMBERING", 16) == 0) {
			if(value.find("on") != -1 && inFile.fileType != _CAA ) inFile.setRecNumbering(true);
			else inFile.setRecNumbering(false);
			continue;
		}
		if(strncmp(QieToUpper(parameter.c_str()), "FILE_NAME", 9) == 0) {
			QieMsg("INFO", "file: ", value.c_str());
			continue;
		}

	} // end while
 
	// validate cef 2 inputs found 
	if(inFile.fileType == _CAA){
		if( inFile.data_until == std::string("") )
			QieMsg( "WARNING", "No DATA_UNTIL parameter found in header. ", " Continuing to attempt read" );
		if( inFile.start_after != std::string("") )
			QieMsg( "WARNING", "START_AFTER parameter should not be used in CAA file. ", " Continuing to read");
		if(inFile.commentMark != '!')
			QieMsg( "WARNING", "CEF 2 format: Comment marker not exclamation mark '!'. ", " Continuing to read" );
		if(inFile.attr_delim != ',')
			QieMsg( "WARNING", "CEF 2 format: Attribute delimiter must be a comma. ", " Continuing to read" );
		if( inFile.data_delim != ',')
			QieMsg( "WARNING", "CEF 2 format: Data delimiter must be a comma. ", " Continuing to read" );
		if (inFile.recNumberingOn)
			QieMsg("WARNING", "CEF 2 format: Record numbering is not supported in CAA", " Continuing to read" );
   	}


	inFile.setEOL( endOfRecord ); // only set this AFTER the header has been read as only records use end_of_line
//   QieFTinit_parsers();  // initalise parsers for SJS free time routines 

        // ensure there is a DEPEND_0 set it time line exists
	ensureDependZero();
	
	return true;
} 

// ------------------------------------------------

bool QieData::readGlobal(const char *globalAttrName)
{
   	// create new QdObjectSeq_var for this global
	// Note globals are sequences of QdObject since
	// globals may change data type in cdf
	// data type is attached to each object
	
   	vector <QdObject_var> GobjList;
	std::string gattr_datatype = "CHAR";
	std::string nextLine;
  
    QdObject_var Gobj;
	
	QdString_var Gname = new QdString(globalAttrName);

   	// read the global block 
   	while(inFile.readLine(nextLine) != EOF){
		if( !parseLine(nextLine) ) {
		      QieMsg( "WARNING", "Failed to parse line in Global block =", nextLine.c_str());
			 // failed to parse line
			continue;
		}
		
       	if(strncmp(QieToUpper(parameter.c_str()), "END_META", 8) == 0){
			if(*Gname != value) {
		      	QieMsg( "WARNING", "Global block start/end names differ", Gname->c_str(), value.c_str());
         	 	}
          	break;
       	}

       	if(strncmp(QieToUpper(parameter.c_str()), "G_ATTR_DATA_TYPE", 16) == 0 ||
        	strncmp(QieToUpper(parameter.c_str()), "VALUE_TYPE", 10) == 0){
				gattr_datatype = QieToUpper(value.c_str());  // deep copy
        	  	continue;
		}

		if(strncmp(QieToUpper(parameter.c_str()), "ENTRY", 5) == 0){

			QdObject_var Gentry;
			int nEntries;
			QdStringSeq_var EntryList = parseEntries(value, inFile.attr_delim);
			
			if( EntryList.is_nil() ){
		      	QieMsg( "WARNING", "Global block unable to parse entry =",value.c_str());
				continue;
			}
			
			nEntries = EntryList->size();
			if(nEntries == 1){
				// single entry
				if(gattr_datatype == "CHAR"  ){
					QdString_var Sentry = new QdString(QieStripQuotes(value));
					Gentry = Sentry;
				}
				else if(gattr_datatype == "DOUBLE"){
					QdRScalar_var Dentry = new QdRScalar(atof((QieStripQuotes(value)).c_str()));
					Gentry = Dentry;
				}
				else if(gattr_datatype == "FLOAT"){
					QdRScalar_var Dentry = new QdRScalar(atof((QieStripQuotes(value)).c_str()));
					Gentry = Dentry;
				}
				else if(gattr_datatype == "INT"){
					QdRScalar_var Dentry = new QdRScalar(atof((QieStripQuotes(value)).c_str()));
					Gentry = Dentry;
				}
				else if(gattr_datatype == "BYTE"){
					QdRScalar_var Dentry = new QdRScalar(atof((QieStripQuotes(value)).c_str()));
					Gentry = Dentry;
				}
				else if(gattr_datatype == "ISO_TIME"){
					QdString_var Sentry = new QdString(QieStripQuotes(value));
					Gentry = Sentry;
				}
				else {
					// use char, but should never get here as default is set
		    		  	QieMsg( "WARNING", "Global block value type not recognised", gattr_datatype.c_str());
					QdString_var Sentry = new QdString(QieStripQuotes(value));
					Gentry = Sentry;
				}
			}
			else{
				// arrays of entries
				if(gattr_datatype == "CHAR"  ){
					QdSMatrix_var SAentry = new QdSMatrix(nEntries);
					for(int i=0; i<nEntries; i++){
						(*SAentry)[i] = (QieStripQuotes(EntryList[i])).c_str();
					}
					Gentry = SAentry;
				}
				else if(gattr_datatype == "DOUBLE"){
					QdRMatrix_var DAentry = new QdRMatrix(nEntries);
					for(int i=0; i<nEntries; i++){
						(*DAentry)[i] = strtod((QieStripQuotes(EntryList[i])).c_str(), NULL);
					}
					Gentry = DAentry;
				}
				else if(gattr_datatype == "FLOAT"){
					QdRMatrix_var DAentry = new QdRMatrix(nEntries);
					for(int i=0; i<nEntries; i++){
						(*DAentry)[i] = strtod((QieStripQuotes(EntryList[i])).c_str(), NULL);
					}
					Gentry = DAentry;
				}
				else if(gattr_datatype == "INT"){
					QdRMatrix_var DAentry = new QdRMatrix(nEntries);
					for(int i=0; i<nEntries; i++){
						(*DAentry)[i] = strtod((QieStripQuotes(EntryList[i])).c_str(), NULL);
					}
					Gentry = DAentry;
				}
				else if(gattr_datatype == "BYTE"){
					QdRMatrix_var DAentry = new QdRMatrix(nEntries);
					for(int i=0; i<nEntries; i++){
						(*DAentry)[i] = strtod((QieStripQuotes(EntryList[i])).c_str(), NULL);
					}
					Gentry = DAentry;
				}
				else if(gattr_datatype == "ISO_TIME"){
					QdSMatrix_var SAentry = new QdSMatrix(nEntries);
					for(int i=0; i<nEntries; i++){
						(*SAentry)[i] = (QieStripQuotes(EntryList[i])).c_str();
					}
					Gentry = SAentry;
				}
				else {
					// use char, but should never get here as default is set
		    		  	QieMsg( "WARNING", "Global block value type not recognised", gattr_datatype.c_str());
					QdSMatrix_var SAentry = new QdSMatrix(nEntries);
					for(int i=0; i<nEntries; i++){
						(*SAentry)[i] = QieStripQuotes(EntryList[i]);
					}
					Gentry = SAentry;
				}
			}
			GobjList.push_back(Gentry);
			continue;
		}
	}
	
	// convert entries into appropriate data objects 
	if(GobjList.size() == 0) return false;
	else if(GobjList.size() == 1){
		// single entry
		Gobj =  GobjList[0];
	}
	else{
		// sequence type
		if(gattr_datatype == "CHAR" || gattr_datatype == "ISO_TIME" ){
			QdStringSeq_var SSv = new QdStringSeq(GobjList.size());
			for(int i=0; i<GobjList.size(); i++){
				QdString_var So = QdString_var::narrow(GobjList[i]);
				if( !So.is_nil() ){
					SSv[i] = *So;
				}
				else{
					QieMsg( "WARNING", "inconsistent G attr types in String Seq");
				}
			}
			Gobj = SSv;
		}
		else {
			// all arrays or all scalars, mixed not handled
			QdRScalar_var Rv = QdRScalar_var::narrow(GobjList[0]);
			QdRMatrix_var Mv = QdRMatrix_var::narrow(GobjList[0]);
			if( !Rv.is_nil() ) {
				QdRScalarSeq_var RSv = new QdRScalarSeq(GobjList.size());
				for(int i=0; i<GobjList.size(); i++){
					QdRScalar_var Ro = QdRScalar_var::narrow(GobjList[i]);
					if( !Ro.is_nil() ){
						RSv[i] = *Ro ;
					}
					else{
						QieMsg( "WARNING", "inconsistent G attr types in Scalar Seq");
					}
				}
				Gobj = RSv;
			}
			if( !Mv.is_nil() ) {
				QdRMatrixSeq_var MSv = new QdRMatrixSeq(GobjList.size());
				for(int i=0; i<GobjList.size(); i++){
					QdRMatrix_var Mo = QdRMatrix_var::narrow(GobjList[i]);
					if( !Mo.is_nil() ){
						MSv[i] = *Mo;
					}
					else{
						QieMsg( "WARNING", "inconsistent G attr types in Matrix Seq");
					}
				}
				Gobj = MSv;
			}
			else{
					QieMsg( "WARNING", "cannot handle G attr type");
			}
		} // end of if on data type
	}
	
	if( Gobj.is_nil() ) {
		return false;
	}
	
	// use globalAttrName first as it is QieData.value string from parse and may change
	Gobj->change_xref("name", Gname);
	QdString_var typeXref = new QdString(gattr_datatype.c_str());
	Gobj->change_xref("value_type", typeXref);

	QieGlobals.push_back(Gobj);

	return true;
}

// ------------------------------------------------

bool QieData::readVariable(const char *variableName)
{

	// initialise variable and format first (set defaults)
	QieVariable V;
	V.name = variableName;
	V.col_start = -1;
	V.col_width = -1;
	V.objDim.clear();
//	V.ftpackets.timeformat = NOT_A_TIME; // SJS re FT - safe initialisation
	V.recVary = true;
	V.datatype = "CHAR";
	V.n_posn = recordNumEntries;
	V.cachedFrom = -1;
	V.cachedTo  = -1;
	V.nRecs = -1;
	
	std::string dataValues=" ";
	std::string nextLine;
	int nEntries = 1;

	while(inFile.readLine(nextLine) != EOF){
		if( !parseLine(nextLine) ) {
			QieMsg( "WARNING", "Failed to parse line in Variable block =", nextLine.c_str());
			continue;
		}

		if(strncmp(QieToUpper(parameter.c_str()), "END_VARIABLE", 12) == 0) {
        		if(V.name != value) {
		     	 	QieMsg( "WARNING", "Variable block start/end names differ", V.name.c_str(), value.c_str());
         		}
          		break; // end variable block, now make object
		}

		if(strncmp(QieToUpper(parameter.c_str()), "DATA_TYPE", 9) == 0 ||
		   strncmp(QieToUpper(parameter.c_str()), "VALUE_TYPE", 10) == 0){
				V.datatype = QieToUpper(value.c_str());  // deep copy
				// Note: first time var is default time tags
				if(time_var_num == -1 && value == "ISO_TIME") time_var_num = QieVars.size(); // not pushed back yet
        	  		continue;
		}
		
        	if(strcmp(QieToUpper(parameter.c_str()), "DATA") == 0) {
        	   /* NOVARY variable */
          		 V.recVary = false;
		   	dataValues = value.c_str(); // deep copy for creating object after while loop completes
           		continue;
         	}

         	if(strncmp(QieToUpper(parameter.c_str()), "COLUMN_START", 12) == 0) {
         		V.col_start = atol(value.c_str());
           		if(inFile.fileType == _CAA) QieMsg( "WARNING", "COLUNM_START detected for '", V.name.c_str(), "' being ignored");
           		continue;
         	}

         	if(strncmp(QieToUpper(parameter.c_str()), "COLUMN_WIDTH", 12) == 0) {
			V.col_width = atol(value.c_str());
			if(inFile.fileType == _CAA) QieMsg( "WARNING", "COLUNM_WIDTH detected for '", V.name.c_str(), "' being ignored");
			continue;
        	}
		 
         	if(strncmp(QieToUpper(parameter.c_str()), "IGNORE_ON_READ", 14) == 0) {
           	if (value[0] == 'y' || value[0] == 'Y') V.ignore = true;
           	if(inFile.fileType == _CAA) QieMsg( "WARNING", "non-CAA option IGNORE_ON_READ detected for '",  V.name.c_str(), "', flag being ignored");
          	continue;
        	}

         	if(strcmp(QieToUpper(parameter.c_str()), "TIME_FORMAT") == 0) {
            		QieMsg( "WARNING", "TIME_FORMAT descriptor detected for  ", V.name.c_str(), " not in CEF 2 syntax");
            		if(strncmp(QieToUpper(value.c_str()),"FREE_TIME_FORMAT", 16 ) == 0) {
   //>>>>             V.ftpackets.timeformat = FREE_TIME_FORMAT;
            		}
           		else if(strncmp(QieToUpper(value.c_str()),"ISO", 3 ) == 0){
     //>>>           V.ftpackets.timeformat = ISO;
            		}             // Non-time vars don't have this, and carry initialised NOT_A_TIME value, SJS re FT */
            		else{
				QieMsg( "WARNING", "Unknown time format designation '", value.c_str(), "' being ignored");
            		}
            		continue;
        	}
			 		 
        	if(strncmp(QieToUpper(parameter.c_str()), "MAXSTRLEN", 9) == 0) {
           		// no longer needed
           		if(inFile.fileType == _CAA) QieMsg( "WARNING", "MAXSTRLEN detected for '", V.name.c_str(), "' not in CEF 2 syntax");
           		continue;
         	}
         	if(strncmp(QieToUpper(parameter.c_str()), "SIZES", 5) == 0) {
				QdStringSeq_var EntryList = parseEntries(value, inFile.attr_delim);
				// set up Dim object 
			
				if(EntryList.is_nil() )  continue;  
		
				V.objDim.clear();
				for(int i=0; i<EntryList->size(); i++){
					int Ndim = strtol(EntryList[i].c_str() , NULL, 10);
					if(Ndim == 0) {
						QieMsg( "ERROR", "Variable dimension size zero from '", value.c_str(), "' (remove quotes?)");
						return false;
					}
					if (Ndim == 0) continue;
					nEntries = nEntries * Ndim;
					V.objDim.push_back(Ndim);
				}
				continue;
        	}

        	 /* else read attribute */
		QdObject_var nextAttr;
		try{
			 nextAttr = this->QieText_to_Attr(V, value); // V needed for type and dimensions
		}
		catch(Exception &e){
				QieMsg( "WARNING", "failed to make attribute from: ", value.c_str());
				e.print_msg();
		}
		 if( !nextAttr.is_nil() ){
			V.attrList.push_back(nextAttr);
			V.attrNameList.push_back(string(parameter.c_str()));
		}
	} // end while loop

	// If NRV object, fill with data now

	if(dataValues != " "){
		V.Vobj = QieText_to_Obj(V, dataValues);
	}
	if(V.Vobj.is_nil() ){
		// object must exist to keep xrefs
		QdRScalar_var dummy = new QdRScalar(1.0);
		V.Vobj = dummy;
	}
	// attach attributes. Note: RV variable will be changed when data fetched, so re-copy later
	for(int i=0; i<V.attrList.size(); i++){
	StringRep string_rep;

		V.Vobj->change_xref(V.attrNameList[i].c_str(), V.attrList[i]); // use change_xref as it is safe against duplicate attr in file
	}

	QieVars.push_back(V);
	
	if (V.recVary) recordNumEntries += nEntries;
	return true;
  
}

// ----------------------------------------------------------

QdObject_var QieData::QieText_to_Attr(QieVariable &V, std::string valueStr)
{
	std::string datatype;
	QdObject_var dobj;

	// create data object from valueStr
	
	DataBuffer_var data_buffer; 
	QdStringSeq_var entryList = parseEntries(valueStr, inFile.attr_delim);
	if( entryList->size() == 0) return dobj;
   
	if(entryList[0][0] == '\"' && entryList[0][entryList[0].size()-1] == '\"' ) datatype = "CHAR";
	
	else if(QieIsISOtime(entryList[0])) datatype = "ISO_TIME";
	 
	else{
       // test for numeric data type 
		char * endPtr;
		strtod(entryList[0].c_str(), &endPtr);
		if( entryList[0] == string(endPtr) ) datatype = "CHAR"; // no numeric value found, use char as safe
		else datatype = "DOUBLE"; 
	}
	
	if(datatype == "CHAR")
	{
		StringVDataBuffer *sdb_p = new StringVDataBuffer();
		// if an array set up 1-D matrix with dimensions
		vector<int> dims;
		dims.push_back(entryList->size());
	
		sdb_p->set_dims( dims );
		sdb_p->set_majority( DBUFFER_ROW_MAJOR );
   
		for(int i=0; i<entryList->size(); i++){
			sdb_p->push_back(QieStripQuotes(entryList[i].c_str()));
		}
   
		data_buffer=sdb_p;
		try{
			dobj = Factory::clone(data_buffer);
		}
		catch(Exception &e){
			e.print_msg();
		}
	}

	else if(datatype == "ISO_TIME")
	{		
		if(entryList->size() == 1){
	
			try{
				if( QieIsISOtime(entryList[0]) ){
					QdTime_var tob = new QdTime( Time(entryList[0].c_str()) ); 
					dobj = tob;
				}
			}
			catch(Exception &e){
				QieMsg( "WARNING", "failed to make time object from  ", entryList[0].c_str());
				e.print_msg();
			}
		}
		else if(entryList->size() == 2){
			// assume to be a time interval (NRV) 
			try{
				if(QieIsISOtime(entryList[0]) && QieIsISOtime(entryList[1])  ){
					Time iv_start(entryList[0].c_str());
					Time iv_end(entryList[1].c_str());
					QdTimeInterval_var ti = new QdTimeInterval(iv_start, iv_end); 
	 
					dobj = ti;
				}
			}
			catch(Exception &e){
				QieMsg( "WARNING", "failed to make time object from  ", entryList[0].c_str());
				e.print_msg();
			}
		}
		else{ 
			// treat as time sequence
 			try{
				QdTimeSeq_var ts = new QdTimeSeq();
				for (int i=0; i <= entryList->size(); i++){
					if(QieIsISOtime(entryList[i]) ) 
						ts->push_back( QdTime( Time(entryList[i].c_str()) ) );
				}
   
				dobj = ts;
			}
			catch(Exception &e){
				QieMsg( "WARNING", "failed to make time object from  ", entryList[0].c_str());
				e.print_msg();
			}
			
		}
	}

   	else if(datatype == "DOUBLE" )
	{
		NumericVDataBuffer *ndb_p = new NumericVDataBuffer();

		// if an array set up matrix with dimensions
   
		vector<int> dims;
		dims.push_back( entryList->size() );
		ndb_p->set_dims( dims );
		ndb_p->set_majority( DBUFFER_ROW_MAJOR );
    
		// if not it will default to a Scalar sequence
       
		for(int i=0; i<entryList->size(); i++){
			ndb_p->push_back( strtod(entryList[i].c_str(), NULL) );
		}
		data_buffer=ndb_p;
		try{
			dobj = Factory::clone(data_buffer);
		}
		catch(Exception &e){
			e.print_msg();
		}
	}
	
     return dobj;
}


// ----------------------------------------------------------

QdObject_var QieData::QieText_to_Obj(QieVariable &V, std::string valueStr)
{
	// used for NRV data, uses attr_delim

	QdObject_var dobj;
	
	// create data object from valueStr
	DataBuffer_var data_buffer; 
	QdStringSeq_var entryList = parseEntries(valueStr, inFile.attr_delim);
	
	for(int k=0; k<entryList->size(); k++) 

	if(V.datatype == "CHAR")
	{
		StringVDataBuffer *sdb_p = new StringVDataBuffer();
	
		if( V.objDim.size() > 0) sdb_p->set_dims( V.objDim );
		sdb_p->set_majority( DBUFFER_ROW_MAJOR );
   
		for(int i=0; i<entryList->size(); i++){
			sdb_p->push_back( QieStripQuotes(entryList[i].c_str()) );
		}
   
		data_buffer=sdb_p;
		dobj = Factory::clone(data_buffer);
	}

	else if(V.datatype == "ISO_TIME")
	{		
		if(entryList->size() == 1){
		    try{
				QdTime_var ti = new QdTime( Time(entryList[0].c_str())); 
				dobj = ti;
			}
			catch(Exception &e){
				e.print_msg();
			}
		}
		else if(entryList->size() == 2){
			// assume to be a time interval (NRV) 
			try{
				if(QieIsISOtime(entryList[0]) && QieIsISOtime(entryList[1])  ){
					Time iv_start(entryList[0].c_str());
					Time iv_end(entryList[1].c_str());
					QdTimeInterval_var ti = new QdTimeInterval(iv_start, iv_end); 
	 
					dobj = ti;
				}
			}
			catch(Exception &e){
				e.print_msg();
			}
		}
		else{ 
			// treat as time sequence
 			try{
				QdTimeSeq_var ts = new QdTimeSeq();
				for (int i=0; i <= entryList->size(); i++){
					ts->push_back( Time(entryList[i].c_str()) );
				}
   
				dobj = ts;
			}
			catch(Exception &e){
				e.print_msg();
			}
		}
	}

   	else if(V.datatype == "DOUBLE" || V.datatype == "FLOAT" || V.datatype == "INT" || V.datatype == "BYTE"  )
	{

		NumericVDataBuffer *ndb_p = new NumericVDataBuffer();

		// if an array set up matrix with dimensions
   
		if( V.objDim.size() > 0 ) ndb_p->set_dims( V.objDim );
		ndb_p->set_majority( DBUFFER_ROW_MAJOR );
    
		// if not it will default to a Scalar sequence
		for(int i=0; i<entryList->size(); i++){
			ndb_p->push_back( strtod(entryList[i].c_str(), NULL) );
		}
		data_buffer=ndb_p;
		dobj = Factory::clone(data_buffer);
	}

     return dobj;

}


// ----------------------------------------------------------

void QieData::clearGetThis()
{
	//  sets the getThis flag to false for all vars 	
	for (int i=0; i< QieVars.size(); i++){
		QieVars[i].getThis = false;
	}
}

// ----------------------------------------------------------

void QieData::setGetThis(int vnum)
{
	// recursively sets the getThis flag to true for var and xrefs
	// NOTE: NRV variables are read with header
	
	QieVariable *V = &(QieVars[vnum]);

	if(V->getThis) return; // already set
	
	 V->getThis = true;
	 
	  // go through xrefs and find vars pointed to
	  for(int i=0; i< V->attrList.size(); i++){
		QdString_var depVarName = QdString_var::narrow(V->attrList[i]);
		if( !depVarName.is_nil() ){
			int attr_num = findVarNumByName('V', depVarName->c_str());
			if(attr_num != -1) {
				// found dependent variable
				setGetThis(attr_num);
			}
		}
	  }
}

// ----------------------------------------------------------

bool QieData::initSequence(QieVariable * V, int nRecs)
{
	// determine dimensionality
	V->nEntries = 1;
	V->nRecs = -1; // set flag in case we fail to read data (was set ro record number by qdi for sliders)
	for(int i=0; i<V->objDim.size(); i++){
		V->nEntries = V->nEntries * V->objDim[i];
	}
		
	// initialise sequence 
	if(V->datatype == "CHAR"){
		if( V->nEntries > 1 ) {
			V->stms = new QdSMatrixSeq(nRecs, Dim(V->objDim));		
			V->Vobj = V->stms;
		}
		else{
			V->sts = new QdStringSeq(nRecs);
			V->Vobj = V->sts;
		}
			
	}
	else if (V->datatype == "ISO_TIME"){
		if(V->nEntries == 1) {
			// time sequence
			V->ts = new QdTimeSeq(nRecs);
			V->Vobj = V->ts;
		}
		else{
			// cannot handle time arrays yet
			return false;
		}
	}
	else{
		// everything else stored as double 
		if( V->nEntries > 1 ) {
			V->ms = new QdRMatrixSeq(nRecs, Dim(V->objDim));		
			V->Vobj = V->ms;
		}
		else{
			V->ss = new QdRScalarSeq(nRecs);
			V->Vobj = V->ss;
		}
	}
	V->nRecs = nRecs;
	return true;
}

// ----------------------------------------------------------

bool QieData::readVarData(int vnum, int recStart, int recEndFlag)
{

	clearGetThis(); // ensure no vars are set to be read
	int recEnd;
	if(recEndFlag == -1) recEnd = dataRecords.size() -1 ; // set to  -1 fetches whole range
	else recEnd = recEndFlag;
	
	QieVariable *V = &(QieVars[vnum]);
	// NOTE: NRV variables are read already in queryFile() but not queryCDFfile()
	if ( !(V->recVary) && inFile.fileType != _CDF) return true; 
	
	setGetThis(vnum); // recursively sets the getThis flag to true for var and record varying xrefs (incl time tags)
	
	// create list of vars to fetch and initialise
	vector <int> getList;
	int nRecs = recEnd - recStart + 1;
	for(int i=0; i<QieVars.size(); i++){
		if(QieVars[i].getThis){

			if(inFile.fileType == _CDF){
			 	getList.push_back(i);
			 }
			 else if( QieVars[i].recVary){
				// excludes NRV vars for flat files as read in header
				
				// Use cached object if possible works for all file types
				if(QieVars[i].cachedFrom == recStart && QieVars[i].cachedTo == recEnd) {
					// variable already read for this exact range
					continue; 
				}
				else if( QieVars[i].cachedFrom != -1 && QieVars[i].cachedFrom <= recStart && QieVars[i].cachedTo >= recEnd){
					// Take G slice
					QieVars[i].Vobj = QuGetSubset(QieVars[i].Vobj, recStart - QieVars[i].cachedFrom, recEnd - QieVars[i].cachedFrom);
					QieVars[i].cachedFrom = recStart;
					QieVars[i].cachedTo = recEnd;

					if( !(QieVars[i].Vobj.is_nil()) )continue; // else fall through and add to list to get
				}	
				else {
					// fall through and add to list of vars to get
					QieVars[i].cachedFrom = recStart;
					QieVars[i].cachedTo = recEnd;
				}
				// not cached if we get here
				if( initSequence(&(QieVars[i]), nRecs) ) getList.push_back(i);
			}
		}
	}

	if(inFile.fileType == _CDF) {
	   // fetch data for variables from CDF file
	   getCDFvarDataList(getList, recStart, recEndFlag);
	}
	else{
		// get records requested for appropriate data type
		int recNum=0;
		for( int recNow=recStart; recNow<dataRecords.size(); recNow++ ){
			currentRecord = dataRecords[recNow];

			// locate start and length of each token in record
			if( !findEntries() ) {
				QieMsg ("ERROR", "failed to find entries in record\n");
				continue;
			}
			for(int nv=0; nv<getList.size(); nv++){
				if ( !(getVarRecordEntry(getList[nv], recNum)) ) {
					// don't continue if data itself fails, continue for metadata
					if(QieVars[getList[nv]].name == QieVars[vnum].name) {
						return false;
					}
				}
			}

			recNum++;
			if(recNow == recEnd) break;
		
		} // end while over records
	
		// The variables needed are read into their Vobj pointers for the records requested
		// The object must be put on the WL and the xrefs attached
	
		for(int i=0; i<QieVars.size(); i++){
			if(QieVars[i].getThis ){
				// also gets attrs for NRV vars
				attachXrefs(&(QieVars[i]));
			}
		}
	}

	return true;
}		


// ------------------------------------------------

void QieData::attachXrefs(QieVariable *V){

	// attach attributes. Note: RV variable will be changed when data fetched, so re-copy later
	
	for(int i=0; i<V->attrList.size(); i++){
		// skip Themis xrefs for epoch no longer needed
		if( V->attrNameList[i] == "VIRTUAL" )   continue;
		if( V->attrNameList[i] == "COMPONENT_0")  continue;
		if( V->attrNameList[i] == "COMPONENT_1" ) continue;
		
		// Find attributes that point to vars
		QdString_var depVarName = QdString_var::narrow(V->attrList[i]);
		if( depVarName.is_nil() ){
		 	// not text so just attach xref
			V->Vobj->change_xref(V->attrNameList[i].c_str(), V->attrList[i]); // use change_xref, safe against duplicates
		}
		else{

			int attr_num = findVarNumByName('V', depVarName->c_str());
			if(attr_num != -1 && QieVars[attr_num].name != V->name && // skip self reference, e.g. Ptr_var_name,
			   V->name != V->attrNameList[i]) { // also skip attr with same text as var name
			
				// use var as attribute
				QdString_var ptrVarName = new QdString(QieVars[attr_num].name.c_str());
				QieVars[attr_num].attrList.push_back(ptrVarName);
				QieVars[attr_num].attrNameList.push_back(string("Ptr_var_name"));
				V->Vobj->change_xref(V->attrNameList[i].c_str(), QieVars[attr_num].Vobj);
				if(V->attrNameList[i] == std::string("DEPEND_0")) {
					// also set the time tags
					QdTimeSeq_var tso = QdTimeSeq_var::narrow(QieVars[attr_num].Vobj);
					if( !tso.is_nil() )V->Vobj->change_xref(TIME_TAGS, tso);
				}
			}
			else{
				// fall back to attribute
				V->Vobj->change_xref(V->attrNameList[i].c_str(), V->attrList[i]); // use change_xref, safe against duplicates
			}
		}
	}

}


// ------------------------------------------------

bool QieData::getVarRecordEntry(int nToGet, int recNum){

	QieVariable *V = &(QieVars[nToGet]);
	
	if(V->datatype == "CHAR")
	{

  
		if( V->nEntries > 1 ) {
					
			if( !getEntriesChar(V->n_posn, V->nEntries, recNum,  V->stms) ){
				QieMsg ("ERROR", "unable to extract variable char array data for\n", V->name.c_str());
				return false;
			}
		}
		else{				

			if( !getEntriesChar(V->n_posn, recNum,  V->sts) ){
				QieMsg ("ERROR", "unable to extract variable char data for\n", V->name.c_str());
				return false;
			}
		}

	}
	else if(V->datatype == "ISO_TIME"){
		if( !getEntriesTime(V->n_posn, recNum,  V->ts) ){
			QieMsg ("ERROR", "unable to extract variable time data for\n", V->name.c_str());
			return false;
		}
		
	}
	else{

		// treat as double
		if( V->objDim.size() == 0 || (V->objDim.size() == 1 && V->objDim[0] == 1) ) {				
		   // scalar
			if( !getEntriesNumeric(V->n_posn, recNum, V->ss) ){
				QieMsg ("ERROR", "unable to extract variable numeric data for\n",  V->name.c_str());
				return false;
			}
		}
		else{
		    // matrix
			if( !getEntriesNumeric(V->n_posn, V->nEntries, recNum, V->ms) ){
				 QieMsg ("ERROR", "unable to extract variable numeric array data for ", V->name.c_str());
				return false;
			}
		}
	}
	return true;
}		

// ------------------------------------------------

bool QieData::parseLine(string &nextLine)
{
  int ptr;
  int startAt=0;
  int tokLen;

  ptr  = nextLine.find("="); // note we remove white space on param later 
  if(ptr == -1) {
	QieMsg ("ERROR", "No '=', cannot parse line\n", nextLine.c_str());
    return false;
  }
  if(ptr == 0) {
	QieMsg ("ERROR", "Line starts '=', cannot parse line\n", nextLine.c_str());
    return false;
  }

  /* remove leading white space in param */
  while( (nextLine[startAt] == ' ' || nextLine[startAt] == '\t') ) startAt++;
 
   /* remove possible trailing white space in param */
  tokLen = ptr-1;  
  while( nextLine[tokLen] == ' ' || nextLine[tokLen] == '\t' ) tokLen--;
  tokLen++;
  tokLen -= startAt;
  if(tokLen <= 0) {
	QieMsg ("ERROR", "No parameter, cannot parse line\n", nextLine.c_str());
    return false;
  }
  parameter = nextLine.substr(startAt, tokLen);
  
  ptr ++;
  // remove leading white space in value 
  while( nextLine[ptr] == ' ' || nextLine[ptr] == '\t')  ptr++;
 
  // remove newline and possible trailing white space in value 
  tokLen = nextLine.length()-1;  
  
  while( nextLine[tokLen] == ' ' || nextLine[tokLen] == '\t' || nextLine[tokLen] == '\n'   || nextLine[tokLen] == '\r'   ) tokLen--;
  tokLen++;
  tokLen -= ptr;
  if(tokLen <= 0) {
	// replace empty RHS with single space
	value = " ";
  }
  else{
	value = nextLine.substr( ptr, tokLen);
  }

  return true;

} 

// -----------------------------------------------
bool QieData::findEntries(){

	vector <int> tokS(recordNumEntries);
	tokenStart = tokS;
	
	// locates start and length of each token
	// tokenStart.clear();
	//tokenLen.clear();
	
	if(currentRecord == "") return false;
	int ptr = 0;
	int startAt=0;
	int NextLineLen = currentRecord.length();
	int i=0;
	
	while(ptr < NextLineLen) {
		// find next delimiter or end of line 
		while(ptr < NextLineLen && currentRecord[ptr] != inFile.data_delim) ptr++;
		tokenStart[i] = startAt;
		currentRecord[ptr] = '\0'; // replace delimiters with nulls
		//tokenLen.push_back(ptr - startAt);
		ptr++;
		startAt = ptr;
		i++;
	}
	
	return true;
	
}


// -----------------------------------------------
/*
bool QieData::getEntriesChar(int nStart, int nEntries, int recNum, QdSMatrixSeq_var stms){
	
	if(nStart + nEntries > tokenStart.size() ) return false;
	
	try{
		valarray <string> v_data(nEntries);
		// peel off entries using shallow copies
		for(register int i=0; i<nEntries; i++){
			std::string token = currentRecord.substr(tokenStart[nStart+i],  tokenLen[nStart+i]);
			v_data[i] = token.c_str();
		}
		stms[recNum].data()[stms[recNum].getgslice()]=v_data;
		return true;	
	}
	catch(Exception &e){
		e.print_msg();
		return false;
	}
	
}
*/
// -----------------------------------------------

bool QieData::getEntriesChar(int nStart, int nEntries, int recNum, QdSMatrixSeq_var stms){
	
	if(nStart + nEntries > tokenStart.size() ) return false;
	
	try{
		valarray <string> v_data(nEntries);
		// peel off entries using shallow copies
		for(register int i=0; i<nEntries; i++){
			v_data[i] = string(&(currentRecord.c_str()[tokenStart[nStart+i]]));
		}
		stms[recNum].data()[stms[recNum].getgslice()]=v_data;
		return true;	
	}
	catch(Exception &e){
		e.print_msg();
		return false;
	}
	
}

// -----------------------------------------------
/*
bool QieData::getEntriesChar(int nStart, int recNum, QdStringSeq_var sts){
	
	if(nStart  > tokenStart.size() ) return false;
	
	try{
		std::string token = currentRecord.substr(tokenStart[nStart],  tokenLen[nStart]);
		sts[recNum] = token.c_str();
		return true;
	}
	catch(Exception &e){
		e.print_msg();
		return false;
	}
	
}
*/
// -----------------------------------------------

bool QieData::getEntriesChar(int nStart, int recNum, QdStringSeq_var sts){
	
	if(nStart  > tokenStart.size() ) return false;
	
	try{
		sts[recNum] = string(&(currentRecord.c_str()[tokenStart[nStart]]));
		return true;
	}
	catch(Exception &e){
		e.print_msg();
		return false;
	}
	
}


// -----------------------------------------------
/*
bool QieData::getEntriesNumeric(int nStart, int nEntries, int recNum, QdRMatrixSeq_var ms){
	if(nStart + nEntries > tokenStart.size() ) return false;

	try{	
		valarray<double> v_data(nEntries);
		std::string token;
		// peel off entries using shallow copies
		for(register int i=0; i<nEntries; i++){
			token = currentRecord.substr(tokenStart[nStart+i],  tokenLen[nStart+i]);
			v_data[i] = ( strtod(token.c_str(), NULL) );
		}
		
		ms[recNum].data()[ms[recNum].getgslice()]=v_data;
		return true;		
	
	}
	catch(Exception &e){
		e.print_msg();
		return false;
	}

}
*/

bool QieData::getEntriesNumeric(int nStart, int nEntries, int recNum, QdRMatrixSeq_var ms){
	if(nStart + nEntries > tokenStart.size() ) return false;

	try{	
		valarray<double> v_data(nEntries);

		// peel off entries using shallow copies
		for(register int i=0; i<nEntries; i++){
			v_data[i] = strtod(&(currentRecord.c_str()[tokenStart[nStart+i]]), NULL);
		}
		
		ms[recNum].data()[ms[recNum].getgslice()]=v_data;
		return true;
	
	}
	catch(Exception &e){
		e.print_msg();
		return false;
	}

}

// -----------------------------------------------

bool QieData::getEntriesNumeric(int nStart, int recNum, QdRScalarSeq_var ss){
	if(nStart > tokenStart.size() ) return false;

	try{	
		ss[recNum] = strtod(&(currentRecord.c_str()[tokenStart[nStart]]), NULL);
		
		return true;		
	
	}
	catch(Exception &e){
		e.print_msg();
		return false;
	}

}
// -----------------------------------------------

bool QieData::getEntriesTime(int nStart, int recNum, QdTimeSeq_var ts){
	if(nStart  > tokenStart.size() ) return false;
	
	
	int tok_start = tokenStart[nStart];
	
	try{
		ts[recNum] = Time(&(currentRecord.c_str()[tok_start]));
	}
	catch(Exception &e){
		e.print_msg();
	}
	
	return true;
		
}
// -----------------------------------------------
/*
bool QieData::getEntriesTime(int nStart, int recNum, QdTimeSeq_var ts){
	
	if(nStart  > tokenStart.size() ) return false;
	
	try{
		string token = currentRecord.substr(tokenStart[nStart],  tokenLen[nStart]);
		ts[recNum] = Time(token.c_str());
	}
	catch(Exception &e){
		e.print_msg();
	}
	
	return true;
		
}
*/

// -----------------------------------------------

QdStringSeq_var QieData::parseEntries(std::string inputStr, const char delim){
	
	QdStringSeq_var strSeq = new QdStringSeq();
	int ptr = 0;
	int startAt=0;
	int NextLineLen = inputStr.length();
	int truncateAt;

	while(ptr < NextLineLen) {
	
		std::string token;
		// peel off entries using shallow copies

		while(ptr < NextLineLen && inputStr[ptr] != delim){
			if(inputStr[ptr] == '\"') {
				// ignore anything INSIDE quotes
				ptr++;
				while(inputStr[ptr] != '\"') ptr++;
			}
			ptr++;
		}

		// remove leading white space
		while(inputStr[startAt] == ' ' && startAt < ptr) startAt++;
		
		// remove trailing white space
		truncateAt = ptr-1;
		while (inputStr[truncateAt] == ' ' && startAt < truncateAt) truncateAt--;

		try{
			if( truncateAt - startAt >= 0 ) 	
				strSeq->push_back(QdString(inputStr.substr(startAt, truncateAt - startAt +1).c_str()));
			else
				strSeq->push_back(QdString(" "));
		}
		catch(Exception &e){
			QieMsg( "WARNING", "failed to make string object from  ", inputStr.c_str());
		    e.print_msg();
		}

		ptr++;
		startAt = ptr;
	}
	
	return strSeq;
	
}

// -----------------------------------------------

int QieData::findVarNumByName(const char VorGlist, const char * name){
	if(VorGlist == 'V'){
		for(int i=0; i<QieVars.size(); i++){
			if(QieVars[i].name == std::string(name) ) return i;
		}
	}
	else if(VorGlist == 'G'){
		for(int i=0; i<QieGlobals.size(); i++){
			if(QieGlobals[i]->xref_exists("name")){
				QdString_var Gs = QdString_var::narrow(QieGlobals[i]->get_xref("name"));
				if( !Gs.is_nil() && *Gs == QdString(name) ) return i;
			}
		}
	}
	return -1; // not found
}

// ------------------------------------------------------

int QieData::readDataRecords(){
	dataRecords.clear();
	bool OK=true;

	if(inFile.fileType == _CDF) return maxCDFnumRecs;
	
	// read one record and set block factor
	
	while(OK){
		string record = inFile.readRecord();
		if(record == "" ) break;
		dataRecords.push_back(record);
		if (record.size() > blockSize) blockSize = record.size(); // try to optimise block size
	}

	return dataRecords.size();
}


// --------------------------------------------------------

QdTimeSeq_var QieData::getTTobj(int vnum){
	QdTimeSeq_var ttobj;
	int tt_num = time_var_num; // pick up file default if set

	if(QieVars[vnum].Vobj->xref_exists("DEPEND_0")){
		QdString_var D0 = QdString_var::narrow(QieVars[vnum].Vobj->get_xref("DEPEND_0"));
		if( !D0.is_nil() ){
			tt_num = findVarNumByName('V', D0->c_str());
		}
		if (tt_num == -1) tt_num = time_var_num; // failed, so reset to default
	}
	if (tt_num != -1) {
		ttobj = QieVars[tt_num].ts; // if DEPEND_0 object not time tags will be is_nil()
	}

	return ttobj;
}

void QieData::ensureDependZero(){

	// ensure an attribute called DEPEND_0 exists if time tags found
	for(int i=0; i< QieVars.size(); i++){
		// NRV variables should not need DEPEND_0 adding, and they will
		if( !(QieVars[i].recVary) ) continue;

		for(int ia=0; ia < QieVars[i].attrNameList.size(); ia++){
			if(QieVars[i].attrNameList[ia] == std::string("DEPEND_0") ){
				// if any variable has DEPEND_0 set assume absence is intentional
				return;
			}
		}

		if (time_var_num != -1){
			// set default tt. Note variable may not be read yet so create ptr to var attribute
			QieVars[i].attrList.push_back( new QdString(QieVars[time_var_num].name.c_str() ));
			QieVars[i].attrNameList.push_back(string("DEPEND_0"));
		}
	}
	return;

}
